@@ -2,8 +2,8 @@
 
   array.c -
 
-  $Author: eban $
-  $Date: 2000/09/12 07:08:40 $
+  $Author: knu $
+  $Date: 2002/10/26 15:57:20 $
   created at: Fri Aug  6 09:46:12 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -14,6 +14,7 @@
 
 #include "ruby.h"
 #include "util.h"
+#include "st.h"
 
 VALUE rb_cArray;
 
@@ -22,7 +23,7 @@ VALUE rb_cArray;
 void
 rb_mem_clear(mem, size)
     register VALUE *mem;
-    register size_t size;
+    register long size;
 {
     while (size--) {
 	*mem++ = Qnil;
@@ -32,7 +33,7 @@ rb_mem_clear(mem, size)
 static void
 memfill(mem, size, val)
     register VALUE *mem;
-    register size_t size;
+    register long size;
     register VALUE val;
 {
     while (size--) {
@@ -79,7 +80,7 @@ rb_ary_new2(len)
     if (len < 0) {
 	rb_raise(rb_eArgError, "negative array size (or size too big)");
     }
-    if (len > 0 && len*sizeof(VALUE) <= 0) {
+    if (len > 0 && len*sizeof(VALUE) <= len) {
 	rb_raise(rb_eArgError, "array size too big");
     }
     ary->len = 0;
@@ -119,7 +120,7 @@ rb_ary_new3(n, va_alist)
     long i;
 
     if (n < 0) {
-	rb_raise(rb_eIndexError, "negative number of items(%d)", n);
+	rb_raise(rb_eIndexError, "negative number of items(%ld)", n);
     }
     ary = rb_ary_new2(n<ARY_DEFAULT_SIZE?ARY_DEFAULT_SIZE:n);
 
@@ -194,7 +195,7 @@ rb_ary_initialize(argc, argv, ary)
     if (len < 0) {
 	rb_raise(rb_eArgError, "negative array size");
     }
-    if (len > 0 && len*sizeof(VALUE) <= 0) {
+    if (len > 0 && len*sizeof(VALUE) <= len) {
 	rb_raise(rb_eArgError, "array size too big");
     }
     if (len > RARRAY(ary)->capa) {
@@ -239,7 +240,7 @@ rb_ary_store(ary, idx, val)
     if (idx < 0) {
 	idx += RARRAY(ary)->len;
 	if (idx < 0) {
-	    rb_raise(rb_eIndexError, "index %d out of array",
+	    rb_raise(rb_eIndexError, "index %ld out of array",
 		     idx - RARRAY(ary)->len);
 	}
     }
@@ -278,6 +279,9 @@ rb_ary_push_m(argc, argv, ary)
     VALUE *argv;
     VALUE ary;
 {
+    if (argc == 0) {
+	rb_raise(rb_eArgError, "wrong # of arguments(at least 1)");
+    }
     if (argc > 0) {
 	long len = RARRAY(ary)->len;
 
@@ -285,8 +289,9 @@ rb_ary_push_m(argc, argv, ary)
 	/* make rooms by copying the last item */
 	rb_ary_store(ary, len + argc, argv[argc]);
 
-	if (argc)		/* if any rest */
+	if (argc) {		/* if any rest */
 	    MEMCPY(RARRAY(ary)->ptr + len, argv, VALUE, argc);
+	}
     }
     return ary;
 }
@@ -396,11 +401,8 @@ rb_ary_subseq(ary, beg, len)
     VALUE ary2;
 
     if (beg > RARRAY(ary)->len) return Qnil;
-    if (beg < 0) {
-	len += beg;
-	beg = 0;
-    }
-    if (len < 0) return Qnil;
+    if (beg < 0 || len < 0) return Qnil;
+
     if (beg + len > RARRAY(ary)->len) {
 	len = RARRAY(ary)->len - beg;
     }
@@ -412,6 +414,7 @@ rb_ary_subseq(ary, beg, len)
     ary2 = rb_ary_new2(len);
     MEMCPY(RARRAY(ary2)->ptr, RARRAY(ary)->ptr+beg, VALUE, len);
     RARRAY(ary2)->len = len;
+    RBASIC(ary2)->klass = rb_obj_class(ary);
 
     return ary2;
 }
@@ -528,13 +531,15 @@ rb_ary_replace(ary, beg, len, rpl)
     VALUE ary, rpl;
     long beg, len;
 {
-    if (len < 0) rb_raise(rb_eIndexError, "negative length %d", len);
+    long rlen;
+
+    if (len < 0) rb_raise(rb_eIndexError, "negative length %ld", len);
     if (beg < 0) {
 	beg += RARRAY(ary)->len;
     }
     if (beg < 0) {
 	beg -= RARRAY(ary)->len;
-	rb_raise(rb_eIndexError, "index %d out of array", beg);
+	rb_raise(rb_eIndexError, "index %ld out of array", beg);
     }
     if (beg + len > RARRAY(ary)->len) {
 	len = RARRAY(ary)->len - beg;
@@ -546,16 +551,17 @@ rb_ary_replace(ary, beg, len, rpl)
     else if (TYPE(rpl) != T_ARRAY) {
 	rpl = rb_ary_new3(1, rpl);
     }
+    rlen = RARRAY(rpl)->len;
 
     rb_ary_modify(ary);
     if (beg >= RARRAY(ary)->len) {
-	len = beg + RARRAY(rpl)->len;
+	len = beg + rlen;
 	if (len >= RARRAY(ary)->capa) {
 	    RARRAY(ary)->capa=len;
 	    REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
 	}
 	rb_mem_clear(RARRAY(ary)->ptr+RARRAY(ary)->len, beg-RARRAY(ary)->len);
-	MEMCPY(RARRAY(ary)->ptr+beg, RARRAY(rpl)->ptr, VALUE, RARRAY(rpl)->len);
+	MEMCPY(RARRAY(ary)->ptr+beg, RARRAY(rpl)->ptr, VALUE, rlen);
 	RARRAY(ary)->len = len;
     }
     else {
@@ -565,18 +571,18 @@ rb_ary_replace(ary, beg, len, rpl)
 	    len = RARRAY(ary)->len - beg;
 	}
 
-	alen = RARRAY(ary)->len + RARRAY(rpl)->len - len;
+	alen = RARRAY(ary)->len + rlen - len;
 	if (alen >= RARRAY(ary)->capa) {
 	    RARRAY(ary)->capa=alen;
 	    REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
 	}
 
 	if (len != RARRAY(rpl)->len) {
-	    MEMMOVE(RARRAY(ary)->ptr+beg+RARRAY(rpl)->len, RARRAY(ary)->ptr+beg+len,
+	    MEMMOVE(RARRAY(ary)->ptr+beg+rlen, RARRAY(ary)->ptr+beg+len,
 		    VALUE, RARRAY(ary)->len-(beg+len));
 	    RARRAY(ary)->len = alen;
 	}
-	MEMCPY(RARRAY(ary)->ptr+beg, RARRAY(rpl)->ptr, VALUE, RARRAY(rpl)->len);
+	MEMMOVE(RARRAY(ary)->ptr+beg, RARRAY(rpl)->ptr, VALUE, rlen);
     }
 }
 
@@ -791,7 +797,7 @@ struct inspect_arg {
     VALUE arg1, arg2;
 };
 
-VALUE
+static VALUE
 inspect_call(arg)
     struct inspect_arg *arg;
 {
@@ -844,7 +850,9 @@ rb_inspecting_p(obj)
 {
     VALUE inspect_tbl;
 
-    if (!inspect_key) return Qfalse;
+    if (!inspect_key) {
+	inspect_key = rb_intern("__inspect_key__");
+    }
     inspect_tbl = rb_thread_local_aref(rb_thread_current(), inspect_key);
     if (NIL_P(inspect_tbl)) return Qfalse;
     return rb_ary_includes(inspect_tbl, rb_obj_id(obj));
@@ -903,9 +911,8 @@ rb_ary_reverse(ary)
 
     while (p1 < p2) {
 	tmp = *p1;
-	*p1 = *p2;
-	*p2 = tmp;
-	p1++; p2--;
+	*p1++ = *p2;
+	*p2-- = tmp;
     }
 
     return ary;
@@ -926,6 +933,20 @@ rb_ary_reverse_m(ary)
     return rb_ary_reverse(rb_obj_dup(ary));
 }
 
+int
+rb_cmpint(cmp)
+    VALUE cmp;
+{
+    if (FIXNUM_P(cmp)) return NUM2LONG(cmp);
+    if (TYPE(cmp) == T_BIGNUM) {
+	if (RBIGNUM(cmp)->sign) return 1;
+	return -1;
+    }
+    if (rb_funcall(cmp, '>', 1, INT2FIX(0))) return 1;
+    if (rb_funcall(cmp, '<', 1, INT2FIX(0))) return -1;
+    return 0;
+}
+
 static ID cmp;
 
 static int
@@ -933,24 +954,27 @@ sort_1(a, b)
     VALUE *a, *b;
 {
     VALUE retval = rb_yield(rb_assoc_new(*a, *b));
-    return NUM2INT(retval);
+    return rb_cmpint(retval);
 }
 
 static int
-sort_2(a, b)
-    VALUE *a, *b;
+sort_2(ap, bp)
+    VALUE *ap, *bp;
 {
     VALUE retval;
+    long a = (long)*ap, b = (long)*bp;
 
-    if (FIXNUM_P(*a)) {
-	if (FIXNUM_P(*b)) return *a - *b;
+    if (FIXNUM_P(a) && FIXNUM_P(b)) {
+	if (a > b) return 1;
+	if (a < b) return -1;
+	return 0;
     }
-    else if (TYPE(*a) == T_STRING && TYPE(*b) == T_STRING) {
-	return rb_str_cmp(*a, *b);
+    if (TYPE(a) == T_STRING && TYPE(b) == T_STRING) {
+	return rb_str_cmp(a, b);
     }
 
-    retval = rb_funcall(*a, cmp, 1, *b);
-    return NUM2INT(retval);
+    retval = rb_funcall(a, cmp, 1, b);
+    return rb_cmpint(retval);
 }
 
 static VALUE
@@ -1167,7 +1191,7 @@ rb_ary_replace_m(ary, ary2)
     return ary;
 }
 
-static VALUE
+VALUE
 rb_ary_clear(ary)
     VALUE ary;
 {
@@ -1236,8 +1260,7 @@ rb_ary_plus(x, y)
 {
     VALUE z;
 
-    Check_Type(y, T_ARRAY);
-
+    y = to_ary(y);
     z = rb_ary_new2(RARRAY(x)->len + RARRAY(y)->len);
     MEMCPY(RARRAY(z)->ptr, RARRAY(x)->ptr, VALUE, RARRAY(x)->len);
     MEMCPY(RARRAY(z)->ptr+RARRAY(x)->len, RARRAY(y)->ptr, VALUE, RARRAY(y)->len);
@@ -1249,9 +1272,20 @@ VALUE
 rb_ary_concat(x, y)
     VALUE x, y;
 {
-    Check_Type(y, T_ARRAY);
+    long xlen = RARRAY(x)->len;
+    long ylen;
 
-    rb_ary_push_m(RARRAY(y)->len, RARRAY(y)->ptr, x);
+    y = to_ary(y);
+    ylen = RARRAY(y)->len;
+    if (ylen > 0) {
+	rb_ary_modify(x);
+	if (xlen + ylen > RARRAY(x)->capa) {
+	    RARRAY(x)->capa = xlen + ylen;
+	    REALLOC_N(RARRAY(x)->ptr, VALUE, RARRAY(x)->capa);
+	}
+	MEMCPY(RARRAY(x)->ptr+xlen, RARRAY(y)->ptr, VALUE, ylen);
+	RARRAY(x)->len = xlen + ylen;
+    }
     return x;
 }
 
@@ -1280,6 +1314,9 @@ rb_ary_times(ary, times)
 	MEMCPY(RARRAY(ary2)->ptr+i, RARRAY(ary)->ptr, VALUE, RARRAY(ary)->len);
     }
 
+    OBJ_INFECT(ary2, ary);
+    RBASIC(ary2)->klass = rb_obj_class(ary);
+
     return ary2;
 }
 
@@ -1325,6 +1362,7 @@ rb_ary_equal(ary1, ary2)
 {
     long i;
 
+    if (ary1 == ary2) return Qtrue;
     if (TYPE(ary2) != T_ARRAY) return Qfalse;
     if (RARRAY(ary1)->len != RARRAY(ary2)->len) return Qfalse;
     for (i=0; i<RARRAY(ary1)->len; i++) {
@@ -1355,7 +1393,8 @@ rb_ary_hash(ary)
     VALUE ary;
 {
     long i;
-    int n, h;
+    VALUE n;
+    long h;
 
     h = RARRAY(ary)->len;
     for (i=0; i<RARRAY(ary)->len; i++) {
@@ -1421,21 +1460,42 @@ rb_ary_diff(ary1, ary2)
     return ary3;
 }
 
+static VALUE
+ary_make_hash(ary1, ary2)
+    VALUE ary1, ary2;
+{
+    VALUE hash = rb_hash_new();
+    int i, n;
+
+    for (i=0; i<RARRAY(ary1)->len; i++) {
+	rb_hash_aset(hash, RARRAY(ary1)->ptr[i], Qtrue);
+    }
+    if (ary2) {
+	for (i=0; i<RARRAY(ary2)->len; i++) {
+	    rb_hash_aset(hash, RARRAY(ary2)->ptr[i], Qtrue);
+	}
+    }
+    return hash;
+}
+
 static VALUE
 rb_ary_and(ary1, ary2)
     VALUE ary1, ary2;
 {
-    VALUE ary3;
+    VALUE hash;
+    VALUE ary3 = rb_ary_new();
     long i;
 
     ary2 = to_ary(ary2);
-    ary3 = rb_ary_new();
+    hash = ary_make_hash(ary2, 0);
+
     for (i=0; i<RARRAY(ary1)->len; i++) {
-	if (rb_ary_includes(ary2, RARRAY(ary1)->ptr[i])
-	    && !rb_ary_includes(ary3, RARRAY(ary1)->ptr[i])) {
+	VALUE v = RARRAY(ary1)->ptr[i];
+	if (st_delete(RHASH(hash)->tbl, &v, 0)) {
 	    rb_ary_push(ary3, RARRAY(ary1)->ptr[i]);
 	}
     }
+
     return ary3;
 }
 
@@ -1443,23 +1503,27 @@ static VALUE
 rb_ary_or(ary1, ary2)
     VALUE ary1, ary2;
 {
-    VALUE ary3;
+    VALUE hash;
+    VALUE ary3 = rb_ary_new();
+    VALUE v;
     long i;
 
-    if (TYPE(ary2) != T_ARRAY) {
-	if (rb_ary_includes(ary1, ary2)) return ary1;
-	else return rb_ary_push(ary1, ary2);
-    }
+    ary2 = to_ary(ary2);
+    hash = ary_make_hash(ary1, ary2);
 
-    ary3 = rb_ary_new();
     for (i=0; i<RARRAY(ary1)->len; i++) {
-	if (!rb_ary_includes(ary3, RARRAY(ary1)->ptr[i]))
-		rb_ary_push(ary3, RARRAY(ary1)->ptr[i]);
+	v = RARRAY(ary1)->ptr[i];
+	if (st_delete(RHASH(hash)->tbl, &v, 0)) {
+	    rb_ary_push(ary3, RARRAY(ary1)->ptr[i]);
+	}
     }
     for (i=0; i<RARRAY(ary2)->len; i++) {
-	if (!rb_ary_includes(ary3, RARRAY(ary2)->ptr[i]))
-		rb_ary_push(ary3, RARRAY(ary2)->ptr[i]);
+	v = RARRAY(ary2)->ptr[i];
+	if (st_delete(RHASH(hash)->tbl, &v, 0)) {
+	    rb_ary_push(ary3, RARRAY(ary2)->ptr[i]);
+	}
     }
+
     return ary3;
 }
 
@@ -1467,27 +1531,23 @@ static VALUE
 rb_ary_uniq_bang(ary)
     VALUE ary;
 {
-    VALUE *p, *q, *t, *end;
-    VALUE v;
+    VALUE hash = ary_make_hash(ary, 0);
+    VALUE *p, *q, *end;
 
-    rb_ary_modify(ary);
-    p = RARRAY(ary)->ptr;
-    end = p + RARRAY(ary)->len;
-
-    while (p < end) {
-	v = *p++;
-	q = t = p;
-	while (q < end) {
-	    if (rb_equal(*q, v)) q++;
-	    else *t++ = *q++;
-	}
-	end = t;
-    }
-    if (RARRAY(ary)->len == (end - RARRAY(ary)->ptr)) {
+    if (RARRAY(ary)->len == RHASH(hash)->tbl->num_entries) {
 	return Qnil;
     }
 
-    RARRAY(ary)->len = (end - RARRAY(ary)->ptr);
+    rb_ary_modify(ary);
+    p = q = RARRAY(ary)->ptr;
+    end = p + RARRAY(ary)->len;
+    while (p < end) {
+	VALUE v = *p++;
+	if (st_delete(RHASH(hash)->tbl, &v, 0)) {
+	    *q++ = *(p-1);
+	}
+    }
+    RARRAY(ary)->len = (q - RARRAY(ary)->ptr);
 
     return ary;
 }
@@ -1548,35 +1608,54 @@ rb_ary_nitems(ary)
     return INT2NUM(n);
 }
 
+static int
+flatten(ary, idx, ary2, memo)
+    VALUE ary;
+    long idx;
+    VALUE ary2, memo;
+{
+    VALUE id;
+    long i = idx;
+    long n, lim = idx + RARRAY(ary2)->len;
+
+    id = rb_obj_id(ary2);
+    if (rb_ary_includes(memo, id)) {
+	rb_raise(rb_eArgError, "tried to flatten recursive array");
+    }
+    rb_ary_push(memo, id);
+    rb_ary_replace(ary, idx, 1, ary2);
+    while (i < lim) {
+	if (TYPE(RARRAY(ary)->ptr[i]) == T_ARRAY) {
+	    n = flatten(ary, i, RARRAY(ary)->ptr[i], memo);
+	    i += n; lim += n;
+	}
+	i++;
+    }
+    rb_ary_pop(memo);
+
+    return lim - idx - 1;	/* returns number of increased items */
+}
+
 static VALUE
 rb_ary_flatten_bang(ary)
     VALUE ary;
 {
-    long i;
+    long i = 0;
     int mod = 0;
-    VALUE flattening = Qnil;
+    VALUE memo = Qnil;
 
     rb_ary_modify(ary);
-    for (i=0; i<RARRAY(ary)->len; i++) {
+    while (i<RARRAY(ary)->len) {
 	VALUE ary2 = RARRAY(ary)->ptr[i];
-	if (TYPE(ary2) == T_ARRAY) {
-	    if (ary == ary2) {
-		ary2 = Qnil;
-	    } else {
-		VALUE id;
 
-		if (NIL_P(flattening)) {
-		    flattening = rb_ary_new();
-		}
-		id = rb_obj_id(ary2);
-		if (rb_ary_includes(flattening, id)) {
-		    rb_raise(rb_eArgError, "tryed to flatten recursive array");
-		}
-		rb_ary_push(flattening, id);
+	if (TYPE(ary2) == T_ARRAY) {
+	    if (NIL_P(memo)) {
+		memo = rb_ary_new();
 	    }
-	    rb_ary_replace(ary, i--, 1, ary2);
+	    i += flatten(ary, i, ary2, memo);
 	    mod = 1;
 	}
+	i++;
     }
     if (mod == 0) return Qnil;
     return ary;
