@@ -3,7 +3,7 @@
   bignum.c -
 
   $Author: matz $
-  $Date: 2000/09/15 06:00:22 $
+  $Date: 2002/10/07 07:44:18 $
   created at: Fri Jun 10 00:48:55 JST 1994
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -20,15 +20,27 @@ VALUE rb_cBignum;
 #define USHORT _USHORT
 #endif
 
-typedef unsigned short USHORT;
+#if SIZEOF_INT*2 <= SIZEOF_LONG_LONG
+typedef unsigned int BDIGIT;
+typedef unsigned long long BDIGIT_DBL;
+typedef long long BDIGIT_DBL_SIGNED;
+#elif SIZEOF_INT*2 <= SIZEOF___INT64
+typedef unsigned int BDIGIT;
+typedef unsigned __int64 BDIGIT_DBL;
+typedef __int64 BDIGIT_DBL_SIGNED;
+#else
+typedef unsigned short BDIGIT;
+typedef unsigned long BDIGIT_DBL;
+typedef long BDIGIT_DBL_SIGNED;
+#endif
 
-#define BDIGITS(x) ((USHORT*)RBIGNUM(x)->digits)
-#define BITSPERDIG (sizeof(short)*CHAR_BIT)
-#define BIGRAD (1L << BITSPERDIG)
-#define DIGSPERLONG ((unsigned int)(sizeof(long)/sizeof(short)))
-#define BIGUP(x) ((unsigned long)(x) << BITSPERDIG)
+#define BDIGITS(x) ((BDIGIT*)RBIGNUM(x)->digits)
+#define BITSPERDIG (sizeof(BDIGIT)*CHAR_BIT)
+#define BIGRAD ((BDIGIT_DBL)1 << BITSPERDIG)
+#define DIGSPERLONG ((unsigned int)(sizeof(long)/sizeof(BDIGIT)))
+#define BIGUP(x) ((BDIGIT_DBL)(x) << BITSPERDIG)
 #define BIGDN(x) RSHIFT(x,BITSPERDIG)
-#define BIGLO(x) ((USHORT)((x) & (BIGRAD-1)))
+#define BIGLO(x) ((BDIGIT)((x) & (BIGRAD-1)))
 
 static VALUE
 bignew_1(klass, len, sign)
@@ -40,7 +52,7 @@ bignew_1(klass, len, sign)
     OBJSETUP(big, klass, T_BIGNUM);
     big->sign = sign;
     big->len = len;
-    big->digits = ALLOC_N(USHORT, len);
+    big->digits = ALLOC_N(BDIGIT, len);
 
     return (VALUE)big;
 }
@@ -53,17 +65,18 @@ rb_big_clone(x)
 {
     VALUE z = bignew_1(CLASS_OF(x), RBIGNUM(x)->len, RBIGNUM(x)->sign);
 
-    MEMCPY(BDIGITS(z), BDIGITS(x), USHORT, RBIGNUM(x)->len);
+    MEMCPY(BDIGITS(z), BDIGITS(x), BDIGIT, RBIGNUM(x)->len);
     return z;
 }
 
-void
-rb_big_2comp(x)			/* get 2's complement */
+static void
+get2comp(x, carry)		/* get 2's complement */
     VALUE x;
+    int carry;
 {
     long i = RBIGNUM(x)->len;
-    USHORT *ds = BDIGITS(x);
-    long num;
+    BDIGIT *ds = BDIGITS(x);
+    BDIGIT_DBL num;
 
     while (i--) ds[i] = ~ds[i];
     i = 0; num = 1;
@@ -72,28 +85,37 @@ rb_big_2comp(x)			/* get 2's complement */
 	ds[i++] = BIGLO(num);
 	num = BIGDN(num);
     } while (i < RBIGNUM(x)->len);
+    if (!carry) return;
     if (ds[0] == 1 || ds[0] == 0) {
+	if (RBIGNUM(x)->len == 1) return;
 	for (i=1; i<RBIGNUM(x)->len; i++) {
 	    if (ds[i] != 0) return;
 	}
-	REALLOC_N(RBIGNUM(x)->digits, USHORT, RBIGNUM(x)->len++);
+	REALLOC_N(RBIGNUM(x)->digits, BDIGIT, ++RBIGNUM(x)->len);
 	ds = BDIGITS(x);
 	ds[RBIGNUM(x)->len-1] = 1;
     }
 }
 
+void
+rb_big_2comp(x)			/* get 2's complement */
+    VALUE x;
+{
+    get2comp(x, Qtrue);
+}
+
 static VALUE
 bignorm(x)
     VALUE x;
 {
     if (!FIXNUM_P(x)) {
 	long len = RBIGNUM(x)->len;
-	USHORT *ds = BDIGITS(x);
+	BDIGIT *ds = BDIGITS(x);
 
 	while (len-- && !ds[len]) ;
 	RBIGNUM(x)->len = ++len;
 
-	if (len*sizeof(USHORT) <= sizeof(VALUE)) {
+	if (len*sizeof(BDIGIT) <= sizeof(VALUE)) {
 	    long num = 0;
 	    while (len--) {
 		num = BIGUP(num) + ds[len];
@@ -120,20 +142,21 @@ VALUE
 rb_uint2big(n)
     unsigned long n;
 {
-    unsigned int i = 0;
-    USHORT *digits;
+    BDIGIT_DBL num = n;
+    long i = 0;
+    BDIGIT *digits;
     VALUE big;
 
     i = 0;
     big = bignew(DIGSPERLONG, 1);
     digits = BDIGITS(big);
     while (i < DIGSPERLONG) {
-	digits[i++] = BIGLO(n);
-	n = BIGDN(n);
+	digits[i++] = BIGLO(num);
+	num = BIGDN(num);
     }
 
     i = DIGSPERLONG;
-    while (i-- && !digits[i]) ;
+    while (--i && !digits[i]) ;
     RBIGNUM(big)->len = i+1;
     return big;
 }
@@ -181,68 +204,69 @@ rb_cstr2inum(str, base)
     char *end;
     int badcheck = (base==0)?1:0;
     char sign = 1, c;
-    unsigned long num;
+    BDIGIT_DBL num;
     long len, blen = 1;
     long i;
     VALUE z;
-    USHORT *zds;
+    BDIGIT *zds;
 
     while (*str && ISSPACE(*str)) str++;
 
-    if (*str == '+') {
+    if (str[0] == '+') {
 	str++;
     }
-    else if (*str == '-') {
+    else if (str[0] == '-') {
 	str++;
 	sign = 0;
     }
+    if (str[0] == '+' || str[0] == '-') {
+	if (badcheck) goto bad;
+	return INT2FIX(0);
+    }
     if (base == 0) {
-	if (*str == '0') {
-	    str++;
-	    if (*str == 'x' || *str == 'X') {
-		str++;
+	if (str[0] == '0') {
+	    if (str[1] == 'x' || str[1] == 'X') {
 		base = 16;
 	    }
-	    else if (*str == 'b' || *str == 'B') {
-		str++;
+	    else if (str[1] == 'b' || str[1] == 'B') {
 		base = 2;
 	    }
 	    else {
 		base = 8;
-		if (!*str) return INT2FIX(0);
 	    }
 	}
 	else {
 	    base = 10;
-	    if (!*str) return INT2FIX(0);
 	}
     }
     if (base == 8) {
-	while (str[0] == '0') str++;
-	len = 3*strlen(str)*sizeof(char);
+	len = 3;
     }
     else {			/* base == 10, 2 or 16 */
 	if (base == 16 && str[0] == '0' && (str[1] == 'x'||str[1] == 'X')) {
 	    str += 2;
 	}
-	if (base == 2 && str[0] == '0' && (str[1] == 'b'||str[1] == 'B')) {
+	else if (base == 2 && str[0] == '0' && (str[1] == 'b'||str[1] == 'B')) {
 	    str += 2;
 	}
-	while (*str && *str == '0') str++;
-	if (!*str) str--;
-	len = 4*strlen(str)*sizeof(char);
+	len = 4;
     }
+    if (*str == '0') {		/* squeeze preceeding 0s */
+	while (*++str == '0');
+	--str;
+    }
+    len *= strlen(str)*sizeof(char);
 
     if (len <= (sizeof(VALUE)*CHAR_BIT)) {
 	unsigned long val = strtoul((char*)str, &end, base);
 
+	if (*end == '_') goto bigparse;
 	if (badcheck) {
-	    if (end == str || *end)
-		goto bad;
+	    if (end == str) goto bad; /* no number */
 	    while (*end && ISSPACE(*end)) end++;
-	    if (*end) {
+	    if (*end) {		      /* trailing garbage */
 	      bad:
-		rb_raise(rb_eArgError, "invalid literal for Integer: %s", s);
+		rb_raise(rb_eArgError, "invalid value for Integer: \"%s\"", s);
 	    }
 	}
 
@@ -259,36 +283,38 @@ rb_cstr2inum(str, base)
 	    return big;
 	}
     }
+  bigparse:
     len = (len/BITSPERDIG)+1;
+    if (badcheck && *str == '_') goto bad;
 
     z = bignew(len, sign);
     zds = BDIGITS(z);
     for (i=len;i--;) zds[i]=0;
     while (c = *str++) {
 	switch (c) {
+	  case '8': case '9':
+	    if (base == 8) {
+		c = base;
+		break;
+	    }
 	  case '0': case '1': case '2': case '3': case '4':
-	  case '5': case '6': case '7': case '8': case '9':
+	  case '5': case '6': case '7': 
 	    c = c - '0';
 	    break;
 	  case 'a': case 'b': case 'c':
 	  case 'd': case 'e': case 'f':
-	    c = c - 'a' + 10;
+	    if (base != 16) c = base;
+	    else c = c - 'a' + 10;
 	    break;
 	  case 'A': case 'B': case 'C':
 	  case 'D': case 'E': case 'F':
-	    c = c - 'A' + 10;
+	    if (base != 16) c = base;
+	    else c = c - 'A' + 10;
 	    break;
+	  case '_':
+	    continue;
 	  default:
 	    c = base;
-	    if (badcheck) {
-		if (ISSPACE(c)) {
-		    while (*str && ISSPACE(*str)) str++;
-		    if (*str) {
-			break;
-		    }
-		}
-		rb_raise(rb_eArgError, "invalid literal for Integer: %s", s);
-	    }
 	    break;
 	}
 	if (c >= base) break;
@@ -296,7 +322,7 @@ rb_cstr2inum(str, base)
 	num = c;
 	for (;;) {
 	    while (i<blen) {
-		num += zds[i]*base;
+		num += (BDIGIT_DBL)zds[i]*base;
 		zds[i++] = BIGLO(num);
 		num = BIGDN(num);
 	    }
@@ -307,6 +333,12 @@ rb_cstr2inum(str, base)
 	    break;
 	}
     }
+    if (badcheck) {
+	str--;
+	if (s+1 < str && str[-1] == '_') goto bad;
+	while (*str && ISSPACE(*str)) str++;
+	if (*str) goto bad;
+    }
 
     return bignorm(z);
 }
@@ -320,6 +352,13 @@ rb_str2inum(str, base)
     int len;
 
     s = rb_str2cstr(str, &len);
+    if (s[len]) {		/* no sentinel somehow */
+	char *p = ALLOCA_N(char, len+1);
+
+	MEMCPY(p, s, char, len);
+	p[len] = '\0';
+	s = p;
+    }
     if (len != strlen(s)) {
 	rb_raise(rb_eArgError, "string for Integer contains null byte");
     }
@@ -332,9 +371,9 @@ rb_big2str(x, base)
     VALUE x;
     int base;
 {
-    VALUE t;
-    USHORT *ds;
-    unsigned long i, j, hbase;
+    volatile VALUE t;
+    BDIGIT *ds;
+    long i, j, hbase;
     VALUE ss;
     char *s, c;
 
@@ -342,21 +381,23 @@ rb_big2str(x, base)
 	return rb_fix2str(x, base);
     }
     i = RBIGNUM(x)->len;
-    if (i == 0) return rb_str_new2("0");
+    if (i == 0 || (i == 1 && BDIGITS(x)[0] == 0)) {
+	return rb_str_new2("0");
+    }
     if (base == 10) {
-	j = (sizeof(USHORT)/sizeof(char)*CHAR_BIT*i*241L)/800+2;
+	j = (sizeof(BDIGIT)/sizeof(char)*CHAR_BIT*i*241L)/800+2;
 	hbase = 10000;
     }
     else if (base == 16) {
-	j = (sizeof(USHORT)/sizeof(char)*CHAR_BIT*i)/4+2;
+	j = (sizeof(BDIGIT)/sizeof(char)*CHAR_BIT*i)/4+2;
 	hbase = 0x10000;
     }
     else if (base == 8) {
-	j = (sizeof(USHORT)/sizeof(char)*CHAR_BIT*i)+2;
+	j = (sizeof(BDIGIT)/sizeof(char)*CHAR_BIT*i)+2;
 	hbase = 010000;
     }
     else if (base == 2) {
-	j = (sizeof(USHORT)*CHAR_BIT*i)+2;
+	j = (sizeof(BDIGIT)*CHAR_BIT*i)+2;
 	hbase = 020;
     }
     else {
@@ -373,10 +414,11 @@ rb_big2str(x, base)
     s[0] = RBIGNUM(x)->sign ? '+' : '-';
     while (i && j) {
 	long k = i;
-	unsigned long num = 0;
+	BDIGIT_DBL num = 0;
+
 	while (k--) {
 	    num = BIGUP(num) + ds[k];
-	    ds[k] = (USHORT)(num / hbase);
+	    ds[k] = (BDIGIT)(num / hbase);
 	    num %= hbase;
 	}
 	if (ds[i-1] == 0) i--;
@@ -408,11 +450,11 @@ big2ulong(x, type)
     VALUE x;
     char *type;
 {
-    unsigned long num;
     long len = RBIGNUM(x)->len;
-    USHORT *ds;
+    BDIGIT_DBL num;
+    BDIGIT *ds;
 
-    if (len > sizeof(long)/sizeof(USHORT))
+    if (len > sizeof(long)/sizeof(BDIGIT))
 	rb_raise(rb_eRangeError, "bignum too big to convert into `%s'", type);
     ds = BDIGITS(x);
     num = 0;
@@ -439,7 +481,7 @@ rb_big2long(x)
 {
     unsigned long num = big2ulong(x, "int");
 
-    if ((long)num < 0) {
+    if ((long)num < 0 && (RBIGNUM(x)->sign || (long)num != LONG_MIN)) {
 	rb_raise(rb_eRangeError, "bignum too big to convert into `int'");
     }
     if (!RBIGNUM(x)->sign) return -(long)num;
@@ -450,9 +492,9 @@ static VALUE
 dbl2big(d)
     double d;
 {
-    unsigned long i = 0;
-    long c;
-    USHORT *digits;
+    long i = 0;
+    BDIGIT c;
+    BDIGIT *digits;
     VALUE z;
     double u = (d < 0)?-d:d;
 
@@ -471,9 +513,9 @@ dbl2big(d)
     digits = BDIGITS(z);
     while (i--) {
 	u *= BIGRAD;
-	c = (long)u;
+	c = (BDIGIT)u;
 	u -= c;
-	digits[i] = (USHORT)c;
+	digits[i] = c;
     }
 
     return z;
@@ -492,11 +534,12 @@ rb_big2dbl(x)
 {
     double d = 0.0;
     long i = RBIGNUM(x)->len;
-    USHORT *ds = BDIGITS(x);
+    BDIGIT *ds = BDIGITS(x);
 
     while (i--) {
 	d = ds[i] + BIGRAD*d;
     }
+    if (isinf(d)) d = HUGE_VAL;
     if (!RBIGNUM(x)->sign) d = -d;
     return d;
 }
@@ -555,14 +598,13 @@ rb_big_eq(x, y)
       case T_BIGNUM:
 	break;
       case T_FLOAT:
-	y = dbl2big(RFLOAT(y)->value);
-	break;
+	return (rb_big2dbl(x) == RFLOAT(y)->value)?Qtrue:Qfalse;
       default:
-	return Qfalse;
+	return rb_equal(y, x);
     }
     if (RBIGNUM(x)->sign != RBIGNUM(y)->sign) return Qfalse;
     if (RBIGNUM(x)->len != RBIGNUM(y)->len) return Qfalse;
-    if (MEMCMP(BDIGITS(x),BDIGITS(y),USHORT,RBIGNUM(y)->len) != 0) return Qfalse;
+    if (MEMCMP(BDIGITS(x),BDIGITS(y),BDIGIT,RBIGNUM(y)->len) != 0) return Qfalse;
     return Qtrue;
 }
 
@@ -583,11 +625,11 @@ rb_big_neg(x)
 {
     VALUE z = rb_big_clone(x);
     long i = RBIGNUM(x)->len;
-    USHORT *ds = BDIGITS(z);
+    BDIGIT *ds = BDIGITS(z);
 
-    if (!RBIGNUM(x)->sign) rb_big_2comp(z);
+    if (!RBIGNUM(x)->sign) get2comp(z, Qtrue);
     while (i--) ds[i] = ~ds[i];
-    if (RBIGNUM(x)->sign) rb_big_2comp(z);
+    if (RBIGNUM(x)->sign) get2comp(z, Qfalse);
     RBIGNUM(z)->sign = !RBIGNUM(z)->sign;
 
     return bignorm(z);
@@ -598,8 +640,8 @@ bigsub(x, y)
     VALUE x, y;
 {
     VALUE z = 0;
-    USHORT *zds;
-    long num;
+    BDIGIT *zds;
+    BDIGIT_DBL_SIGNED num;
     long i;
 
     i = RBIGNUM(x)->len;
@@ -624,7 +666,7 @@ bigsub(x, y)
     zds = BDIGITS(z);
 
     for (i = 0, num = 0; i < RBIGNUM(y)->len; i++) { 
-	num += (long)BDIGITS(x)[i] - BDIGITS(y)[i];
+	num += (BDIGIT_DBL_SIGNED)BDIGITS(x)[i] - BDIGITS(y)[i];
 	zds[i] = BIGLO(num);
 	num = BIGDN(num);
     } 
@@ -647,7 +689,7 @@ bigadd(x, y, sign)
     char sign;
 {
     VALUE z;
-    long num;
+    BDIGIT_DBL num;
     long i, len;
 
     sign = (sign == RBIGNUM(y)->sign);
@@ -667,7 +709,7 @@ bigadd(x, y, sign)
 
     len = RBIGNUM(x)->len;
     for (i = 0, num = 0; i < len; i++) {
-	num += BDIGITS(x)[i] + BDIGITS(y)[i];
+	num += (BDIGIT_DBL)BDIGITS(x)[i] + BDIGITS(y)[i];
 	BDIGITS(z)[i] = BIGLO(num);
 	num = BIGDN(num);
     }
@@ -681,7 +723,7 @@ bigadd(x, y, sign)
 	BDIGITS(z)[i] = BDIGITS(y)[i];
 	i++;
     }
-    BDIGITS(z)[i] = (USHORT)num;
+    BDIGITS(z)[i] = (BDIGIT)num;
 
     return z;
 }
@@ -729,9 +771,9 @@ rb_big_mul(x, y)
     VALUE x, y;
 {
     long i, j;
-    unsigned long n = 0;
+    BDIGIT_DBL n = 0;
     VALUE z;
-    USHORT *zds;
+    BDIGIT *zds;
 
     if (FIXNUM_P(x)) x = rb_int2big(FIX2LONG(x));
     switch (TYPE(y)) {
@@ -754,11 +796,11 @@ rb_big_mul(x, y)
     zds = BDIGITS(z);
     while (j--) zds[j] = 0;
     for (i = 0; i < RBIGNUM(x)->len; i++) {
-	unsigned long dd = BDIGITS(x)[i]; 
+	BDIGIT_DBL dd = BDIGITS(x)[i]; 
 	if (dd == 0) continue;
 	n = 0;
 	for (j = 0; j < RBIGNUM(y)->len; j++) {
-	    int ee = n + dd * BDIGITS(y)[j];
+	    BDIGIT_DBL ee = n + (BDIGIT_DBL)dd * BDIGITS(y)[j];
 	    n = zds[i + j] + ee;
 	    if (ee) zds[i + j] = BIGLO(n);
 	    n = BIGDN(n);
@@ -779,10 +821,10 @@ bigdivrem(x, y, divp, modp)
     long nx = RBIGNUM(x)->len, ny = RBIGNUM(y)->len;
     long i, j;
     VALUE yy, z;
-    USHORT *xds, *yds, *zds, *tds;
-    unsigned long t2;
-    long num;
-    USHORT dd, q;
+    BDIGIT *xds, *yds, *zds, *tds;
+    BDIGIT_DBL t2;
+    BDIGIT_DBL_SIGNED num;
+    BDIGIT dd, q;
 
     yds = BDIGITS(y);
     if (ny == 0 && yds[0] == 0) rb_num_zerodiv();
@@ -799,12 +841,14 @@ bigdivrem(x, y, divp, modp)
 	t2 = 0; i = nx;
 	while (i--) {
 	    t2 = BIGUP(t2) + zds[i];
-	    zds[i] = (USHORT)(t2 / dd);
+	    zds[i] = (BDIGIT)(t2 / dd);
 	    t2 %= dd;
 	}
 	RBIGNUM(z)->sign = RBIGNUM(x)->sign==RBIGNUM(y)->sign;
-	if (!RBIGNUM(x)->sign) t2 = -(long)t2;
-	if (modp) *modp = rb_int2big((long)t2);
+	if (modp) {
+	    *modp = rb_uint2big((unsigned long)t2);
+	    RBIGNUM(*modp)->sign = RBIGNUM(x)->sign;
+	}
 	if (divp) *divp = z;
 	return;
     }
@@ -812,52 +856,60 @@ bigdivrem(x, y, divp, modp)
     zds = BDIGITS(z);
     if (nx==ny) zds[nx+1] = 0;
     while (!yds[ny-1]) ny--;
-    if ((dd = BIGRAD/(int)(yds[ny-1]+1)) != 1) {
+
+    dd = 0;
+    q = yds[ny-1];
+    while ((q & (1<<(BITSPERDIG-1))) == 0) {
+	q <<= 1;
+	dd++;
+    }
+    if (dd) {
 	yy = rb_big_clone(y);
 	tds = BDIGITS(yy);
 	j = 0;
-	num = 0;
+	t2 = 0;
 	while (j<ny) {
-	    num += (long)yds[j]*dd;
-	    tds[j++] = BIGLO(num);
-	    num = BIGDN(num);
+	    t2 += (BDIGIT_DBL)yds[j]<<dd;
+	    tds[j++] = BIGLO(t2);
+	    t2 = BIGDN(t2);
 	}
 	yds = tds;
 	j = 0;
-	num = 0;
+	t2 = 0;
 	while (j<nx) {
-	    num += (long)xds[j]*dd;
-	    zds[j++] = BIGLO(num);
-	    num = BIGDN(num);
+	    t2 += (BDIGIT_DBL)xds[j]<<dd;
+	    zds[j++] = BIGLO(t2);
+	    t2 = BIGDN(t2);
 	}
-	zds[j] = (USHORT)num;
+	zds[j] = (BDIGIT)t2;
     }
     else {
 	zds[nx] = 0;
 	j = nx;
 	while (j--) zds[j] = xds[j];
     }
+
     j = nx==ny?nx+1:nx;
     do {
 	if (zds[j] ==  yds[ny-1]) q = BIGRAD-1;
-	else q = (USHORT)((BIGUP(zds[j]) + zds[j-1])/yds[ny-1]);
+	else q = (BDIGIT)((BIGUP(zds[j]) + zds[j-1])/yds[ny-1]);
 	if (q) {
 	    i = 0; num = 0; t2 = 0;
 	    do {			/* multiply and subtract */
-		int ee;
-		t2 += (long)yds[i] * q;
+		BDIGIT_DBL ee;
+		t2 += (BDIGIT_DBL)yds[i] * q;
 		ee = num - BIGLO(t2);
-		num = zds[j - ny + i] + ee;
+		num = (BDIGIT_DBL)zds[j - ny + i] + ee;
 		if (ee) zds[j - ny + i] = BIGLO(num);
 		num = BIGDN(num);
 		t2 = BIGDN(t2);
 	    } while (++i < ny);
-	    num += zds[j - ny + i] - t2; /* borrow from high digit; don't update */
+	    num += zds[j - ny + i] - t2;/* borrow from high digit; don't update */
 	    while (num) {		/* "add back" required */
 		i = 0; num = 0; q--;
 		do {
-		    int ee = num + yds[i];
-		    num = (long) zds[j - ny + i] + ee;
+		    BDIGIT_DBL ee = num + yds[i];
+		    num = (BDIGIT_DBL)zds[j - ny + i] + ee;
 		    if (ee) zds[j - ny + i] = BIGLO(num);
 		    num = BIGDN(num);
 		} while (++i < ny);
@@ -875,13 +927,15 @@ bigdivrem(x, y, divp, modp)
     }
     if (modp) {			/* just normalize remainder */
 	*modp = rb_big_clone(z);
+	zds = BDIGITS(*modp);
+	while (--ny && !zds[ny]); ++ny;
 	if (dd) {
-	    zds = BDIGITS(*modp);
 	    t2 = 0; i = ny;
 	    while(i--) {
-		t2 = BIGUP(t2) + zds[i];
-		zds[i] = (USHORT)(t2 / dd);
-		t2 %= dd;
+		t2 = (t2 | zds[i]) >> dd;
+		q = zds[i];
+		zds[i] = BIGLO(t2);
+		t2 = BIGUP(q);
 	    }
 	}
 	RBIGNUM(*modp)->len = ny;
@@ -897,7 +951,8 @@ bigdivmod(x, y, divp, modp)
     VALUE mod;
 
     bigdivrem(x, y, divp, &mod);
-    if (RBIGNUM(x)->sign != RBIGNUM(y)->sign && RBIGNUM(mod)->len > 0) {
+    if (RBIGNUM(x)->sign != RBIGNUM(y)->sign &&
+	!(RBIGNUM(mod)->len == 1 && BDIGITS(mod)[0] == 0)) {
 	if (divp) *divp = bigadd(*divp, rb_int2big(1), 0);
 	if (modp) *modp = bigadd(mod, y, 1);
     }
@@ -1049,7 +1104,7 @@ rb_big_and(x, y)
     VALUE x, y;
 {
     VALUE z;
-    USHORT *ds1, *ds2, *zds;
+    BDIGIT *ds1, *ds2, *zds;
     long i, l1, l2;
     char sign;
 
@@ -1062,11 +1117,11 @@ rb_big_and(x, y)
 
     if (!RBIGNUM(y)->sign) {
 	y = rb_big_clone(y);
-	rb_big_2comp(y);
+	get2comp(y, Qtrue);
     }
     if (!RBIGNUM(x)->sign) {
 	x = rb_big_clone(x);
-	rb_big_2comp(x);
+	get2comp(x, Qtrue);
     }
     if (RBIGNUM(x)->len > RBIGNUM(y)->len) {
 	l1 = RBIGNUM(y)->len;
@@ -1091,7 +1146,7 @@ rb_big_and(x, y)
     for (; i<l2; i++) {
 	zds[i] = sign?0:ds2[i];
     }
-    if (!RBIGNUM(z)->sign) rb_big_2comp(z);
+    if (!RBIGNUM(z)->sign) get2comp(z, Qfalse);
     return bignorm(z);
 }
 
@@ -1100,8 +1155,8 @@ rb_big_or(x, y)
     VALUE x, y;
 {
     VALUE z;
-    USHORT *ds1, *ds2, *zds;
-    unsigned long i, l1, l2;
+    BDIGIT *ds1, *ds2, *zds;
+    long i, l1, l2;
     char sign;
 
     if (FIXNUM_P(y)) {
@@ -1113,11 +1168,11 @@ rb_big_or(x, y)
 
     if (!RBIGNUM(y)->sign) {
 	y = rb_big_clone(y);
-	rb_big_2comp(y);
+	get2comp(y, Qtrue);
     }
     if (!RBIGNUM(x)->sign) {
 	x = rb_big_clone(x);
-	rb_big_2comp(x);
+	get2comp(x, Qtrue);
     }
     if (RBIGNUM(x)->len > RBIGNUM(y)->len) {
 	l1 = RBIGNUM(y)->len;
@@ -1142,7 +1197,7 @@ rb_big_or(x, y)
     for (; i<l2; i++) {
 	zds[i] = sign?ds2[i]:(BIGRAD-1);
     }
-    if (!RBIGNUM(z)->sign) rb_big_2comp(z);
+    if (!RBIGNUM(z)->sign) get2comp(z, Qfalse);
 
     return bignorm(z);
 }
@@ -1152,8 +1207,8 @@ rb_big_xor(x, y)
     VALUE x, y;
 {
     VALUE z;
-    USHORT *ds1, *ds2, *zds;
-    unsigned int i, l1, l2;
+    BDIGIT *ds1, *ds2, *zds;
+    long i, l1, l2;
     char sign;
 
     if (FIXNUM_P(y)) {
@@ -1165,11 +1220,11 @@ rb_big_xor(x, y)
 
     if (!RBIGNUM(y)->sign) {
 	y = rb_big_clone(y);
-	rb_big_2comp(y);
+	get2comp(y, Qtrue);
     }
     if (!RBIGNUM(x)->sign) {
 	x = rb_big_clone(x);
-	rb_big_2comp(x);
+	get2comp(x, Qtrue);
     }
     if (RBIGNUM(x)->len > RBIGNUM(y)->len) {
 	l1 = RBIGNUM(y)->len;
@@ -1196,7 +1251,7 @@ rb_big_xor(x, y)
     for (; i<l2; i++) {
 	zds[i] = sign?ds2[i]:~ds2[i];
     }
-    if (!RBIGNUM(z)->sign) rb_big_2comp(z);
+    if (!RBIGNUM(z)->sign) get2comp(z, Qfalse);
 
     return bignorm(z);
 }
@@ -1207,12 +1262,12 @@ VALUE
 rb_big_lshift(x, y)
     VALUE x, y;
 {
-    USHORT *xds, *zds;
+    BDIGIT *xds, *zds;
     int shift = NUM2INT(y);
     int s1 = shift/BITSPERDIG;
     int s2 = shift%BITSPERDIG;
     VALUE z;
-    unsigned long num = 0;
+    BDIGIT_DBL num = 0;
     long len, i;
 
     if (shift < 0) return rb_big_rshift(x, INT2FIX(-shift));
@@ -1224,7 +1279,7 @@ rb_big_lshift(x, y)
     }
     xds = BDIGITS(x);
     for (i=0; i<len; i++) {
-	num = num | *xds++<<s2;
+	num = num | (BDIGIT_DBL)*xds++<<s2;
 	*zds++ = BIGLO(num);
 	num = BIGDN(num);
     }
@@ -1236,12 +1291,12 @@ static VALUE
 rb_big_rshift(x, y)
     VALUE x, y;
 {
-    USHORT *xds, *zds;
+    BDIGIT *xds, *zds;
     int shift = NUM2INT(y);
     int s1 = shift/BITSPERDIG;
     int s2 = shift%BITSPERDIG;
     VALUE z;
-    unsigned long num = 0;
+    BDIGIT_DBL num = 0;
     long i = RBIGNUM(x)->len;
     long j;
 
@@ -1252,15 +1307,25 @@ rb_big_rshift(x, y)
 	else
 	    return INT2FIX(-1);
     }
+    if (!RBIGNUM(x)->sign) {
+	x = rb_big_clone(x);
+	get2comp(x, Qtrue);
+    }
     xds = BDIGITS(x);
     i = RBIGNUM(x)->len; j = i - s1;
     z = bignew(j, RBIGNUM(x)->sign);
+    if (!RBIGNUM(x)->sign) {
+	num = ((BDIGIT_DBL)~0) << BITSPERDIG;
+    }
     zds = BDIGITS(z);
     while (i--, j--) {
 	num = (num | xds[i]) >> s2;
 	zds[j] = BIGLO(num);
 	num = BIGUP(xds[i]);
     }
+    if (!RBIGNUM(x)->sign) {
+	get2comp(z, Qfalse);
+    }
     return bignorm(z);
 }
 
@@ -1268,10 +1333,16 @@ static VALUE
 rb_big_aref(x, y)
     VALUE x, y;
 {
-    USHORT *xds;
-    int shift = NUM2INT(y);
+    BDIGIT *xds;
+    int shift;
     int s1, s2;
 
+    if (TYPE(y) == T_BIGNUM) {
+	if (!RBIGNUM(y)->sign || RBIGNUM(x)->sign)
+	    return INT2FIX(0);
+	return INT2FIX(1);
+    }
+    shift = NUM2INT(y);
     if (shift < 0) return INT2FIX(0);
     s1 = shift/BITSPERDIG;
     s2 = shift%BITSPERDIG;
@@ -1279,7 +1350,7 @@ rb_big_aref(x, y)
     if (!RBIGNUM(x)->sign) {
 	if (s1 >= RBIGNUM(x)->len) return INT2FIX(1);
 	x = rb_big_clone(x);
-	rb_big_2comp(x);
+	get2comp(x, Qtrue);
     }
     else {
 	if (s1 >= RBIGNUM(x)->len) return INT2FIX(0);
@@ -1294,12 +1365,11 @@ static VALUE
 rb_big_hash(x)
     VALUE x;
 {
-    long i, len;
-    int key;
-    USHORT *digits;
+    long i, len, key;
+    BDIGIT *digits;
 
-    key = 0; digits = BDIGITS(x);
-    for (i=0,len=RBIGNUM(x)->len; i<RBIGNUM(x)->len; i++) {
+    key = 0; digits = BDIGITS(x); len = RBIGNUM(x)->len;
+    for (i=0; i<len; i++) {
 	key ^= *digits++;
     }
     return INT2FIX(key);
@@ -1346,7 +1416,7 @@ rb_big_rand(max, rand)
     len = RBIGNUM(max)->len;
     v = bignew(len,1);
     while (len--) {
-	BDIGITS(v)[len] = ((USHORT)~0) * rand;
+	BDIGITS(v)[len] = ((BDIGIT)~0) * rand;
     }
 
     return rb_big_modulo((VALUE)v, max);
@@ -1356,7 +1426,7 @@ static VALUE
 rb_big_size(big)
     VALUE big;
 {
-    return INT2FIX(RBIGNUM(big)->len*sizeof(USHORT));
+    return INT2FIX(RBIGNUM(big)->len*sizeof(BDIGIT));
 }
 
 static VALUE
@@ -1371,8 +1441,6 @@ Init_Bignum()
 {
     rb_cBignum = rb_define_class("Bignum", rb_cInteger);
 
-    rb_undef_method(CLASS_OF(rb_cBignum), "new");
-
     rb_define_method(rb_cBignum, "to_s", rb_big_to_s, 0);
     rb_define_method(rb_cBignum, "coerce", rb_big_coerce, 1);
     rb_define_method(rb_cBignum, "-@", rb_big_uminus, 0);
@@ -1381,6 +1449,7 @@ Init_Bignum()
     rb_define_method(rb_cBignum, "*", rb_big_mul, 1);
     rb_define_method(rb_cBignum, "/", rb_big_div, 1);
     rb_define_method(rb_cBignum, "%", rb_big_modulo, 1);
+    rb_define_method(rb_cBignum, "div", rb_big_div, 1);
     rb_define_method(rb_cBignum, "divmod", rb_big_divmod, 1);
     rb_define_method(rb_cBignum, "modulo", rb_big_modulo, 1);
     rb_define_method(rb_cBignum, "remainder", rb_big_remainder, 1);
