@@ -1,14 +1,14 @@
-/************************************************
+/**********************************************************************
 
   compar.c -
 
   $Author: matz $
-  $Date: 2000/01/17 08:23:53 $
+  $Date: 2002/12/02 18:19:01 $
   created at: Thu Aug 26 14:39:48 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
 
-************************************************/
+**********************************************************************/
 
 #include "ruby.h"
 
@@ -21,9 +21,9 @@ cmp_eq(a)
     VALUE *a;
 {
     VALUE c = rb_funcall(a[0], cmp, 1, a[1]);
-    int t = NUM2INT(c);
 
-    if (t == 0) return Qtrue;
+    if (NIL_P(c)) return Qfalse;
+    if (rb_cmpint(c) == 0) return Qtrue;
     return Qfalse;
 }
 
@@ -42,7 +42,8 @@ cmp_equal(x, y)
     if (x == y) return Qtrue;
 
     a[0] = x; a[1] = y;
-    return rb_rescue(cmp_eq, (VALUE)a, cmp_failed, 0);
+    return rb_rescue2(cmp_eq, (VALUE)a, cmp_failed, 0,
+		      rb_eStandardError, rb_eNameError, 0);
 }
 
 static VALUE
@@ -50,9 +51,9 @@ cmp_gt(x, y)
     VALUE x, y;
 {
     VALUE c = rb_funcall(x, cmp, 1, y);
-    int t = NUM2INT(c);
 
-    if (t > 0) return Qtrue;
+    if (NIL_P(c)) return Qfalse;
+    if (rb_cmpint(c) > 0) return Qtrue;
     return Qfalse;
 }
 
@@ -61,9 +62,9 @@ cmp_ge(x, y)
     VALUE x, y;
 {
     VALUE c = rb_funcall(x, cmp, 1, y);
-    int t = NUM2INT(c);
 
-    if (t >= 0) return Qtrue;
+    if (NIL_P(c)) return Qfalse;
+    if (rb_cmpint(c) >= 0) return Qtrue;
     return Qfalse;
 }
 
@@ -72,9 +73,9 @@ cmp_lt(x, y)
     VALUE x, y;
 {
     VALUE c = rb_funcall(x, cmp, 1, y);
-    int t = NUM2INT(c);
 
-    if (t < 0) return Qtrue;
+    if (NIL_P(c)) return Qfalse;
+    if (rb_cmpint(c) < 0) return Qtrue;
     return Qfalse;
 }
 
@@ -83,9 +84,9 @@ cmp_le(x, y)
     VALUE x, y;
 {
     VALUE c = rb_funcall(x, cmp, 1, y);
-    int t = NUM2INT(c);
 
-    if (t <= 0) return Qtrue;
+    if (NIL_P(c)) return Qfalse;
+    if (rb_cmpint(c) <= 0) return Qtrue;
     return Qfalse;
 }
 
@@ -93,13 +94,8 @@ static VALUE
 cmp_between(x, min, max)
     VALUE x, min, max;
 {
-    VALUE c = rb_funcall(x, cmp, 1, min);
-    long t = NUM2LONG(c);
-    if (t < 0) return Qfalse;
-
-    c = rb_funcall(x, cmp, 1, max);
-    t = NUM2LONG(c);
-    if (t > 0) return Qfalse;
+    if (cmp_lt(x, min)) return Qfalse;
+    if (cmp_gt(x, max)) return Qfalse;
     return Qtrue;
 }
 
