@@ -2,8 +2,8 @@
 
   dir.c -
 
-  $Author: eban $
-  $Date: 2000/09/25 07:00:53 $
+  $Author: matz $
+  $Date: 2002/12/16 07:36:14 $
   created at: Wed Jan  5 09:51:01 JST 1994
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -26,10 +26,10 @@
 #include <unistd.h>
 #endif
 
-#if HAVE_DIRENT_H
+#if defined HAVE_DIRENT_H && !defined NT
 # include <dirent.h>
 # define NAMLEN(dirent) strlen((dirent)->d_name)
-#elif HAVE_DIRECT_H
+#elif defined HAVE_DIRECT_H && !defined NT
 # include <direct.h>
 # define NAMLEN(dirent) strlen((dirent)->d_name)
 #else
@@ -44,8 +44,8 @@
 # if HAVE_NDIR_H
 #  include <ndir.h>
 # endif
-# if defined(NT) && defined(_MSC_VER)
-#  include "missing/dir.h"
+# if defined(NT)
+#  include "win32/dir.h"
 # endif
 #endif
 
@@ -61,6 +61,10 @@ char *strchr _((char*,char));
 
 #include <ctype.h>
 
+#ifndef HAVE_LSTAT
+#define lstat rb_sys_stat
+#endif
+
 #define FNM_NOESCAPE	0x01
 #define FNM_PATHNAME	0x02
 #define FNM_PERIOD	0x04
@@ -69,7 +73,7 @@ char *strchr _((char*,char));
 #define FNM_NOMATCH	1
 #define FNM_ERROR	2
 
-#define downcase(c) (nocase && isupper(c) ? tolower(c) : (c))
+#define downcase(c) (nocase && ISUPPER(c) ? tolower(c) : (c))
 
 #if defined DOSISH
 #define isdirsep(c) ((c) == '/' || (c) == '\\')
@@ -126,17 +130,18 @@ range(pat, test, flags)
     return 0;
 }
 
+#define ISDIRSEP(c) (pathname && isdirsep(c))
 #define PERIOD(s) (period && *(s) == '.' && \
-		  ((s) == string || pathname && isdirsep(*(s))))
+		  ((s) == string || ISDIRSEP((s)[-1])))
 static int
 fnmatch(pat, string, flags)
-    char *pat;
-    char *string;
+    const char *pat;
+    const char *string;
     int flags;
 {
     int c;
     int test;
-    char *s = string;
+    const char *s = string;
     int escape = !(flags & FNM_NOESCAPE);
     int pathname = flags & FNM_PATHNAME;
     int period = flags & FNM_PERIOD;
@@ -145,7 +150,7 @@ fnmatch(pat, string, flags)
     while (c = *pat++) {
 	switch (c) {
 	case '?':
-	    if (!*s || pathname && isdirsep(*s) || PERIOD(s))
+	    if (!*s || ISDIRSEP(*s) || PERIOD(s))
 		return FNM_NOMATCH;
 	    s++;
 	    break;
@@ -162,10 +167,12 @@ fnmatch(pat, string, flags)
 		else
 		    return 0;
 	    }
-	    else if (pathname && isdirsep(c)) {
+	    else if (ISDIRSEP(c)) {
 		s = find_dirsep(s);
-		if (s)
+		if (s) {
+                    s++;
 		    break;
+                }
 		return FNM_NOMATCH;
 	    }
 
@@ -176,14 +183,14 @@ fnmatch(pat, string, flags)
 		if ((c == '[' || downcase(*s) == test) &&
 		    !fnmatch(pat, s, flags & ~FNM_PERIOD))
 		    return 0;
-		else if (pathname && isdirsep(*s))
+		else if (ISDIRSEP(*s))
 		    break;
 		s++;
 	    }
 	    return FNM_NOMATCH;
       
 	case '[':
-	    if (!*s || pathname && isdirsep(*s) || PERIOD(s))
+	    if (!*s || ISDIRSEP(*s) || PERIOD(s))
 		return FNM_NOMATCH;
 	    pat = range(pat, *s, flags);
 	    if (!pat)
@@ -207,7 +214,7 @@ fnmatch(pat, string, flags)
 
 	default:
 #if defined DOSISH
-	    if (pathname && isdirsep(c) && isdirsep(*s))
+	    if (ISDIRSEP(c) && isdirsep(*s))
 		;
 	    else
 #endif
@@ -361,6 +368,14 @@ dir_seek(dir, pos)
 #endif
 }
 
+static VALUE
+dir_set_pos(dir, pos)
+    VALUE dir, pos;
+{
+    dir_seek(dir, pos);
+    return pos;
+}
+
 static VALUE
 dir_rewind(dir)
     VALUE dir;
@@ -404,6 +419,7 @@ dir_s_chdir(argc, argv, obj)
 	dist = getenv("HOME");
 	if (!dist) {
 	    dist = getenv("LOGDIR");
+	    if (!dist) rb_raise(rb_eArgError, "HOME/LOGDIR not set");
 	}
     }
 
@@ -413,20 +429,29 @@ dir_s_chdir(argc, argv, obj)
     return INT2FIX(0);
 }
 
+static char *
+my_getcwd()
+{
+    int size = MAXPATHLEN;
+    char *buf = xmalloc(size);
+
+    while (!getcwd(buf, size)) {
+	if (errno != ERANGE) rb_sys_fail(NULL);
+	size *= 2;
+	buf = xrealloc(buf, size);
+    }
+    return buf;
+}
+
 static VALUE
 dir_s_getwd(dir)
     VALUE dir;
 {
-    char path[MAXPATHLEN];
+    char *path = my_getcwd();
+    VALUE cwd = rb_tainted_str_new2(path);
 
-#ifdef HAVE_GETCWD
-    if (getcwd(path, sizeof(path)) == 0) rb_sys_fail(path);
-#else
-    extern char *getwd();
-    if (getwd(path) == 0) rb_sys_fail(path);
-#endif
-
-    return rb_tainted_str_new2(path);
+    free(path);
+    return cwd;
 }
 
 static VALUE
@@ -490,12 +515,14 @@ dir_s_rmdir(obj, dir)
 
 /* Return nonzero if S has any special globbing chars in it.  */
 static int
-has_magic(s, send)
+has_magic(s, send, flags)
      char *s, *send;
+     int flags;
 {
     register char *p = s;
     register char c;
     int open = 0;
+    int escape = !(flags & FNM_NOESCAPE);
 
     while ((c = *p++) != '\0') {
 	switch (c) {
@@ -512,7 +539,7 @@ has_magic(s, send)
 	    continue;
 
 	  case '\\':
-	    if (*p++ == '\0')
+	    if (escape && *p++ == '\0')
 		return Qfalse;
 	}
 
@@ -531,7 +558,11 @@ extract_path(p, pend)
     len = pend - p;
     alloc = ALLOC_N(char, len+1);
     memcpy(alloc, p, len);
-    if (len > 1 && pend[-1] == '/') {
+    if (len > 1 && pend[-1] == '/'
+#if defined DOSISH
+    && pend[-2] != ':'
+#endif
+    ) {
 	alloc[len-1] = 0;
     }
     else {
@@ -558,26 +589,28 @@ extract_elem(path)
 #endif
 
 void
-rb_glob(path, func, arg)
+rb_glob_helper(path, sub, flags, func, arg)
     char *path;
-    void (*func)();
+    char *sub;
+    int flags;
+    void (*func) _((const char*, VALUE));
     VALUE arg;
 {
     struct stat st;
     char *p, *m;
 
-    if (!has_magic(path, 0)) {
-	if (stat(path, &st) == 0) {
+    p = sub ? sub : path;
+    if (!has_magic(p, 0, flags)) {
+	if (lstat(path, &st) == 0) {
 	    (*func)(path, arg);
 	}
 	return;
     }
 
-    p = path;
     while (p) {
 	if (*p == '/') p++;
 	m = strchr(p, '/');
-	if (has_magic(p, m)) {
+	if (has_magic(p, m, flags)) {
 	    char *dir, *base, *magic, *buf;
 	    DIR *dirp;
 	    struct dirent *dp;
@@ -593,31 +626,55 @@ rb_glob(path, func, arg)
 	    else dir = base;
 
 	    magic = extract_elem(p);
-	    if (m && strcmp(magic, "**") == 0) {
-		recursive = 1;
-		buf = ALLOC_N(char, strlen(base)+strlen(m)+3);
-		sprintf(buf, "%s%s%s", base, (*base)?"":".", m);
-		rb_glob(buf, func, arg);
-		free(buf);
+	    if (rb_sys_stat(dir, &st) < 0) {
+	        free(base);
+	        break;
 	    }
-	    dirp = opendir(dir);
-	    if (dirp == NULL) {
-		free(base);
-		break;
+	    if (S_ISDIR(st.st_mode)) {
+		if (m && strcmp(magic, "**") == 0) {
+		    int n = strlen(base);
+		    recursive = 1;
+		    buf = ALLOC_N(char, n+strlen(m)+3);
+		    sprintf(buf, "%s%s%s", base, (*base)?"":".", m);
+		    rb_glob_helper(buf, buf+n, flags, func, arg);
+		    free(buf);
+		}
+		dirp = opendir(dir);
+		if (dirp == NULL) {
+		    free(base);
+		    break;
+		}
 	    }
+	    else {
+	      free(base);
+	      break;
+	    }
+	    
+#if defined DOSISH
+#define BASE (*base && !((isdirsep(*base) && !base[1]) || (base[1] == ':' && isdirsep(base[2]) && !base[3])))
+#else
 #define BASE (*base && !(*base == '/' && !base[1]))
+#endif
 
 	    for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
 		if (recursive) {
 		    if (strcmp(".", dp->d_name) == 0 || strcmp("..", dp->d_name) == 0)
 			continue;
 		    buf = ALLOC_N(char, strlen(base)+NAMLEN(dp)+strlen(m)+6);
-		    sprintf(buf, "%s%s%s/**%s", base, (BASE)?"/":"", dp->d_name, m);
-		    rb_glob(buf, func, arg);
+		    sprintf(buf, "%s%s%s", base, (BASE)?"/":"", dp->d_name);
+		    if (lstat(buf, &st) < 0) {
+			continue;
+		    }
+		    if (S_ISDIR(st.st_mode)) {
+			char *t = buf+strlen(buf);
+		        strcpy(t, "/**");
+			strcpy(t+3, m);
+			rb_glob_helper(buf, t, flags, func, arg);
+		    }
 		    free(buf);
 		    continue;
 		}
-		if (fnmatch(magic, dp->d_name, FNM_PERIOD|FNM_PATHNAME) == 0) {
+		if (fnmatch(magic, dp->d_name, flags) == 0) {
 		    buf = ALLOC_N(char, strlen(base)+NAMLEN(dp)+2);
 		    sprintf(buf, "%s%s%s", base, (BASE)?"/":"", dp->d_name);
 		    if (!m) {
@@ -642,7 +699,7 @@ rb_glob(path, func, arg)
 		    char *t = ALLOC_N(char, len+mlen+1);
 
 		    sprintf(t, "%s%s", link->path, m);
-		    rb_glob(t, func, arg);
+		    rb_glob_helper(t, t+len, flags, func, arg);
 		    free(t);
 		}
 		tmp = link;
@@ -655,12 +712,39 @@ rb_glob(path, func, arg)
     }
 }
 
+void
+rb_glob(path, func, arg)
+    char *path;
+    void (*func) _((const char*, VALUE));
+    VALUE arg;
+{
+    rb_glob_helper(path, 0, FNM_PERIOD, func, arg);
+}
+
+void
+rb_iglob(path, func, arg)
+    char *path;
+    void (*func) _((const char*, VALUE));
+    VALUE arg;
+{
+    rb_glob_helper(path, 0, FNM_PERIOD|FNM_NOCASE, func, arg);
+}
+
+static void push_pattern _((const char *, VALUE));
+
 static void
 push_pattern(path, ary)
-    char *path;
+    const char *path;
     VALUE ary;
 {
-    rb_ary_push(ary, rb_tainted_str_new2(path));
+    VALUE str = rb_tainted_str_new2(path);
+
+    if (ary) {
+	rb_ary_push(ary, str);
+    }
+    else {
+	rb_yield(str);
+    }
 }
 
 static void
@@ -731,15 +815,19 @@ dir_s_glob(dir, str)
     VALUE dir, str;
 {
     char *p, *pend;
-    char buffer[MAXPATHLEN], *buf = buffer;
+    char buffer[MAXPATHLEN], *buf;
     char *t;
     int nest;
-    VALUE ary;
+    VALUE ary = 0;
 
     Check_SafeStr(str);
-    ary = rb_ary_new();
+    if (!rb_block_given_p()) {
+	ary = rb_ary_new();
+    }
     if (RSTRING(str)->len >= MAXPATHLEN)
 	buf = xmalloc(RSTRING(str)->len + 1);
+    else
+	buf = buffer;
 
     p = RSTRING(str)->ptr;
     pend = p + RSTRING(str)->len;
@@ -764,14 +852,7 @@ dir_s_glob(dir, str)
     }
     if (buf != buffer)
 	free(buf);
-    if (rb_block_given_p()) {
-	long len = RARRAY(ary)->len;
-	VALUE *ptr = RARRAY(ary)->ptr;
-
-	while (len--) {
-	    rb_yield(*ptr++);
-	}
-    }
+    if (!ary) return Qnil;
     return ary;
 }
 
@@ -815,7 +896,7 @@ Init_Dir()
     rb_define_method(rb_cDir,"tell", dir_tell, 0);
     rb_define_method(rb_cDir,"seek", dir_seek, 1);
     rb_define_method(rb_cDir,"pos", dir_tell, 0);
-    rb_define_method(rb_cDir,"pos=", dir_seek, 1);
+    rb_define_method(rb_cDir,"pos=", dir_set_pos, 1);
     rb_define_method(rb_cDir,"close", dir_close, 0);
 
     rb_define_singleton_method(rb_cDir,"chdir", dir_s_chdir, -1);
