@@ -1,7 +1,11 @@
 require 'mkmf'
 
+dir_config('curses')
+dir_config('ncurses')
+dir_config('termcap')
+
 make=false
-have_library("mytinfo", "tgetent") if /bow/ =~ PLATFORM
+have_library("mytinfo", "tgetent") if /bow/ =~ RUBY_PLATFORM
 if have_header("ncurses.h") and have_library("ncurses", "initscr")
   make=true
 elsif have_header("ncurses/curses.h") and have_library("ncurses", "initscr")
@@ -10,7 +14,7 @@ elsif have_header("curses_colr/curses.h") and have_library("cur_colr", "initscr"
   make=true
 else
   have_library("termcap", "tgetent") 
-  if have_library("curses", "initscr")
+  if have_header("curses.h") and have_library("curses", "initscr")
     make=true
   end
 end
