@@ -9,5 +9,8 @@ have_library("ncurses", "tgetnum") or
 if have_header("readline/readline.h") and
     have_header("readline/history.h") and
     have_library("readline", "readline")
+  if have_func("rl_filename_completion_function")
+    $CFLAGS += "-DREADLINE_42_OR_LATER"
+  end
   create_makefile("readline")
 end
