@@ -1,6 +1,7 @@
 /* readline.c -- GNU Readline module
    Copyright (C) 1997-1998  Shugo Maeda */
 
+#include <errno.h>
 #include <stdio.h>
 #include <readline/readline.h>
 #include <readline/history.h>
@@ -15,11 +16,18 @@ static VALUE mReadline;
 #define COMPLETION_PROC "completion_proc"
 #define COMPLETION_CASE_FOLD "completion_case_fold"
 
+#ifndef READLINE_42_OR_LATER
+# define rl_filename_completion_function filename_completion_function
+# define rl_username_completion_function username_completion_function
+# define rl_completion_matches completion_matches
+#endif
+
 static int
 readline_event()
 {
     CHECK_INTS;
     rb_thread_schedule();
+    return 0;
 }
 
 static VALUE
@@ -35,6 +43,9 @@ readline_readline(argc, argv, self)
     if (rb_scan_args(argc, argv, "02", &tmp, &add_hist) > 0) {
 	prompt = STR2CSTR(tmp);
     }
+
+    if (!isatty(0) && errno == EBADF) rb_raise(rb_eIOError, "stdin closed");
+
     buff = readline(prompt);
     if (RTEST(add_hist) && buff) {
 	add_history(buff);
@@ -160,6 +171,36 @@ readline_s_emacs_editing_mode(self)
     return Qnil;
 }
 
+static VALUE
+readline_s_set_completion_append_character(self, str)
+    VALUE self, str;
+{
+    if (NIL_P(str)) {
+	rl_completion_append_character = '\0';
+    } else {
+	Check_Type(str, T_STRING);
+
+	rl_completion_append_character = RSTRING(str)->ptr[0];
+    }
+
+    return self;
+}
+
+static VALUE
+readline_s_get_completion_append_character(self)
+    VALUE self;
+{
+    VALUE str;
+
+    if (rl_completion_append_character == '\0')
+	return Qnil;
+
+    str = rb_str_new("", 1);
+    RSTRING(str)->ptr[0] = rl_completion_append_character;
+
+    return str;
+}
+
 static VALUE
 hist_to_s(self)
     VALUE self;
@@ -321,8 +362,8 @@ filename_completion_proc_call(self, str)
     char **matches;
     int i;
 
-    matches = completion_matches(STR2CSTR(str),
-				 filename_completion_function);
+    matches = rl_completion_matches(STR2CSTR(str),
+				    rl_filename_completion_function);
     if (matches) {
 	result = rb_ary_new();
 	for (i = 0; matches[i]; i++) {
@@ -348,8 +389,8 @@ username_completion_proc_call(self, str)
     char **matches;
     int i;
 
-    matches = completion_matches(STR2CSTR(str),
-				 username_completion_function);
+    matches = rl_completion_matches(STR2CSTR(str),
+				    rl_username_completion_function);
     if (matches) {
 	result = rb_ary_new();
 	for (i = 0; matches[i]; i++) {
@@ -388,6 +429,10 @@ Init_readline()
 			       readline_s_vi_editing_mode, 0);
     rb_define_singleton_method(mReadline, "emacs_editing_mode",
 			       readline_s_emacs_editing_mode, 0);
+    rb_define_singleton_method(mReadline, "completion_append_character=",
+			       readline_s_set_completion_append_character, 1);
+    rb_define_singleton_method(mReadline, "completion_append_character",
+			       readline_s_get_completion_append_character, 0);
 
     histary = rb_obj_alloc(rb_cObject);
     rb_extend_object(histary, rb_mEnumerable);
