@@ -2,8 +2,8 @@
 
   socket.c -
 
-  $Author: matz $
-  $Date: 2000/08/24 06:21:42 $
+  $Author: knu $
+  $Date: 2002/10/26 15:57:26 $
   created at: Thu Mar 31 12:21:29 JST 1994
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -154,19 +154,9 @@ rb_getaddrinfo(nodename, servname, hints, res)
 #define getaddrinfo(node,serv,hints,res) rb_getaddrinfo((node),(serv),(hints),(res))
 #endif
 
-#ifdef NT
-static void
-sock_finalize(fptr)
-    OpenFile *fptr;
-{
-    SOCKET s;
-
-    if (!fptr->f) return;
-    s = get_osfhandle(fileno(fptr->f));
-    myfdclose(fptr->f);
-    if (fptr->f2) myfdclose(fptr->f2);
-    closesocket(s);
-}
+#ifdef HAVE_CLOSESOCKET
+#undef close
+#define close closesocket
 #endif
 
 static VALUE
@@ -180,11 +170,7 @@ sock_new(class, fd)
 
     MakeOpenFile(sock, fp);
     fp->f = rb_fdopen(fd, "r");
-#ifdef NT
-    fp->finalize = sock_finalize;
-#else
     fd = dup(fd);
-#endif
     fp->f2 = rb_fdopen(fd, "w");
     fp->mode = FMODE_READWRITE;
     rb_io_synchronized(fp);
@@ -237,11 +223,7 @@ bsock_close_read(sock)
     }
     rb_thread_fd_close(fileno(fptr->f));
     fptr->mode &= ~FMODE_READABLE;
-#ifdef NT
-    myfdclose(fptr->f);
-#else
     fclose(fptr->f);
-#endif
     fptr->f = fptr->f2;
     fptr->f2 = 0;
 
@@ -263,11 +245,7 @@ bsock_close_write(sock)
     }
     shutdown(fileno(fptr->f2), 1);
     fptr->mode &= ~FMODE_WRITABLE;
-#ifdef NT
-    myfdclose(fptr->f2);
-#else
     fclose(fptr->f2);
-#endif
     fptr->f2 = 0;
 
     return Qnil;
@@ -328,7 +306,7 @@ bsock_getsockopt(sock, lev, optname)
     if (getsockopt(fileno(fptr->f), level, option, buf, &len) < 0)
 	rb_sys_fail(fptr->path);
 
-    return rb_tainted_str_new(buf, len);
+    return rb_str_new(buf, len);
 #else
     rb_notimplement();
 #endif
@@ -345,7 +323,7 @@ bsock_getsockname(sock)
     GetOpenFile(sock, fptr);
     if (getsockname(fileno(fptr->f), (struct sockaddr*)buf, &len) < 0)
 	rb_sys_fail("getsockname(2)");
-    return rb_tainted_str_new(buf, len);
+    return rb_str_new(buf, len);
 }
 
 static VALUE
@@ -359,7 +337,7 @@ bsock_getpeername(sock)
     GetOpenFile(sock, fptr);
     if (getpeername(fileno(fptr->f), (struct sockaddr*)buf, &len) < 0)
 	rb_sys_fail("getpeername(2)");
-    return rb_tainted_str_new(buf, len);
+    return rb_str_new(buf, len);
 }
 
 static VALUE
@@ -398,12 +376,6 @@ bsock_send(argc, argv, sock)
 	  case EINTR:
 	    rb_thread_schedule();
 	    goto retry;
-	  case EWOULDBLOCK:
-#if EAGAIN != EWOULDBLOCK
-	  case EAGAIN:
-#endif
-	    rb_thread_fd_writable(fd);
-	    goto retry;
 	}
 	rb_sys_fail("send(2)");
     }
@@ -434,6 +406,7 @@ s_recvfrom(sock, argc, argv, from)
     char buf[1024];
     socklen_t alen = sizeof buf;
     VALUE len, flg;
+    long slen;
     int fd, flags;
 
     rb_scan_args(argc, argv, "11", &len, &flg);
@@ -447,30 +420,27 @@ s_recvfrom(sock, argc, argv, from)
     }
     fd = fileno(fptr->f);
 
-    str = rb_tainted_str_new(0, NUM2INT(len));
+    slen = NUM2INT(len);
+    str = rb_tainted_str_new(0, slen);
 
-    rb_thread_wait_fd(fd);
   retry:
+    rb_thread_wait_fd(fd);
     TRAP_BEG;
-    RSTRING(str)->len = recvfrom(fd, RSTRING(str)->ptr, RSTRING(str)->len, flags,
-				 (struct sockaddr*)buf, &alen);
+    slen = recvfrom(fd, RSTRING(str)->ptr, slen, flags, (struct sockaddr*)buf, &alen);
     TRAP_END;
 
-    if (RSTRING(str)->len < 0) {
+    if (slen < 0) {
 	switch (errno) {
 	  case EINTR:
 	    rb_thread_schedule();
 	    goto retry;
-
-	  case EWOULDBLOCK:
-#if EAGAIN != EWOULDBLOCK
-	  case EAGAIN:
-#endif
-	    rb_thread_wait_fd(fd);
-	    goto retry;
 	}
 	rb_sys_fail("recvfrom(2)");
     }
+    if (slen < RSTRING(str)->len) {
+	RSTRING(str)->len = slen;
+	RSTRING(str)->ptr[slen] = '\0';
+    }
     rb_obj_taint(str);
     switch (from) {
       case RECV_RECV:
@@ -487,7 +457,9 @@ s_recvfrom(sock, argc, argv, from)
 	return rb_assoc_new(str, unixaddr((struct sockaddr_un *)buf));
 #endif
       case RECV_SOCKET:
-	return rb_assoc_new(str, rb_tainted_str_new(buf, alen));
+	return rb_assoc_new(str, rb_str_new(buf, alen));
+      default:
+	rb_bug("s_recvfrom called with bad value");
     }
 }
 
@@ -509,6 +481,7 @@ bsock_do_not_rev_lookup()
 static VALUE
 bsock_do_not_rev_lookup_set(self, val)
 {
+    rb_secure(4);
     do_not_reverse_lookup = RTEST(val);
     return val;
 }
@@ -534,7 +507,7 @@ mkipaddr(addr)
     char buf[1024];
 
     mkipaddr0(addr, buf, sizeof(buf));
-    return rb_tainted_str_new2(buf);
+    return rb_str_new2(buf);
 }
 
 static void
@@ -593,7 +566,7 @@ ip_addrsetup(host, port)
 	portp = 0;
     }
     else if (FIXNUM_P(port)) {
-	snprintf(pbuf, sizeof(pbuf), "%ld", FIX2INT(port));
+	snprintf(pbuf, sizeof(pbuf), "%d", FIX2INT(port));
 	portp = pbuf;
     }
     else {
@@ -668,14 +641,14 @@ ipaddr(sockaddr)
 	if (error) {
 	    rb_raise(rb_eSocket, "getnameinfo: %s", gai_strerror(error));
 	}
-	addr1 = rb_tainted_str_new2(hbuf);
+	addr1 = rb_str_new2(hbuf);
     }
     error = getnameinfo(sockaddr, SA_LEN(sockaddr), hbuf, sizeof(hbuf),
 			pbuf, sizeof(pbuf), NI_NUMERICHOST | NI_NUMERICSERV);
     if (error) {
-	rb_raise(rb_eSocket, "getnameinfo %s", gai_strerror(error));
+	rb_raise(rb_eSocket, "getnameinfo: %s", gai_strerror(error));
     }
-    addr2 = rb_tainted_str_new2(hbuf);
+    addr2 = rb_str_new2(hbuf);
     if (do_not_reverse_lookup) {
 	addr1 = addr2;
     }
@@ -721,6 +694,9 @@ ruby_connect(fd, sockaddr, len, socks)
 {
     int status;
     int mode;
+#if defined __CYGWIN__
+    int wait_in_progress = -1;
+#endif
 
 #if defined(HAVE_FCNTL)
     mode = fcntl(fd, F_GETFL, 0);
@@ -755,79 +731,120 @@ ruby_connect(fd, sockaddr, len, socks)
 	      case EAGAIN:
 #ifdef EINPROGRESS
 	      case EINPROGRESS:
+#if defined __CYGWIN__
+	      case EALREADY:
+		wait_in_progress = 10;
+#endif
 #endif
 		thread_write_select(fd);
 		continue;
 
+#if defined __CYGWIN__
+	      case EINVAL:
+		if (wait_in_progress-- > 0) {
+		    struct timeval tv = {0, 100000};
+		    rb_thread_wait_for(tv);
+		    continue;
+		}
+		break;
+#endif
+
 #ifdef EISCONN
 	      case EISCONN:
 		status = 0;
 		errno = 0;
 		break;
 #endif
+	      default:
+		break;
 	    }
 	}
 #ifdef HAVE_FCNTL
-	mode &= ~NONBLOCKING;
 	fcntl(fd, F_SETFL, mode);
 #endif
 	return status;
     }
 }
 
-static VALUE
-open_inet(class, h, serv, type)
-    VALUE class, h, serv;
-    int type;
+struct inetsock_arg
 {
-    struct addrinfo hints, *res, *res0;
+    VALUE sock;
+    VALUE host, serv;
+    struct addrinfo *res;
+    int type;
+    int fd;
+};
+
+static VALUE
+inetsock_cleanup(arg)
+    struct inetsock_arg *arg;
+{
+    if (arg->res) {
+	freeaddrinfo(arg->res);
+	arg->res = 0;
+    }
+    if (arg->fd >= 0) {
+	close(arg->fd);
+    }
+    return Qnil;
+}
+
+static VALUE
+open_inet_internal(arg)
+    struct inetsock_arg *arg;
+{
+    int type = arg->type;
+    struct addrinfo hints, *res;
     int fd, status;
     char *syscall;
     char pbuf[1024], *portp;
     char *host;
     int error;
 
-    if (h) {
-	Check_SafeStr(h);
-	host = RSTRING(h)->ptr;
+    if (arg->host) {
+	Check_SafeStr(arg->host);
+	host = RSTRING(arg->host)->ptr;
     }
     else {
 	host = NULL;
     }
-    if (FIXNUM_P(serv)) {
-	snprintf(pbuf, sizeof(pbuf), "%ld", FIX2UINT(serv));
+    if (FIXNUM_P(arg->serv)) {
+	snprintf(pbuf, sizeof(pbuf), "%ld", FIX2UINT(arg->serv));
 	portp = pbuf;
     }
     else {
-	Check_SafeStr(serv);
-	if (RSTRING(serv)->len >= sizeof(pbuf))
-	    rb_raise(rb_eArgError, "servicename too long (%d)", RSTRING(serv)->len);
-	strcpy(pbuf, RSTRING(serv)->ptr);
+	Check_SafeStr(arg->serv);
+	if (RSTRING(arg->serv)->len >= sizeof(pbuf))
+	    rb_raise(rb_eArgError, "servicename too long (%ld)", RSTRING(arg->serv)->len);
+	strcpy(pbuf, RSTRING(arg->serv)->ptr);
 	portp = pbuf;
     }
     MEMZERO(&hints, struct addrinfo, 1);
     hints.ai_family = PF_UNSPEC;
     hints.ai_socktype = SOCK_STREAM;
-    if (type == INET_SERVER) {
+    if (arg->type == INET_SERVER) {
 	hints.ai_flags = AI_PASSIVE;
     }
-    error = getaddrinfo(host, portp, &hints, &res0);
+    error = getaddrinfo(host, portp, &hints, &arg->res);
     if (error) {
 	rb_raise(rb_eSocket, "getaddrinfo: %s", gai_strerror(error));
     }
 
     fd = -1;
-    for (res = res0; res; res = res->ai_next) {
+    for (res = arg->res; res; res = res->ai_next) {
 	status = ruby_socket(res->ai_family,res->ai_socktype,res->ai_protocol);
 	syscall = "socket(2)";
 	fd = status;
 	if (fd < 0) {
 	    continue;
 	}
+	arg->fd = fd;
 	if (type == INET_SERVER) {
+#ifndef NT
 	    status = 1;
 	    setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
 		       (char*)&status, sizeof(status));
+#endif
 	    status = bind(fd, res->ai_addr, res->ai_addrlen);
 	    syscall = "bind(2)";
 	}
@@ -838,33 +855,39 @@ open_inet(class, h, serv, type)
 	}
 
 	if (status < 0) {
-#if defined(HAVE_CLOSESOCKET)
-	    closesocket(fd);
-#else
 	    close(fd);
-#endif
-	    fd = -1;
+	    arg->fd = fd = -1;
 	    continue;
 	} else
 	    break;
     }
     if (status < 0) {
-	if (fd >= 0)
-#if defined(HAVE_CLOSESOCKET)
-	    closesocket(fd);
-#else
-	    close(fd);
-#endif
-	freeaddrinfo(res0);
 	rb_sys_fail(syscall);
     }
 
+    arg->fd = -1;
+
     if (type == INET_SERVER)
 	listen(fd, 5);
 
     /* create new instance */
-    freeaddrinfo(res0);
-    return sock_new(class, fd);
+    return sock_new(arg->sock, fd);
+}
+
+static VALUE
+open_inet(class, h, serv, type)
+    VALUE class, h, serv;
+    int type;
+{
+    struct inetsock_arg arg;
+    arg.sock = class;
+    arg.host = h;
+    arg.serv = serv;
+    arg.res = 0;
+    arg.type = type;
+    arg.fd = -1;
+    return rb_ensure(open_inet_internal, (VALUE)&arg,
+		     inetsock_cleanup, (VALUE)&arg);
 }
 
 static VALUE
@@ -970,11 +993,11 @@ tcp_s_gethostbyname(obj, host)
 #endif
     }
     ary = rb_ary_new();
-    rb_ary_push(ary, rb_tainted_str_new2(h->h_name));
+    rb_ary_push(ary, rb_str_new2(h->h_name));
     names = rb_ary_new();
     rb_ary_push(ary, names);
     for (pch = h->h_aliases; *pch; pch++) {
-	rb_ary_push(names, rb_tainted_str_new2(*pch));
+	rb_ary_push(names, rb_str_new2(*pch));
     }
     rb_ary_push(ary, INT2NUM(h->h_addrtype));
 #ifdef h_addr
@@ -1049,9 +1072,13 @@ s_accept(class, fd, sockaddr, len)
     rb_secure(3);
   retry:
     rb_thread_wait_fd(fd);
+#if defined(_nec_ews)
+    fd2 = accept(fd, sockaddr, len);
+#else
     TRAP_BEG;
     fd2 = accept(fd, sockaddr, len);
     TRAP_END;
+#endif
     if (fd2 < 0) {
 	switch (errno) {
 	  case EMFILE:
@@ -1063,13 +1090,6 @@ s_accept(class, fd, sockaddr, len)
 	  case EINTR:
 	    rb_thread_schedule();
 	    goto retry;
-
-	  case EWOULDBLOCK:
-#if EAGAIN != EWOULDBLOCK
-	  case EAGAIN:
-#endif
-	    rb_thread_wait_fd(fd);
-	    goto retry;
 	}
 	rb_sys_fail(0);
     }
@@ -1091,6 +1111,19 @@ tcp_accept(sock)
 }
 
 #ifdef HAVE_SYS_UN_H
+struct unixsock_arg {
+    struct sockaddr_un *sockaddr;
+    int fd;
+};
+
+static VALUE
+unixsock_connect_internal(arg)
+    struct unixsock_arg *arg;
+{
+    return (VALUE)ruby_connect(arg->fd, arg->sockaddr, sizeof(*arg->sockaddr),
+			       0);
+}
+
 static VALUE
 open_unix(class, path, server)
     VALUE class;
@@ -1117,15 +1150,19 @@ open_unix(class, path, server)
         status = bind(fd, (struct sockaddr*)&sockaddr, sizeof(sockaddr));
     }
     else {
-        status = ruby_connect(fd, (struct sockaddr*)&sockaddr, sizeof(sockaddr), 0);
+	int prot;
+	struct unixsock_arg arg;
+	arg.sockaddr = &sockaddr;
+	arg.fd = fd;
+        status = rb_protect(unixsock_connect_internal, (VALUE)&arg, &prot);
+	if (prot) {
+	    close(fd);
+	    rb_jump_tag(prot);
+	}
     }
 
     if (status < 0) {
-#if defined(HAVE_CLOSESOCKET)
-	closesocket(fd);
-#else
 	close(fd);
-#endif
 	rb_sys_fail(sockaddr.sun_path);
     }
 
@@ -1210,27 +1247,44 @@ udp_s_open(argc, argv, class)
     return sock_new(class, fd);
 }
 
+struct udp_arg
+{
+    struct addrinfo *res;
+    int fd;
+};
+
+static VALUE
+udp_connect_internal(arg)
+    struct udp_arg *arg;
+{
+    int fd = arg->fd;
+    struct addrinfo *res;
+
+    for (res = arg->res; res; res = res->ai_next) {
+	if (ruby_connect(fd, res->ai_addr, res->ai_addrlen, 0) >= 0) {
+	    return Qtrue;
+	}
+    }
+    return Qfalse;
+}
+
 static VALUE
 udp_connect(sock, host, port)
     VALUE sock, host, port;
 {
     OpenFile *fptr;
     int fd;
-    struct addrinfo *res0, *res;
+    struct udp_arg arg;
+    VALUE ret;
 
     rb_secure(3);
     GetOpenFile(sock, fptr);
     fd = fileno(fptr->f);
-    res0 = ip_addrsetup(host, port);
-    for (res = res0; res; res = res->ai_next) {
-	if (ruby_connect(fd, res->ai_addr, res->ai_addrlen, 0) >= 0) {
-	    freeaddrinfo(res0);
-	    return INT2FIX(0);
-	}
-    }
-
-    freeaddrinfo(res0);
-    rb_sys_fail("connect(2)");
+    arg.res = ip_addrsetup(host, port);
+    arg.fd = fileno(fptr->f);
+    ret = rb_ensure(udp_connect_internal, (VALUE)&arg,
+		    RUBY_METHOD_FUNC(freeaddrinfo), (VALUE)arg.res);
+    if (!ret) rb_sys_fail("connect(2)");
     return INT2FIX(0);
 }
 
@@ -1292,13 +1346,6 @@ udp_send(argc, argv, sock)
 	  case EINTR:
 	    rb_thread_schedule();
 	    goto retry;
-
-	  case EWOULDBLOCK:
-#if EAGAIN != EWOULDBLOCK
-	  case EAGAIN:
-#endif
-	    rb_thread_fd_writable(fileno(f));
-	    goto retry;
 	}
     }
     freeaddrinfo(res0);
@@ -1328,7 +1375,7 @@ unix_path(sock)
 	    rb_sys_fail(0);
 	fptr->path = strdup(addr.sun_path);
     }
-    return rb_tainted_str_new2(fptr->path);
+    return rb_str_new2(fptr->path);
 }
 
 static VALUE
@@ -1366,7 +1413,7 @@ unixaddr(sockaddr)
     struct sockaddr_un *sockaddr;
 {
     return rb_assoc_new(rb_str_new2("AF_UNIX"),
-			rb_tainted_str_new2(sockaddr->sun_path));
+			rb_str_new2(sockaddr->sun_path));
 }
 
 static VALUE
@@ -1381,6 +1428,8 @@ unix_addr(sock)
 
     if (getsockname(fileno(fptr->f), (struct sockaddr*)&addr, &len) < 0)
 	rb_sys_fail("getsockname(2)");
+    if (len == 0)
+        addr.sun_path[0] = '\0';
     return unixaddr(&addr);
 }
 
@@ -1396,6 +1445,8 @@ unix_peeraddr(sock)
 
     if (getpeername(fileno(fptr->f), (struct sockaddr*)&addr, &len) < 0)
 	rb_sys_fail("getsockname(2)");
+    if (len == 0)
+        addr.sun_path[0] = '\0';
     return unixaddr(&addr);
 }
 #endif
@@ -1513,7 +1564,7 @@ static VALUE
 sock_s_socketpair(class, domain, type, protocol)
     VALUE class, domain, type, protocol;
 {
-#if !defined(NT) && !defined(__BEOS__) && !defined(__EMX__)
+#if !defined(NT) && !defined(__BEOS__) && !defined(__EMX__) && !defined(__QNXNTO__)
     int d, t, sp[2];
 
     setup_domain_and_type(domain, &d, type, &t);
@@ -1617,7 +1668,7 @@ sock_gethostname(obj)
 	rb_sys_fail("gethostname");
 
     buf[sizeof buf - 1] = '\0';
-    return rb_tainted_str_new2(buf);
+    return rb_str_new2(buf);
 }
 #else
 #ifdef HAVE_UNAME
@@ -1632,7 +1683,7 @@ sock_gethostname(obj)
 
     rb_secure(3);
     uname(&un);
-    return rb_tainted_str_new2(un.nodename);
+    return rb_str_new2(un.nodename);
 }
 #else
 static VALUE
@@ -1660,19 +1711,19 @@ mkhostent(h)
 #endif
     }
     ary = rb_ary_new();
-    rb_ary_push(ary, rb_tainted_str_new2(h->h_name));
+    rb_ary_push(ary, rb_str_new2(h->h_name));
     names = rb_ary_new();
     rb_ary_push(ary, names);
     for (pch = h->h_aliases; *pch; pch++) {
-	rb_ary_push(names, rb_tainted_str_new2(*pch));
+	rb_ary_push(names, rb_str_new2(*pch));
     }
     rb_ary_push(ary, INT2NUM(h->h_addrtype));
 #ifdef h_addr
     for (pch = h->h_addr_list; *pch; pch++) {
-	rb_ary_push(ary, rb_tainted_str_new(*pch, h->h_length));
+	rb_ary_push(ary, rb_str_new(*pch, h->h_length));
     }
 #else
-    rb_ary_push(ary, rb_tainted_str_new(h->h_addr, h->h_length));
+    rb_ary_push(ary, rb_str_new(h->h_addr, h->h_length));
 #endif
 
     return ary;
@@ -1930,7 +1981,7 @@ sock_s_getnameinfo(argc, argv)
 	    }
 	}
 	else {
-	    rb_raise(rb_eArgError, "array size should be 3 or 4, %d given",
+	    rb_raise(rb_eArgError, "array size should be 3 or 4, %ld given",
 		     RARRAY(sa)->len);
 	}
 	/* host */
@@ -1998,7 +2049,7 @@ sock_s_getnameinfo(argc, argv)
 	}
 	freeaddrinfo(res);
     }
-    return rb_assoc_new(rb_tainted_str_new2(hbuf), rb_tainted_str_new2(pbuf));
+    return rb_assoc_new(rb_str_new2(hbuf), rb_str_new2(pbuf));
 
   error_exit_addr:
     if (res) freeaddrinfo(res);
@@ -2195,6 +2246,165 @@ Init_socket()
     sock_define_const("SOL_UDP", SOL_UDP);
 #endif
 
+#ifdef	IPPROTO_IP
+    sock_define_const("IPPROTO_IP", IPPROTO_IP);
+#else
+    sock_define_const("IPPROTO_IP", 0);
+#endif
+#ifdef	IPPROTO_ICMP
+    sock_define_const("IPPROTO_ICMP", IPPROTO_ICMP);
+#else
+    sock_define_const("IPPROTO_ICMP", 1);
+#endif
+#ifdef	IPPROTO_IGMP
+    sock_define_const("IPPROTO_IGMP", IPPROTO_IGMP);
+#endif
+#ifdef	IPPROTO_GGP
+    sock_define_const("IPPROTO_GGP", IPPROTO_GGP);
+#endif
+#ifdef	IPPROTO_TCP
+    sock_define_const("IPPROTO_TCP", IPPROTO_TCP);
+#else
+    sock_define_const("IPPROTO_TCP", 6);
+#endif
+#ifdef	IPPROTO_EGP
+    sock_define_const("IPPROTO_EGP", IPPROTO_EGP);
+#endif
+#ifdef	IPPROTO_PUP
+    sock_define_const("IPPROTO_PUP", IPPROTO_PUP);
+#endif
+#ifdef	IPPROTO_UDP
+    sock_define_const("IPPROTO_UDP", IPPROTO_UDP);
+#else
+    sock_define_const("IPPROTO_UDP", 17);
+#endif
+#ifdef	IPPROTO_IDP
+    sock_define_const("IPPROTO_IDP", IPPROTO_IDP);
+#endif
+#ifdef	IPPROTO_HELLO
+    sock_define_const("IPPROTO_HELLO", IPPROTO_HELLO);
+#endif
+#ifdef	IPPROTO_ND
+    sock_define_const("IPPROTO_ND", IPPROTO_ND);
+#endif
+#ifdef	IPPROTO_TP
+    sock_define_const("IPPROTO_TP", IPPROTO_TP);
+#endif
+#ifdef	IPPROTO_XTP
+    sock_define_const("IPPROTO_XTP", IPPROTO_XTP);
+#endif
+#ifdef	IPPROTO_EON
+    sock_define_const("IPPROTO_EON", IPPROTO_EON);
+#endif
+#ifdef	IPPROTO_BIP
+    sock_define_const("IPPROTO_BIP", IPPROTO_BIP);
+#endif
+/**/
+#ifdef	IPPROTO_RAW
+    sock_define_const("IPPROTO_RAW", IPPROTO_RAW);
+#else
+    sock_define_const("IPPROTO_RAW", 255);
+#endif
+#ifdef	IPPROTO_MAX
+    sock_define_const("IPPROTO_MAX", IPPROTO_MAX);
+#endif
+
+	/* Some port configuration */
+#ifdef	IPPORT_RESERVED
+    sock_define_const("IPPORT_RESERVED", IPPORT_RESERVED);
+#else
+    sock_define_const("IPPORT_RESERVED", 1024);
+#endif
+#ifdef	IPPORT_USERRESERVED
+    sock_define_const("IPPORT_USERRESERVED", IPPORT_USERRESERVED);
+#else
+    sock_define_const("IPPORT_USERRESERVED", 5000);
+#endif
+	/* Some reserved IP v.4 addresses */
+#ifdef	INADDR_ANY
+    sock_define_const("INADDR_ANY", INADDR_ANY);
+#else
+    sock_define_const("INADDR_ANY", 0x00000000);
+#endif
+#ifdef	INADDR_BROADCAST
+    sock_define_const("INADDR_BROADCAST", INADDR_BROADCAST);
+#else
+    sock_define_const("INADDR_BROADCAST", 0xffffffff);
+#endif
+#ifdef	INADDR_LOOPBACK
+    sock_define_const("INADDR_LOOPBACK", INADDR_LOOPBACK);
+#else
+    sock_define_const("INADDR_LOOPBACK", 0x7F000001);
+#endif
+#ifdef	INADDR_UNSPEC_GROUP
+    sock_define_const("INADDR_UNSPEC_GROUP", INADDR_UNSPEC_GROUP);
+#else
+    sock_define_const("INADDR_UNSPEC_GROUP", 0xe0000000);
+#endif
+#ifdef	INADDR_ALLHOSTS_GROUP
+    sock_define_const("INADDR_ALLHOSTS_GROUP", INADDR_ALLHOSTS_GROUP);
+#else
+    sock_define_const("INADDR_ALLHOSTS_GROUP", 0xe0000001);
+#endif
+#ifdef	INADDR_MAX_LOCAL_GROUP
+    sock_define_const("INADDR_MAX_LOCAL_GROUP", INADDR_MAX_LOCAL_GROUP);
+#else
+    sock_define_const("INADDR_MAX_LOCAL_GROUP", 0xe00000ff);
+#endif
+#ifdef	INADDR_NONE
+    sock_define_const("INADDR_NONE", INADDR_NONE);
+#else
+    sock_define_const("INADDR_NONE", 0xffffffff);
+#endif
+	/* IP [gs]etsockopt options */
+#ifdef	IP_OPTIONS
+    sock_define_const("IP_OPTIONS", IP_OPTIONS);
+#endif
+#ifdef	IP_HDRINCL
+    sock_define_const("IP_HDRINCL", IP_HDRINCL);
+#endif
+#ifdef	IP_TOS
+    sock_define_const("IP_TOS", IP_TOS);
+#endif
+#ifdef	IP_TTL
+    sock_define_const("IP_TTL", IP_TTL);
+#endif
+#ifdef	IP_RECVOPTS
+    sock_define_const("IP_RECVOPTS", IP_RECVOPTS);
+#endif
+#ifdef	IP_RECVRETOPTS
+    sock_define_const("IP_RECVRETOPTS", IP_RECVRETOPTS);
+#endif
+#ifdef	IP_RECVDSTADDR
+    sock_define_const("IP_RECVDSTADDR", IP_RECVDSTADDR);
+#endif
+#ifdef	IP_RETOPTS
+    sock_define_const("IP_RETOPTS", IP_RETOPTS);
+#endif
+#ifdef	IP_MULTICAST_IF
+    sock_define_const("IP_MULTICAST_IF", IP_MULTICAST_IF);
+#endif
+#ifdef	IP_MULTICAST_TTL
+    sock_define_const("IP_MULTICAST_TTL", IP_MULTICAST_TTL);
+#endif
+#ifdef	IP_MULTICAST_LOOP
+    sock_define_const("IP_MULTICAST_LOOP", IP_MULTICAST_LOOP);
+#endif
+#ifdef	IP_ADD_MEMBERSHIP
+    sock_define_const("IP_ADD_MEMBERSHIP", IP_ADD_MEMBERSHIP);
+#endif
+#ifdef	IP_DROP_MEMBERSHIP
+    sock_define_const("IP_DROP_MEMBERSHIP", IP_DROP_MEMBERSHIP);
+#endif
+#ifdef	IP_DEFAULT_MULTICAST_TTL
+    sock_define_const("IP_DEFAULT_MULTICAST_TTL", IP_DEFAULT_MULTICAST_TTL);
+#endif
+#ifdef	IP_DEFAULT_MULTICAST_LOOP
+    sock_define_const("IP_DEFAULT_MULTICAST_LOOP", IP_DEFAULT_MULTICAST_LOOP);
+#endif
+#ifdef	IP_MAX_MEMBERSHIPS
+    sock_define_const("IP_MAX_MEMBERSHIPS", IP_MAX_MEMBERSHIPS);
+#endif
 #ifdef SO_DEBUG
     sock_define_const("SO_DEBUG", SO_DEBUG);
 #endif
@@ -2232,6 +2442,50 @@ Init_socket()
 #ifdef SO_LINGER
     sock_define_const("SO_LINGER", SO_LINGER);
 #endif
+#ifdef SO_PASSCRED
+    sock_define_const("SO_PASSCRED", SO_PASSCRED);
+#endif
+#ifdef SO_PEERCRED
+    sock_define_const("SO_PEERCRED", SO_PEERCRED);
+#endif
+#ifdef SO_RCVLOWAT
+    sock_define_const("SO_RCVLOWAT", SO_RCVLOWAT);
+#endif
+#ifdef SO_SNDLOWAT
+    sock_define_const("SO_SNDLOWAT", SO_SNDLOWAT);
+#endif
+#ifdef SO_RCVTIMEO
+    sock_define_const("SO_RCVTIMEO", SO_RCVTIMEO);
+#endif
+#ifdef SO_SNDTIMEO
+    sock_define_const("SO_SNDTIMEO", SO_SNDTIMEO);
+#endif
+
+#ifdef SO_SECURITY_AUTHENTICATION
+    sock_define_const("SO_SECURITY_AUTHENTICATION", SO_SECURITY_AUTHENTICATION);
+#endif
+#ifdef SO_SECURITY_ENCRYPTION_TRANSPORT
+    sock_define_const("SO_SECURITY_ENCRYPTION_TRANSPORT", SO_SECURITY_ENCRYPTION_TRANSPORT);
+#endif
+#ifdef SO_SECURITY_ENCRYPTION_NETWORK
+    sock_define_const("SO_SECURITY_ENCRYPTION_NETWORK", SO_SECURITY_ENCRYPTION_NETWORK);
+#endif
+
+#ifdef SO_BINDTODEVICE
+    sock_define_const("SO_BINDTODEVICE", SO_BINDTODEVICE);
+#endif
+#ifdef SO_ATTACH_FILTER
+    sock_define_const("SO_ATTACH_FILTER", SO_ATTACH_FILTER);
+#endif
+#ifdef SO_DETACH_FILTER
+    sock_define_const("SO_DETACH_FILTER", SO_DETACH_FILTER);
+#endif
+#ifdef SO_PEERNAME
+    sock_define_const("SO_PEERNAME", SO_PEERNAME);
+#endif
+#ifdef SO_TIMESTAMP
+    sock_define_const("SO_TIMESTAMP", SO_TIMESTAMP);
+#endif
 
 #ifdef SOPRI_INTERACTIVE
     sock_define_const("SOPRI_INTERACTIVE", SOPRI_INTERACTIVE);
@@ -2243,29 +2497,6 @@ Init_socket()
     sock_define_const("SOPRI_BACKGROUND", SOPRI_BACKGROUND);
 #endif
 
-#ifdef IP_MULTICAST_IF
-    sock_define_const("IP_MULTICAST_IF", IP_MULTICAST_IF);
-#endif
-#ifdef IP_MULTICAST_TTL
-    sock_define_const("IP_MULTICAST_TTL", IP_MULTICAST_TTL);
-#endif
-#ifdef IP_MULTICAST_LOOP
-    sock_define_const("IP_MULTICAST_LOOP", IP_MULTICAST_LOOP);
-#endif
-#ifdef IP_ADD_MEMBERSHIP
-    sock_define_const("IP_ADD_MEMBERSHIP", IP_ADD_MEMBERSHIP);
-#endif
-
-#ifdef IP_DEFAULT_MULTICAST_TTL
-    sock_define_const("IP_DEFAULT_MULTICAST_TTL", IP_DEFAULT_MULTICAST_TTL);
-#endif
-#ifdef IP_DEFAULT_MULTICAST_LOOP
-    sock_define_const("IP_DEFAULT_MULTICAST_LOOP", IP_DEFAULT_MULTICAST_LOOP);
-#endif
-#ifdef IP_MAX_MEMBERSHIPS
-    sock_define_const("IP_MAX_MEMBERSHIPS", IP_MAX_MEMBERSHIPS);
-#endif
-
 #ifdef IPX_TYPE
     sock_define_const("IPX_TYPE", IPX_TYPE);
 #endif
