@@ -2,8 +2,8 @@
 
   file.c -
 
-  $Author: matz $
-  $Date: 2000/09/25 09:14:59 $
+  $Author: knu $
+  $Date: 2002/10/26 15:57:23 $
   created at: Mon Nov 15 12:24:34 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -67,7 +67,7 @@ char *strrchr _((const char*,const char));
 #include <sys/stat.h>
 
 #ifndef HAVE_LSTAT
-#define lstat stat
+#define lstat rb_sys_stat
 #endif
  
 VALUE rb_cFile;
@@ -76,7 +76,7 @@ static VALUE rb_cStat;
 
 static int
 apply2files(func, vargs, arg)
-    int (*func)();
+    void (*func)();
     VALUE vargs;
     void *arg;
 {
@@ -90,8 +90,7 @@ apply2files(func, vargs, arg)
 
     for (i=0; i<args->len; i++) {
 	path = args->ptr[i];
-	if ((*func)(RSTRING(path)->ptr, arg) < 0)
-	    rb_sys_fail(RSTRING(path)->ptr);
+	(*func)(RSTRING(path)->ptr, arg);
     }
 
     return args->len;
@@ -113,7 +112,8 @@ rb_file_path(obj)
 #endif
 
 static VALUE
-stat_new(st)
+stat_new_0(klass, st)
+    VALUE klass;
     struct stat *st;
 {
     struct stat *nst;
@@ -121,7 +121,14 @@ stat_new(st)
 
     nst = ALLOC(struct stat);
     *nst = *st;
-    return Data_Wrap_Struct(rb_cStat, NULL, free, nst);
+    return Data_Wrap_Struct(klass, NULL, free, nst);
+}
+
+static VALUE
+stat_new(st)
+    struct stat *st;
+{
+    return stat_new_0(rb_cStat, st);
 }
 
 static struct stat*
@@ -138,56 +145,59 @@ static VALUE
 rb_stat_cmp(self, other)
     VALUE self, other;
 {
-    time_t t1 = get_stat(self)->st_mtime;
-    time_t t2 = get_stat(other)->st_mtime;
-    if (t1 == t2)
-	return INT2FIX(0);
-    else if (t1 < t2)
-	return INT2FIX(-1);
-    else
-	return INT2FIX(1);
+    if (rb_obj_is_kind_of(other, rb_obj_class(self))) {
+	time_t t1 = get_stat(self)->st_mtime;
+	time_t t2 = get_stat(other)->st_mtime;
+	if (t1 == t2)
+	    return INT2FIX(0);
+	else if (t1 < t2)
+	    return INT2FIX(-1);
+	else
+	    return INT2FIX(1);
+    }
+    rb_raise(rb_eTypeError, "operand is not File::Stat");
 }
 
 static VALUE
 rb_stat_dev(self)
     VALUE self;
 {
-    return INT2FIX((int)get_stat(self)->st_dev);
+    return INT2NUM(get_stat(self)->st_dev);
 }
 
 static VALUE
 rb_stat_ino(self)
     VALUE self;
 {
-    return INT2FIX((int)get_stat(self)->st_ino);
+    return UINT2NUM(get_stat(self)->st_ino);
 }
 
 static VALUE
 rb_stat_mode(self)
     VALUE self;
 {
-    return INT2FIX((int)get_stat(self)->st_mode);
+    return UINT2NUM(get_stat(self)->st_mode);
 }
 
 static VALUE
 rb_stat_nlink(self)
     VALUE self;
 {
-    return INT2FIX((int)get_stat(self)->st_nlink);
+    return UINT2NUM(get_stat(self)->st_nlink);
 }
 
 static VALUE
 rb_stat_uid(self)
     VALUE self;
 {
-    return INT2FIX((int)get_stat(self)->st_uid);
+    return UINT2NUM(get_stat(self)->st_uid);
 }
 
 static VALUE
 rb_stat_gid(self)
     VALUE self;
 {
-    return INT2FIX((int)get_stat(self)->st_gid);
+    return UINT2NUM(get_stat(self)->st_gid);
 }
 
 static VALUE
@@ -195,7 +205,7 @@ rb_stat_rdev(self)
     VALUE self;
 {
 #ifdef HAVE_ST_RDEV
-    return INT2FIX((int)get_stat(self)->st_rdev);
+    return INT2NUM(get_stat(self)->st_rdev);
 #else
     return INT2FIX(0);
 #endif
@@ -205,7 +215,7 @@ static VALUE
 rb_stat_size(self)
     VALUE self;
 {
-    return INT2FIX((int)get_stat(self)->st_size);
+    return INT2NUM(get_stat(self)->st_size);
 }
 
 static VALUE
@@ -213,7 +223,7 @@ rb_stat_blksize(self)
     VALUE self;
 {
 #ifdef HAVE_ST_BLKSIZE
-    return INT2FIX((int)get_stat(self)->st_blksize);
+    return UINT2NUM(get_stat(self)->st_blksize);
 #else
     return INT2FIX(0);
 #endif
@@ -224,7 +234,7 @@ rb_stat_blocks(self)
     VALUE self;
 {
 #ifdef HAVE_ST_BLOCKS
-    return INT2FIX((int)get_stat(self)->st_blocks);
+    return UINT2NUM(get_stat(self)->st_blocks);
 #else
     return INT2FIX(0);
 #endif
@@ -257,7 +267,7 @@ rb_stat_inspect(self)
 {
     VALUE str;
     int i;
-    struct {
+    static struct {
         char *name;
         VALUE (*func)();
     } member[] = {
@@ -314,17 +324,17 @@ rb_stat(file, st)
 #if defined DJGPP
     if (RSTRING(file)->len == 0) return -1;
 #endif
-    return stat(RSTRING(file)->ptr, st);
+    return rb_sys_stat(RSTRING(file)->ptr, st);
 }
 
 static VALUE
-rb_file_s_stat(obj, fname)
-    VALUE obj, fname;
+rb_file_s_stat(klass, fname)
+    VALUE klass, fname;
 {
     struct stat st;
 
     Check_SafeStr(fname);
-    if (stat(RSTRING(fname)->ptr, &st) == -1) {
+    if (rb_sys_stat(RSTRING(fname)->ptr, &st) == -1) {
 	rb_sys_fail(RSTRING(fname)->ptr);
     }
     return stat_new(&st);
@@ -345,8 +355,8 @@ rb_io_stat(obj)
 }
 
 static VALUE
-rb_file_s_lstat(obj, fname)
-    VALUE obj, fname;
+rb_file_s_lstat(klass, fname)
+    VALUE klass, fname;
 {
 #ifdef HAVE_LSTAT
     struct stat st;
@@ -357,7 +367,7 @@ rb_file_s_lstat(obj, fname)
     }
     return stat_new(&st);
 #else
-    return rb_file_s_stat(obj, fname);
+    return rb_file_s_stat(klass, fname);
 #endif
 }
 
@@ -386,7 +396,7 @@ group_member(gid)
     GETGROUPS_T gid;
 {
 #if !defined(NT)
-    if (getgid() ==  gid || getegid() == gid)
+    if (getgid() ==  gid)
 	return Qtrue;
 
 # ifdef HAVE_GETGROUPS
@@ -418,31 +428,30 @@ eaccess(path, mode)
 {
 #ifdef S_IXGRP
   struct stat st;
-  static int euid = -1;
+  int euid;
 
-  if (stat(path, &st) < 0) return (-1);
+  if (rb_sys_stat(path, &st) < 0) return (-1);
 
-  if (euid == -1)
-    euid = geteuid ();
-
-  if (euid == 0)
-    {
+  euid = geteuid();
+  if (euid == 0) {
       /* Root can read or write any file. */
-      if (mode != X_OK)
+      if (!(mode & X_OK))
 	return 0;
 
       /* Root can execute any file that has any one of the execute
 	 bits set. */
       if (st.st_mode & S_IXUGO)
 	return 0;
-    }
+
+      return -1;
+  }
 
   if (st.st_uid == euid)        /* owner */
-    mode <<= 6;
-  else if (group_member (st.st_gid))
-    mode <<= 3;
+      mode <<= 6;
+  else if (getegid() == st.st_gid || group_member(st.st_gid))
+      mode <<= 3;
 
-  if (st.st_mode & mode) return 0;
+  if ((st.st_mode & mode) == mode) return 0;
 
   return -1;
 #else  /* !NT */
@@ -722,7 +731,7 @@ check3rdbyte(file, mode)
 {
     struct stat st;
 
-    if (stat(file, &st) < 0) return Qfalse;
+    if (rb_sys_stat(file, &st) < 0) return Qfalse;
     if (st.st_mode & mode) return Qtrue;
     return Qfalse;
 }
@@ -757,15 +766,16 @@ test_sticky(obj, fname)
     VALUE obj, fname;
 {
 #ifdef S_ISVTX
-    return check3rdbyte(STR2CSTR(fname), S_ISVTX);
+    Check_SafeStr(fname);
+    return check3rdbyte(RSTRING(fname)->ptr, S_ISVTX);
 #else
     return Qnil;
 #endif
 }
 
 static VALUE
-rb_file_s_size(obj, fname)
-    VALUE obj, fname;
+rb_file_s_size(klass, fname)
+    VALUE klass, fname;
 {
     struct stat st;
 
@@ -815,8 +825,8 @@ rb_file_ftype(st)
 }
 
 static VALUE
-rb_file_s_ftype(obj, fname)
-    VALUE obj, fname;
+rb_file_s_ftype(klass, fname)
+    VALUE klass, fname;
 {
     struct stat st;
 
@@ -829,8 +839,8 @@ rb_file_s_ftype(obj, fname)
 }
 
 static VALUE
-rb_file_s_atime(obj, fname)
-    VALUE obj, fname;
+rb_file_s_atime(klass, fname)
+    VALUE klass, fname;
 {
     struct stat st;
 
@@ -854,8 +864,8 @@ rb_file_atime(obj)
 }
 
 static VALUE
-rb_file_s_mtime(obj, fname)
-    VALUE obj, fname;
+rb_file_s_mtime(klass, fname)
+    VALUE klass, fname;
 {
     struct stat st;
 
@@ -879,8 +889,8 @@ rb_file_mtime(obj)
 }
 
 static VALUE
-rb_file_s_ctime(obj, fname)
-    VALUE obj, fname;
+rb_file_s_ctime(klass, fname)
+    VALUE klass, fname;
 {
     struct stat st;
 
@@ -940,12 +950,12 @@ rb_file_chmod(obj, vmode)
     mode = NUM2INT(vmode);
 
     GetOpenFile(obj, fptr);
-#if defined(DJGPP) || defined(NT) || defined(__BEOS__) || defined(__EMX__)
-    if (!fptr->path) return Qnil;
-    if (chmod(fptr->path, mode) == -1)
+#ifdef HAVE_FCHMOD
+    if (fchmod(fileno(fptr->f), mode) == -1)
 	rb_sys_fail(fptr->path);
 #else
-    if (fchmod(fileno(fptr->f), mode) == -1)
+    if (!fptr->path) return Qnil;
+    if (chmod(fptr->path, mode) == -1)
 	rb_sys_fail(fptr->path);
 #endif
 
@@ -1097,8 +1107,8 @@ rb_file_s_utime(argc, argv)
 #endif
 
 static VALUE
-rb_file_s_link(obj, from, to)
-    VALUE obj, from, to;
+rb_file_s_link(klass, from, to)
+    VALUE klass, from, to;
 {
     Check_SafeStr(from);
     Check_SafeStr(to);
@@ -1109,8 +1119,8 @@ rb_file_s_link(obj, from, to)
 }
 
 static VALUE
-rb_file_s_symlink(obj, from, to)
-    VALUE obj, from, to;
+rb_file_s_symlink(klass, from, to)
+    VALUE klass, from, to;
 {
 #ifdef HAVE_SYMLINK
     Check_SafeStr(from);
@@ -1126,8 +1136,8 @@ rb_file_s_symlink(obj, from, to)
 }
 
 static VALUE
-rb_file_s_readlink(obj, path)
-    VALUE obj, path;
+rb_file_s_readlink(klass, path)
+    VALUE klass, path;
 {
 #ifdef HAVE_READLINK
     char buf[MAXPATHLEN];
@@ -1154,24 +1164,30 @@ unlink_internal(path)
 }
 
 static VALUE
-rb_file_s_unlink(obj, args)
-    VALUE obj, args;
+rb_file_s_unlink(klass, args)
+    VALUE klass, args;
 {
     int n;
 
+    rb_secure(2);
     n = apply2files(unlink_internal, args, 0);
     return INT2FIX(n);
 }
 
 static VALUE
-rb_file_s_rename(obj, from, to)
-    VALUE obj, from, to;
+rb_file_s_rename(klass, from, to)
+    VALUE klass, from, to;
 {
     Check_SafeStr(from);
     Check_SafeStr(to);
 
-    if (rename(RSTRING(from)->ptr, RSTRING(to)->ptr) < 0)
+    if (rename(RSTRING(from)->ptr, RSTRING(to)->ptr) < 0) {
+#if defined __CYGWIN__
+	extern unsigned long __attribute__((stdcall)) GetLastError();
+	errno = GetLastError(); /* This is a Cygwin bug */
+#endif
 	rb_sys_fail(RSTRING(from)->ptr);
+    }
 
     return INT2FIX(0);
 }
@@ -1211,6 +1227,7 @@ rb_file_s_expand_path(argc, argv)
     VALUE fname, dname;
     char *s, *p;
     char buf[MAXPATHLEN+2];
+    char *bend = buf + sizeof(buf) - 2;
     int tainted;
 
     rb_scan_args(argc, argv, "11", &fname, &dname);
@@ -1225,6 +1242,7 @@ rb_file_s_expand_path(argc, argv)
 	    if (!dir) {
 		rb_raise(rb_eArgError, "couldn't find HOME environment -- expanding `%s'", s);
 	    }
+	    if (strlen(dir) > MAXPATHLEN) goto toolong;
 	    strcpy(buf, dir);
 	    p = &buf[strlen(buf)];
 	    s++;
@@ -1237,6 +1255,7 @@ rb_file_s_expand_path(argc, argv)
 #endif
 	    while (*s && !isdirsep(*s)) {
 		*p++ = *s++;
+		if (p >= bend) goto toolong;
 	    }
 	    *p = '\0';
 #ifdef HAVE_PWD_H
@@ -1245,24 +1264,27 @@ rb_file_s_expand_path(argc, argv)
 		endpwent();
 		rb_raise(rb_eArgError, "user %s doesn't exist", buf);
 	    }
+	    if (strlen(pwPtr->pw_dir) > MAXPATHLEN) goto toolong;
 	    strcpy(buf, pwPtr->pw_dir);
 	    p = &buf[strlen(buf)];
 	    endpwent();
 #endif
 	}
     }
-#if defined DOSISH
+#if defined DOSISH || __CYGWIN__
     /* skip drive letter */
-    else if (isalpha(s[0]) && s[1] == ':' && isdirsep(s[2])) {
+    else if (ISALPHA(s[0]) && s[1] == ':' && isdirsep(s[2])) {
 	while (*s && !isdirsep(*s)) {
 	    *p++ = *s++;
+	    if (p >= bend) goto toolong;
 	}
     }
 #endif
     else if (!isdirsep(*s)) {
-	if (argc == 2) {
+	if (!NIL_P(dname)) {
 	    dname = rb_file_s_expand_path(1, &dname);
 	    if (OBJ_TAINTED(dname)) tainted = 1;
+	    if (strlen(RSTRING(dname)->ptr) > MAXPATHLEN) goto toolong;
 	    strcpy(buf, RSTRING(dname)->ptr);
 	}
 	else {
@@ -1279,6 +1301,7 @@ rb_file_s_expand_path(argc, argv)
     else {
 	while (*s && isdirsep(*s)) {
 	    *p++ = '/';
+	    if (p >= bend) goto toolong;
 	    s++;
 	}
 	if (p > buf && *s) p--;
@@ -1288,7 +1311,10 @@ rb_file_s_expand_path(argc, argv)
     for ( ; *s; s++) {
 	switch (*s) {
 	  case '.':
-	    if (*(s+1)) {
+	    if (!isdirsep(*p)) {
+		*++p = '.';
+	    }
+	    else if (*(s+1)) {
 		switch (*++s) {
 		  case '.':
 		    if (*(s+1) == '\0' || isdirsep(*(s+1))) { 
@@ -1298,8 +1324,8 @@ rb_file_s_expand_path(argc, argv)
 		    }
 		    else {
 			*++p = '.';
-			do *++p = '.'; while (*++s == '.');
-			--s;
+			*++p = *s;
+			if (p >= bend) goto toolong;
 		    }
 		    break;
 		  case '/':
@@ -1320,6 +1346,7 @@ rb_file_s_expand_path(argc, argv)
 	    if (!isdirsep(*p)) *++p = '/'; break;
 	  default:
 	    *++p = *s;
+	    if (p >= bend) goto toolong;
 	}
     }
   
@@ -1330,6 +1357,10 @@ rb_file_s_expand_path(argc, argv)
     fname = rb_str_new2(buf);
     if (tainted) OBJ_TAINT(fname);
     return fname;
+
+  toolong:
+    rb_raise(rb_eArgError, "argument too long (size=%ld)", RSTRING(fname)->len);
+    return Qnil;		/* not reached */
 }
 
 static int
@@ -1388,8 +1419,8 @@ rb_file_s_basename(argc, argv)
 }
 
 static VALUE
-rb_file_s_dirname(obj, fname)
-    VALUE obj, fname;
+rb_file_s_dirname(klass, fname)
+    VALUE klass, fname;
 {
     char *name, *p;
     VALUE dirname;
@@ -1407,8 +1438,8 @@ rb_file_s_dirname(obj, fname)
 }
 
 static VALUE
-rb_file_s_split(obj, path)
-    VALUE obj, path;
+rb_file_s_split(klass, path)
+    VALUE klass, path;
 {
     return rb_assoc_new(rb_file_s_dirname(Qnil, path), rb_file_s_basename(1,&path));
 }
@@ -1416,15 +1447,15 @@ rb_file_s_split(obj, path)
 static VALUE separator;
 
 static VALUE
-rb_file_s_join(obj, args)
-    VALUE obj, args;
+rb_file_s_join(klass, args)
+    VALUE klass, args;
 {
     return rb_ary_join(args, separator);
 }
 
 static VALUE
-rb_file_s_truncate(obj, path, len)
-    VALUE obj, path, len;
+rb_file_s_truncate(klass, path, len)
+    VALUE klass, path, len;
 {
     rb_secure(2);
     Check_SafeStr(path);
@@ -1509,11 +1540,14 @@ rb_thread_flock(fd, op, fptr)
     op |= LOCK_NB;
     while (flock(fd, op) < 0) {
 	switch (errno) {
-	  case EINTR:			/* can be happen? */
+          case EAGAIN:
+          case EACCES:
+#if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
 	  case EWOULDBLOCK:
+#endif
 	    rb_thread_polling();	/* busy wait */
 	    rb_io_check_closed(fptr);
-	    break;
+            continue;
 	  default:
 	    return -1;
 	}
@@ -1530,6 +1564,7 @@ rb_file_flock(obj, operation)
 {
 #ifndef __CHECKER__
     OpenFile *fptr;
+    int ret;
 
     rb_secure(2);
     GetOpenFile(obj, fptr);
@@ -1537,12 +1572,18 @@ rb_file_flock(obj, operation)
     if (fptr->mode & FMODE_WRITABLE) {
 	fflush(GetWriteFile(fptr));
     }
-    if (flock(fileno(fptr->f), NUM2INT(operation)) < 0) {
-#ifdef EWOULDBLOCK
-	if (errno == EWOULDBLOCK) {
-	    return Qfalse;
-	}
+    TRAP_BEG;
+    ret = flock(fileno(fptr->f), NUM2INT(operation));
+    TRAP_END;
+    if (ret < 0) {
+        switch (errno) {
+          case EAGAIN:
+          case EACCES:
+#if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
+          case EWOULDBLOCK:
 #endif
+              return Qfalse;
+        }
 	rb_sys_fail(fptr->path);
     }
 #endif
@@ -1711,6 +1752,30 @@ rb_f_test(argc, argv)
     return Qnil;		/* not reached */
 }
 
+static VALUE
+rb_stat_s_new(klass, fname)
+    VALUE klass, fname;
+{
+    VALUE s;
+    struct stat st;
+
+    Check_SafeStr(fname);
+    if (rb_sys_stat(RSTRING(fname)->ptr, &st) == -1) {
+	rb_sys_fail(RSTRING(fname)->ptr);
+    }
+    s = stat_new_0(klass, &st);
+    rb_obj_call_init(s, 1, &fname);
+    return s;
+}
+
+static VALUE
+rb_stat_init(klass, fname)
+    VALUE klass, fname;
+{
+    /* do nothing */
+    return Qnil;
+}
+
 static VALUE
 rb_stat_ftype(obj)
     VALUE obj;
@@ -1995,13 +2060,14 @@ is_absolute_path(path)
     const char *path;
 {
     if (path[0] == '/') return 1;
-# if defined DOSISH
+# if defined DOSISH || defined __CYGWIN__
     if (path[0] == '\\') return 1;
     if (strlen(path) > 2 && path[1] == ':') return 1;
 # endif
     return 0;
 }
 
+#ifndef DOSISH
 static int
 path_check_1(path)
     char *path;
@@ -2025,7 +2091,8 @@ path_check_1(path)
     }
     for (;;) {
 	if (stat(path, &st) == 0 && (st.st_mode & 002)) {
-	    return 0;
+           if (p) *p = '/';
+	   return 0;
 	}
 	s = strrchr(path, '/');
 	if (p) *p = '/';
@@ -2034,11 +2101,13 @@ path_check_1(path)
 	*p = '\0';
     }
 }
+#endif
 
 int
 rb_path_check(path)
     char *path;
 {
+#ifndef DOSISH
     char *p, *pend;
     const char sep = PATH_SEP_CHAR;
 
@@ -2052,12 +2121,16 @@ rb_path_check(path)
 
 	if (pend) *pend = '\0';
 	safe = path_check_1(p);
-	if (!safe) return 0;
+	if (!safe) {
+	    if (pend) *pend = sep;
+	    return 0;
+	}
 	if (!pend) break;
 	*pend = sep;
 	p = pend + 1;
 	pend = strchr(p, sep);
     }
+#endif
     return 1;
 }
 
@@ -2071,76 +2144,145 @@ is_macos_native_path(path)
 }
 #endif
 
-char*
-rb_find_file(file)
+static int
+file_load_ok(file)
     char *file;
 {
-    extern VALUE rb_load_path;
-    volatile VALUE vpath;
+    FILE *f;
+
+    if (!file) return 0;
+    f = fopen(file, "r");
+    if (f == NULL) return 0;
+    fclose(f);
+    return 1;
+}
+
+extern VALUE rb_load_path;
+
+int
+rb_find_file_ext(filep, ext)
+    VALUE *filep;
+    const char *const *ext;
+{
+    char *path, *e, *found;
+    char *f = RSTRING(*filep)->ptr;
     VALUE fname;
-    char *path;
+    int i, j;
+
+    if (f[0] == '~') {
+	fname = *filep;
+	fname = rb_file_s_expand_path(1, &fname);
+	if (rb_safe_level() >= 2 && OBJ_TAINTED(fname)) {
+	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
+	}
+	f = STR2CSTR(fname);
+	*filep = fname;
+    }
+
+    if (is_absolute_path(f)) {
+	for (i=0; ext[i]; i++) {
+	    fname = rb_str_dup(*filep);
+	    rb_str_cat2(fname, ext[i]);
+	    if (file_load_ok(RSTRING(fname)->ptr)) {
+		*filep = fname;
+		return i+1;
+	    }
+	}
+	return 0;
+    }
+
+    if (!rb_load_path) return 0;
+
+    Check_Type(rb_load_path, T_ARRAY);
+    for (i=0;i<RARRAY(rb_load_path)->len;i++) {
+	VALUE str = RARRAY(rb_load_path)->ptr[i];
+
+	Check_SafeStr(str);
+	if (RSTRING(str)->len == 0) return 0;
+	path = RSTRING(str)->ptr;
+	for (j=0; ext[j]; j++) {
+	    fname = rb_str_dup(*filep);
+	    rb_str_cat2(fname, ext[j]);
+	    found = dln_find_file(RSTRING(fname)->ptr, path);
+	    if (found && file_load_ok(found)) {
+		*filep = fname;
+		return j+1;
+	    }
+	}
+    }
+    return 0;
+}
+
+VALUE
+rb_find_file(path)
+    VALUE path;
+{
+    VALUE tmp, fname;
+    char *f = RSTRING(path)->ptr;
+    char *lpath;
     struct stat st;
 
-#if defined(__MACOS__) || defined(riscos)
-    if (is_macos_native_path(file)) {
-	FILE *f;
-
-	if (safe_level >= 2 && !rb_path_check(file)) {
-	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", file);
+    if (f[0] == '~') {
+	path = rb_file_s_expand_path(1, &path);
+	if (rb_safe_level() >= 2 && OBJ_TAINTED(path)) {
+	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
 	}
-	f= fopen(file, "r");
-	if (f == NULL) return 0;
-	fclose(f);
-	return file;
+	f = STR2CSTR(path);
+    }
+
+#if defined(__MACOS__) || defined(riscos)
+    if (is_macos_native_path(f)) {
+	if (rb_safe_level() >= 2 && !rb_path_check(f)) {
+	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
+	}
+	if (file_load_ok(f)) return path;
     }
 #endif
 
-    if (is_absolute_path(file)) {
-	FILE *f;
-
-	if (rb_safe_level() >= 2 && !rb_path_check(file)) {
-	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", file);
+    if (is_absolute_path(f)) {
+	if (rb_safe_level() >= 2 && !rb_path_check(f)) {
+	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
 	}
-	f = fopen(file, "r");
-	if (f == NULL) return 0;
-	fclose(f);
-	return file;
+	if (file_load_ok(f)) return path;
     }
 
-    if (file[0] == '~') {
-	fname = rb_str_new2(file);
-	fname = rb_file_s_expand_path(1, &fname);
-	if (rb_safe_level() >= 2 && OBJ_TAINTED(fname)) {
-	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", file);
-	}
-	file = STR2CSTR(fname);
+    if (rb_safe_level() >= 4) {
+	rb_raise(rb_eSecurityError, "loading from non-absolute path %s", f);
     }
 
     if (rb_load_path) {
 	int i;
 
 	Check_Type(rb_load_path, T_ARRAY);
-	vpath = rb_ary_new();
+	tmp = rb_ary_new();
 	for (i=0;i<RARRAY(rb_load_path)->len;i++) {
 	    VALUE str = RARRAY(rb_load_path)->ptr[i];
 	    Check_SafeStr(str);
 	    if (RSTRING(str)->len > 0) {
-		rb_ary_push(vpath, str);
+		rb_ary_push(tmp, str);
 	    }
 	}
-	vpath = rb_ary_join(vpath, rb_str_new2(PATH_SEP));
-	path = STR2CSTR(vpath);
-	if (rb_safe_level() >= 2 && !rb_path_check(path)) {
-	    rb_raise(rb_eSecurityError, "loading from unsafe path %s", path);
+	tmp = rb_ary_join(tmp, rb_str_new2(PATH_SEP));
+	if (RSTRING(tmp)->len == 0) {
+	    lpath = 0;
+	}
+	else {
+	    lpath = STR2CSTR(tmp);
+	    if (rb_safe_level() >= 2 && !rb_path_check(lpath)) {
+		rb_raise(rb_eSecurityError, "loading from unsafe path %s", lpath);
+	    }
 	}
     }
     else {
-	path = 0;
+	lpath = 0;
     }
 
-    path = dln_find_file(file, path);
-    if (path && stat(path, &st) == 0) {
-	return path;
+    if (!lpath) {
+	return 0;		/* no path, no load */
+    }
+    f = dln_find_file(f, lpath);
+    if (file_load_ok(f)) {
+	return rb_str_new2(f);
     }
     return 0;
 }
@@ -2155,6 +2297,7 @@ define_filetest_function(name, func, argc)
     rb_define_singleton_method(rb_cFile, name, func, argc);
 }
 
+void
 Init_File()
 {
     rb_mFileTest = rb_define_module("FileTest");
@@ -2172,7 +2315,7 @@ Init_File()
     define_filetest_function("file?", test_f, 1);
     define_filetest_function("zero?", test_z, 1);
     define_filetest_function("size?", test_s, 1);
-    define_filetest_function("size", test_s, 1);
+    define_filetest_function("size", rb_file_s_size, 1);
     define_filetest_function("owned?", test_owned, 1);
     define_filetest_function("grpowned?", test_grpowned, 1);
 
@@ -2194,7 +2337,6 @@ Init_File()
     rb_define_singleton_method(rb_cFile, "atime", rb_file_s_atime, 1);
     rb_define_singleton_method(rb_cFile, "mtime", rb_file_s_mtime, 1);
     rb_define_singleton_method(rb_cFile, "ctime", rb_file_s_ctime, 1);
-    rb_define_singleton_method(rb_cFile, "size",  rb_file_s_size, 1);
 
     rb_define_singleton_method(rb_cFile, "utime", rb_file_s_utime, -1);
     rb_define_singleton_method(rb_cFile, "chmod", rb_file_s_chmod, -1);
@@ -2213,18 +2355,18 @@ Init_File()
     rb_define_singleton_method(rb_cFile, "basename", rb_file_s_basename, -1);
     rb_define_singleton_method(rb_cFile, "dirname", rb_file_s_dirname, 1);
 
-    separator = rb_str_new2("/");
+    separator = rb_obj_freeze(rb_str_new2("/"));
     rb_define_const(rb_cFile, "Separator", separator);
     rb_define_const(rb_cFile, "SEPARATOR", separator);
     rb_define_singleton_method(rb_cFile, "split",  rb_file_s_split, 1);
     rb_define_singleton_method(rb_cFile, "join",   rb_file_s_join, -2);
 
 #if defined DOSISH && !defined __CYGWIN__
-    rb_define_const(rb_cFile, "ALT_SEPARATOR", rb_str_new2("\\"));
+    rb_define_const(rb_cFile, "ALT_SEPARATOR", rb_obj_freeze(rb_str_new2("\\")));
 #else
     rb_define_const(rb_cFile, "ALT_SEPARATOR", Qnil);
 #endif
-    rb_define_const(rb_cFile, "PATH_SEPARATOR", rb_str_new2(PATH_SEP));
+    rb_define_const(rb_cFile, "PATH_SEPARATOR", rb_obj_freeze(rb_str_new2(PATH_SEP)));
 
     rb_define_method(rb_cIO, "stat",  rb_io_stat, 0); /* this is IO's method */
     rb_define_method(rb_cFile, "lstat",  rb_file_lstat, 0);
@@ -2251,6 +2393,8 @@ Init_File()
     rb_define_global_function("test", rb_f_test, -1);
 
     rb_cStat = rb_define_class_under(rb_cFile, "Stat", rb_cObject);
+    rb_define_singleton_method(rb_cStat, "new",  rb_stat_s_new, 1);
+    rb_define_method(rb_cStat, "initialize", rb_stat_init, 1);
 
     rb_include_module(rb_cStat, rb_mComparable);
 
