@@ -2,8 +2,8 @@
 
   hash.c -
 
-  $Author: matz $
-  $Date: 2000/09/12 05:37:23 $
+  $Author: knu $
+  $Date: 2002/10/17 16:27:15 $
   created at: Mon Nov 22 18:51:18 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -17,6 +17,10 @@
 #include "util.h"
 #include "rubysig.h"
 
+#ifdef __APPLE__
+#include <crt_externs.h>
+#endif
+
 #define HASH_DELETED  FL_USER1
 
 static void
@@ -59,12 +63,17 @@ rb_any_cmp(a, b)
     VALUE a, b;
 {
     VALUE args[2];
-    if (FIXNUM_P(a)) {
-	if (FIXNUM_P(b)) return a != b;
+    if (FIXNUM_P(a) && FIXNUM_P(b)) {
+	return a != b;
     }
-    else if (TYPE(a) == T_STRING) {
-	if (TYPE(b) == T_STRING) return rb_str_cmp(a, b);
+    if (TYPE(a) == T_STRING && RBASIC(a)->klass == rb_cString &&
+	TYPE(b) == T_STRING && RBASIC(b)->klass == rb_cString) {
+	return rb_str_cmp(a, b);
     }
+    if (SYMBOL_P(a) && SYMBOL_P(b)) {
+	return a != b;
+    }
+    if (a == Qundef || b == Qundef) return -1;
 
     args[0] = a;
     args[1] = b;
@@ -75,27 +84,27 @@ static int
 rb_any_hash(a)
     VALUE a;
 {
-    unsigned int hval;
+    VALUE hval;
 
     switch (TYPE(a)) {
       case T_FIXNUM:
-	hval = a;
+      case T_SYMBOL:
+	return (int)a;
 	break;
 
       case T_STRING:
-	hval = rb_str_hash(a);
+	return rb_str_hash(a);
 	break;
 
       default:
 	DEFER_INTS;
 	hval = rb_funcall(a, hash, 0);
 	if (!FIXNUM_P(hval)) {
-	    hval = rb_funcall(hval, '%', 1, INT2FIX(65439));
+	    hval = rb_funcall(hval, '%', 1, INT2FIX(536870923));
 	}
 	ENABLE_INTS;
-	hval = FIX2LONG(hval);
+	return (int)FIX2LONG(hval);
     }
-    return  hval;
 }
 
 static struct st_hash_type objhash = {
@@ -235,7 +244,7 @@ rb_hash_s_create(argc, argv, klass)
     hash = rb_hash_new2(klass);
 
     for (i=0; i<argc; i+=2) {
-	st_insert(RHASH(hash)->tbl, argv[i], argv[i+1]);
+        rb_hash_aset(hash, argv[i], argv[i + 1]);
     }
 
     return hash;
@@ -310,7 +319,7 @@ rb_hash_fetch(argc, argv, hash)
     if (!st_lookup(RHASH(hash)->tbl, key, &val)) {
 	if (rb_block_given_p()) {
 	    if (argc > 1) {
-		rb_raise(rb_eArgError, "wrong # of arguments", argc);
+		rb_raise(rb_eArgError, "wrong # of arguments");
 	    }
 	    return rb_yield(key);
 	}
@@ -333,8 +342,9 @@ static VALUE
 rb_hash_set_default(hash, ifnone)
     VALUE hash, ifnone;
 {
+    rb_hash_modify(hash);
     RHASH(hash)->ifnone = ifnone;
-    return hash;
+    return ifnone;
 }
 
 static int
@@ -356,7 +366,7 @@ rb_hash_index(hash, value)
     VALUE args[2];
 
     args[0] = value;
-    args[1] = RHASH(hash)->ifnone;
+    args[1] = Qnil;
 
     st_foreach(RHASH(hash)->tbl, index_i, args);
 
@@ -505,7 +515,10 @@ static int
 replace_i(key, val, hash)
     VALUE key, val, hash;
 {
-    rb_hash_aset(hash, key, val);
+    if (key != Qundef) {
+	rb_hash_aset(hash, key, val);
+    }
+
     return ST_CONTINUE;
 }
 
@@ -792,9 +805,12 @@ rb_hash_equal(hash1, hash2)
 {
     struct equal_data data;
 
+    if (hash1 == hash2) return Qtrue;
     if (TYPE(hash2) != T_HASH) return Qfalse;
     if (RHASH(hash1)->tbl->num_entries != RHASH(hash2)->tbl->num_entries)
 	return Qfalse;
+    if (!rb_equal(RHASH(hash1)->ifnone, RHASH(hash2)->ifnone))
+	return Qfalse;
 
     data.tbl = RHASH(hash2)->tbl;
     data.result = Qtrue;
@@ -844,10 +860,23 @@ rb_hash_update(hash1, hash2)
 
 static int path_tainted = -1;
 
-#ifndef NT
-extern char **environ;
-#endif
 static char **origenviron;
+#ifdef NT
+#define GET_ENVIRON(e) (e = win32_get_environ())
+#define FREE_ENVIRON(e) win32_free_environ(e)
+static char **my_environ;
+#undef environ
+#define environ my_environ
+#elif defined(__APPLE__)
+#undef environ
+#define environ (*_NSGetEnviron())
+#define GET_ENVIRON(e) (e)
+#define FREE_ENVIRON(e)
+#else
+extern char **environ;
+#define GET_ENVIRON(e) (e)
+#define FREE_ENVIRON(e)
+#endif
 
 static VALUE
 env_delete(obj, name)
@@ -956,17 +985,20 @@ envix(nam)
 char *nam;
 {
     register int i, len = strlen(nam);
+    char **env;
 
-    for (i = 0; environ[i]; i++) {
+    env = GET_ENVIRON(environ);
+    for (i = 0; env[i]; i++) {
 	if (
 #ifdef WIN32
-	    strnicmp(environ[i],nam,len) == 0
+	    strnicmp(env[i],nam,len) == 0
 #else
-	    memcmp(environ[i],nam,len) == 0
+	    memcmp(env[i],nam,len) == 0
 #endif
-	    && environ[i][len] == '=')
+	    && env[i][len] == '=')
 	    break;			/* memcmp must come first to avoid */
     }					/* potential SEGV's */
+    FREE_ENVIRON(environ);
     return i;
 }
 
@@ -1055,8 +1087,12 @@ ruby_setenv(name, value)
 	environ = tmpenv;		/* tell exec where it is now */
     }
     if (!value) {
-	if (environ[i] != origenviron[i])
-	    free(environ[i]);
+	if (environ != origenviron) {
+	    char **envp = origenviron;
+	    while (*envp && *envp != environ[i]) envp++;
+	    if (!*envp)
+		free(environ[i]);
+	}
 	while (environ[i]) {
 	    environ[i] = environ[i+1];
 	    i++;
@@ -1122,13 +1158,13 @@ rb_f_setenv(obj, nm, val)
 	if (OBJ_TAINTED(val)) {
 	    /* already tainted, no check */
 	    path_tainted = 1;
-	    return Qtrue;
+	    return val;
 	}
 	else {
 	    path_tainted_p(value);
 	}
     }
-    return Qtrue;
+    return val;
 }
 
 static VALUE
@@ -1137,7 +1173,7 @@ env_keys()
     char **env;
     VALUE ary = rb_ary_new();
 
-    env = environ;
+    env = GET_ENVIRON(environ);
     while (*env) {
 	char *s = strchr(*env, '=');
 	if (s) {
@@ -1145,6 +1181,7 @@ env_keys()
 	}
 	env++;
     }
+    FREE_ENVIRON(environ);
     return ary;
 }
 
@@ -1154,7 +1191,7 @@ env_each_key(hash)
 {
     char **env;
 
-    env = environ;
+    env = GET_ENVIRON(environ);
     while (*env) {
 	char *s = strchr(*env, '=');
 	if (s) {
@@ -1162,6 +1199,7 @@ env_each_key(hash)
 	}
 	env++;
     }
+    FREE_ENVIRON(environ);
     return Qnil;
 }
 
@@ -1171,7 +1209,7 @@ env_values()
     char **env;
     VALUE ary = rb_ary_new();
 
-    env = environ;
+    env = GET_ENVIRON(environ);
     while (*env) {
 	char *s = strchr(*env, '=');
 	if (s) {
@@ -1179,6 +1217,7 @@ env_values()
 	}
 	env++;
     }
+    FREE_ENVIRON(environ);
     return ary;
 }
 
@@ -1188,7 +1227,7 @@ env_each_value(hash)
 {
     char **env;
 
-    env = environ;
+    env = GET_ENVIRON(environ);
     while (*env) {
 	char *s = strchr(*env, '=');
 	if (s) {
@@ -1196,6 +1235,7 @@ env_each_value(hash)
 	}
 	env++;
     }
+    FREE_ENVIRON(environ);
     return Qnil;
 }
 
@@ -1205,7 +1245,7 @@ env_each(hash)
 {
     char **env;
 
-    env = environ;
+    env = GET_ENVIRON(environ);
     while (*env) {
 	char *s = strchr(*env, '=');
 	if (s) {
@@ -1214,6 +1254,7 @@ env_each(hash)
 	}
 	env++;
     }
+    FREE_ENVIRON(environ);
     return Qnil;
 }
 
@@ -1256,13 +1297,43 @@ env_to_s()
     return rb_str_new2("ENV");
 }
 
+static VALUE
+env_inspect()
+{
+    char **env;
+    VALUE str = rb_str_new2("{");
+    VALUE i;
+
+    env = GET_ENVIRON(environ);
+    while (*env) {
+	char *s = strchr(*env, '=');
+
+	if (env != environ) {
+	    rb_str_cat2(str, ", ");
+	}
+	if (s) {
+	    rb_str_cat2(str, "\"");
+	    rb_str_cat(str, *env, s-*env);
+	    rb_str_cat2(str, "\"=>");
+	    i = rb_inspect(rb_str_new2(s+1));
+	    rb_str_append(str, i);
+	}
+	env++;
+    }
+    FREE_ENVIRON(environ);
+    rb_str_cat2(str, "}");
+    OBJ_TAINT(str);
+
+    return str;
+}
+
 static VALUE
 env_to_a()
 {
     char **env;
     VALUE ary = rb_ary_new();
 
-    env = environ;
+    env = GET_ENVIRON(environ);
     while (*env) {
 	char *s = strchr(*env, '=');
 	if (s) {
@@ -1271,6 +1342,7 @@ env_to_a()
 	}
 	env++;
     }
+    FREE_ENVIRON(environ);
     return ary;
 }
 
@@ -1284,16 +1356,26 @@ static VALUE
 env_size()
 {
     int i;
+    char **env;
 
-    for(i=0; environ[i]; i++)
+    env = GET_ENVIRON(environ);
+    for(i=0; env[i]; i++)
 	;
+    FREE_ENVIRON(environ);
     return INT2FIX(i);
 }
 
 static VALUE
 env_empty_p()
 {
-    if (environ[0] == 0) return Qtrue;
+    char **env;
+
+    env = GET_ENVIRON(environ);
+    if (env[0] == 0) {
+	FREE_ENVIRON(environ);
+	return Qtrue;
+    }
+    FREE_ENVIRON(environ);
     return Qfalse;
 }
 
@@ -1313,15 +1395,18 @@ env_has_value(dmy, value)
     char **env;
 
     if (TYPE(value) != T_STRING) return Qfalse;
-    env = environ;
+    env = GET_ENVIRON(environ);
     while (*env) {
 	char *s = strchr(*env, '=')+1;
 	if (s) {
-	    if (strncmp(s, RSTRING(value)->ptr, strlen(s)) == 0)
+	    if (strncmp(s, RSTRING(value)->ptr, strlen(s)) == 0) {
+		FREE_ENVIRON(environ);
 		return Qtrue;
+	    }
 	}
 	env++;
     }
+    FREE_ENVIRON(environ);
     return Qfalse;
 }
 
@@ -1330,18 +1415,22 @@ env_index(dmy, value)
     VALUE dmy, value;
 {
     char **env;
+    VALUE str;
 
     if (TYPE(value) != T_STRING) return Qnil;
-    env = environ;
+    env = GET_ENVIRON(environ);
     while (*env) {
 	char *s = strchr(*env, '=')+1;
 	if (s) {
 	    if (strncmp(s, RSTRING(value)->ptr, strlen(s)) == 0) {
-		return rb_tainted_str_new(*env, s-*env-1);
+		str = rb_tainted_str_new(*env, s-*env-1);
+		FREE_ENVIRON(environ);
+		return str;
 	    }
 	}
 	env++;
     }
+    FREE_ENVIRON(environ);
     return Qnil;
 }
 
@@ -1376,7 +1465,7 @@ env_to_hash()
     char **env;
     VALUE hash = rb_hash_new();
 
-    env = environ;
+    env = GET_ENVIRON(environ);
     while (*env) {
 	char *s = strchr(*env, '=');
 	if (s) {
@@ -1385,6 +1474,7 @@ env_to_hash()
 	}
 	env++;
     }
+    FREE_ENVIRON(environ);
     return hash;
 }
 
@@ -1473,6 +1563,7 @@ Init_Hash()
     rb_define_singleton_method(envtbl,"reject", env_reject, 0);
     rb_define_singleton_method(envtbl,"reject!", env_reject_bang, 0);
     rb_define_singleton_method(envtbl,"to_s", env_to_s, 0);
+    rb_define_singleton_method(envtbl,"inspect", env_inspect, 0);
     rb_define_singleton_method(envtbl,"rehash", env_none, 0);
     rb_define_singleton_method(envtbl,"to_a", env_to_a, 0);
     rb_define_singleton_method(envtbl,"index", env_index, 1);
