@@ -2,8 +2,8 @@
 
   io.c -
 
-  $Author: matz $
-  $Date: 2000/09/21 09:30:55 $
+  $Author: nobu $
+  $Date: 2002/08/01 04:54:35 $
   created at: Fri Oct 15 18:08:59 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -27,6 +27,10 @@
 # define NO_LONG_FNAME
 #endif
 
+#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(sun) || defined(_nec_ews)
+# define USE_SETVBUF
+#endif
+
 #include <sys/types.h>
 #if !defined(DJGPP) && !defined(NT) && !defined(__human68k__)
 #include <sys/ioctl.h>
@@ -47,14 +51,11 @@ struct timeval {
 };
 #endif
 #endif
-#ifdef HAVE_VFORK_H
-#include <vfork.h>
-#endif
 
 #include <sys/stat.h>
 
-/* EMX has sys/parm.h, but.. */
-#if defined(HAVE_SYS_PARAM_H) && !defined(__EMX__)
+/* EMX has sys/param.h, but.. */
+#if defined(HAVE_SYS_PARAM_H) && !(defined(__EMX__) || defined(__HIUX_MPP__))
 # include <sys/param.h>
 #else
 # define NOFILE 64
@@ -110,10 +111,10 @@ static VALUE lineno;
 #  define READ_DATA_PENDING(fp) ((fp)->FILE_COUNT > 0)
 #elif defined(__BEOS__)
 #  define READ_DATA_PENDING(fp) (fp->_state._eof == 0)
+#elif defined(__UCLIBC__) && !defined(_UC_IOFBF)
+#  define READ_DATA_PENDING(fp) ((fp)->bufpos < (fp)->bufread)
 #else
-/* requires systems own version of the ReadDataPending() */
-extern int ReadDataPending();
-#  define READ_DATA_PENDING(fp) ReadDataPending(fp)
+#  define READ_DATA_PENDING(fp) (!feof(fp))
 #endif
 
 #define READ_CHECK(fp) do {\
@@ -129,6 +130,15 @@ rb_eof_error()
     rb_raise(rb_eEOFError, "End of file reached");
 }
 
+VALUE
+rb_io_taint_check(io)
+    VALUE io;
+{
+    if (!OBJ_TAINTED(io) && rb_safe_level() >= 4)
+	rb_raise(rb_eSecurityError, "Insecure: operation on untainted IO");
+    return io;
+}
+
 void
 rb_io_check_closed(fptr)
     OpenFile *fptr;
@@ -194,6 +204,20 @@ rb_dup(orig)
     return fd;
 }
 
+static void
+io_fflush(f, path)
+    FILE *f;
+    const char *path;
+{
+    int n;
+
+    rb_thread_fd_writable(fileno(f));
+    TRAP_BEG;
+    n = fflush(f);
+    TRAP_END;
+    if (n == EOF) rb_sys_fail(path);
+}
+
 /* writing functions */
 static VALUE
 io_write(io, str)
@@ -223,19 +247,19 @@ io_write(io, str)
 	n = (int)RSTRING(str)->len;
 	while (--n >= 0)
 	    if (fputc(*ptr++, f) == EOF)
-		rb_sys_fail(fptr->path);
+		break;
 	n = ptr - RSTRING(str)->ptr;
     }
-    if (ferror(f))
+    if (n != RSTRING(str)->len && ferror(f))
 	rb_sys_fail(fptr->path);
 #else
     n = fwrite(RSTRING(str)->ptr, 1, RSTRING(str)->len, f);
-    if (ferror(f)) {
+    if (n != RSTRING(str)->len && ferror(f)) {
 	rb_sys_fail(fptr->path);
     }
 #endif
     if (fptr->mode & FMODE_SYNC) {
-	if (fflush(f) == EOF) rb_sys_fail(fptr->path);
+	io_fflush(f, fptr->path);
     }
 
     return INT2FIX(n);
@@ -248,7 +272,7 @@ rb_io_write(io, str)
     return rb_funcall(io, id_write, 1, str);
 }
 
-static VALUE
+VALUE
 rb_io_addstr(io, str)
     VALUE io, str;
 {
@@ -266,8 +290,8 @@ rb_io_flush(io)
     GetOpenFile(io, fptr);
     rb_io_check_writable(fptr);
     f = GetWriteFile(fptr);
-
-    if (fflush(f) == EOF) rb_sys_fail(fptr->path);
+    
+    io_fflush(f, fptr->path);
 
     return io;
 }
@@ -430,16 +454,16 @@ rb_io_to_io(io)
 
 /* reading functions */
 
-static size_t
+static long
 io_fread(ptr, len, f)
     char *ptr;
-    size_t len;
+    long len;
     FILE *f;
 {
-    size_t n = len;
+    long n = len;
     int c;
 
-    while (n--) {
+    while (n > 0) {
 	if (!READ_DATA_PENDING(f)) {
 	    rb_thread_wait_fd(fileno(f));
 	}
@@ -449,15 +473,20 @@ io_fread(ptr, len, f)
 	if (c == EOF) {
 	    if (ferror(f)) {
 		if (errno == EINTR) continue;
-		rb_sys_fail(0);
+		if (errno == EAGAIN) break;
+#if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
+		if (errno == EWOULDBLOCK) break;
+#endif
+		return 0;
 	    }
 	    *ptr = '\0';
 	    break;
 	}
 	*ptr++ = c;
+	--n;
     }
 
-    return len - n - 1;
+    return len - n;
 }
 
 #ifndef S_ISREG
@@ -476,6 +505,7 @@ read_all(port)
     long siz = BUFSIZ;
     long bytes = 0;
     int n;
+    long pos = 0;
 
     GetOpenFile(port, fptr);
     rb_io_check_readable(fptr);
@@ -487,22 +517,16 @@ read_all(port)
 #endif
 	)
     {
-	if (st.st_size == 0) {
-	    getc(fptr->f);	/* force EOF */
-	    return rb_str_new(0, 0);
-	}
-	else {
-	    long pos = ftell(fptr->f);
-	    if (st.st_size > pos && pos >= 0) {
-		siz = st.st_size - pos + 1;
-	    }
+	pos = ftell(fptr->f);
+	if (st.st_size > pos && pos >= 0) {
+	    siz = st.st_size - pos + 1;
 	}
     }
     str = rb_tainted_str_new(0, siz);
     READ_CHECK(fptr->f);
     for (;;) {
 	n = io_fread(RSTRING(str)->ptr+bytes, siz-bytes, fptr->f);
-	if (n == 0 && bytes == 0) {
+	if (pos > 0 && n == 0 && bytes == 0) {
 	    if (feof(fptr->f)) return Qnil;
 	    rb_sys_fail(fptr->path);
 	}
@@ -892,7 +916,7 @@ rb_io_each_byte(io)
 	rb_yield(INT2FIX(c & 0xff));
     }
     if (ferror(f)) rb_sys_fail(fptr->path);
-    return Qnil;
+    return io;
 }
 
 VALUE
@@ -961,7 +985,7 @@ rb_io_ungetc(io, c)
     GetOpenFile(io, fptr);
     rb_io_check_readable(fptr);
 
-    if (ungetc(cc, fptr->f) == EOF)
+    if (ungetc(cc, fptr->f) == EOF && cc != EOF)
 	rb_sys_fail(fptr->path);
     return Qnil;
 }
@@ -990,13 +1014,10 @@ fptr_finalize(fptr)
     }
 }
 
-void
-rb_io_fptr_finalize(fptr)
+static void
+rb_io_fptr_cleanup(fptr)
     OpenFile *fptr;
 {
-    if (!fptr) return;
-    if (!fptr->f && !fptr->f2) return;
-
     if (fptr->finalize) {
 	(*fptr->finalize)(fptr);
     }
@@ -1011,6 +1032,17 @@ rb_io_fptr_finalize(fptr)
     }
 }
 
+void
+rb_io_fptr_finalize(fptr)
+    OpenFile *fptr;
+{
+    if (!fptr) return;
+    if (!fptr->f && !fptr->f2) return;
+    if (fileno(fptr->f) < 3) return;
+
+    rb_io_fptr_cleanup(fptr);
+}
+
 static void
 rb_io_fptr_close(fptr)
     OpenFile *fptr;
@@ -1021,7 +1053,7 @@ rb_io_fptr_close(fptr)
     if (!fptr->f && !fptr->f2) return;
 
     fd = fileno(fptr->f);
-    rb_io_fptr_finalize(fptr);
+    rb_io_fptr_cleanup(fptr);
     rb_thread_fd_close(fd);
 }
 
@@ -1032,9 +1064,6 @@ rb_io_close(io)
     OpenFile *fptr;
 
     fptr = RFILE(io)->fptr;
-    if (fptr->mode & FMODE_WRITABLE) {
-	rb_io_flush(io);
-    }
     rb_io_fptr_close(fptr);
     if (fptr->pid) {
 	rb_syswait(fptr->pid);
@@ -1051,6 +1080,7 @@ rb_io_close_m(io)
     if (rb_safe_level() >= 4 && !OBJ_TAINTED(io)) {
 	rb_raise(rb_eSecurityError, "Insecure: can't close");
     }
+    rb_io_check_closed(RFILE(io)->fptr);
     rb_io_close(io);
     return Qnil;
 }
@@ -1065,7 +1095,7 @@ rb_io_closed(io)
     return (fptr->f || fptr->f2)?Qfalse:Qtrue;
 }
 
-VALUE
+static VALUE
 rb_io_close_read(io)
     VALUE io;
 {
@@ -1334,6 +1364,10 @@ rb_fopen(fname, mode)
 	    rb_sys_fail(fname);
 	}
     }
+#ifdef USE_SETVBUF
+    if (setvbuf(file, NULL, _IOFBF, 0) != 0)
+	rb_warn("setvbuf() can't be honered for %s", fname);
+#endif
 #ifdef __human68k__
     fmode(file, _IOTEXT);
 #endif
@@ -1357,6 +1391,11 @@ rb_fdopen(fd, mode)
 	    rb_sys_fail(0);
 	}
     }
+#ifdef USE_SETVBUF
+    if (setvbuf(file, NULL, _IOFBF, 0) != 0)
+	rb_warn("setvbuf() can't be honered (fd=%d)", fd);
+#endif
+
     return file;
 }
 
@@ -1487,7 +1526,7 @@ pipe_finalize(fptr)
 	status = pclose(fptr->f2);
     }
     fptr->f = fptr->f2 = 0;
-#if defined DJGPP || (defined NT && !defined __BORLANDC__)
+#if defined DJGPP
     status <<= 8;
 #endif
     rb_last_status = INT2FIX(status);
@@ -1555,7 +1594,7 @@ pipe_open(pname, mode)
     }
 
   retry:
-    switch (pid = (doexec?vfork():fork())) {
+    switch ((pid = fork())) {
       case 0:			/* child */
 	if (modef & FMODE_READABLE) {
 	    close(pr[0]);
@@ -1639,6 +1678,9 @@ rb_io_popen(str, argc, argv, klass)
     if (rb_scan_args(argc, argv, "11", &pname, &pmode) == 1) {
 	mode = "r";
     }
+    else if (FIXNUM_P(pmode)) {
+	mode = rb_io_flags_mode(NUM2INT(pmode));
+    }
     else {
 	mode = STR2CSTR(pmode);
     }
@@ -1779,10 +1821,10 @@ io_reopen(io, nfile)
 	pos = ftell(orig->f);
     }
     if (orig->f2) {
-	if (fflush(orig->f2) == EOF) rb_sys_fail(orig->path);
+	io_fflush(orig->f2, orig->path);
     }
     else if (orig->mode & FMODE_WRITABLE) {
-	if (fflush(orig->f) == EOF) rb_sys_fail(orig->path);
+	io_fflush(orig->f, orig->path);
     }
     rb_thread_fd_close(fileno(fptr->f));
 
@@ -1874,12 +1916,18 @@ rb_io_reopen(argc, argv, file)
 	    fclose(fptr->f2);
 	    fptr->f2 = 0;
 	}
+
 	return file;
     }
 
     if (freopen(RSTRING(fname)->ptr, mode, fptr->f) == 0) {
 	rb_sys_fail(fptr->path);
     }
+#ifdef USE_SETVBUF
+    if (setvbuf(fptr->f, NULL, _IOFBF, 0) != 0)
+	rb_warn("setvbuf() can't be honered for %s", RSTRING(fname)->ptr);
+#endif
+
     if (fptr->f2) {
 	if (freopen(RSTRING(fname)->ptr, "w", fptr->f2) == 0) {
 	    rb_sys_fail(fptr->path);
@@ -1905,10 +1953,10 @@ rb_io_clone(io)
     MakeOpenFile(clone, fptr);
 
     if (orig->f2) {
-	if (fflush(orig->f2) == EOF) rb_sys_fail(orig->path);
+	io_fflush(orig->f2, orig->path);
     }
     else if (orig->mode & FMODE_WRITABLE) {
-	if (fflush(orig->f) == EOF) rb_sys_fail(orig->path);
+	io_fflush(orig->f, orig->path);
     }
 
     /* copy OpenFile structure */
@@ -1931,9 +1979,9 @@ rb_io_clone(io)
     }
     fd = rb_dup(fileno(orig->f));
     fptr->f = rb_fdopen(fd, mode);
-    if (fptr->f2) {
+    if (orig->f2) {
 	fd = rb_dup(fileno(orig->f2));
-	fptr->f = rb_fdopen(fd, "w");
+	fptr->f2 = rb_fdopen(fd, "w");
     }
     if (fptr->mode & FMODE_BINMODE) {
 	rb_io_binmode((VALUE)clone);
@@ -1942,7 +1990,7 @@ rb_io_clone(io)
     return (VALUE)clone;
 }
 
-static VALUE
+VALUE
 rb_io_printf(argc, argv, out)
     int argc;
     VALUE argv[];
@@ -1973,7 +2021,7 @@ rb_f_printf(argc, argv)
     return Qnil;
 }
 
-static VALUE
+VALUE
 rb_io_print(argc, argv, out)
     int argc;
     VALUE *argv;
@@ -2033,7 +2081,7 @@ rb_io_putc(io, ch)
     if (fputc(c, f) == EOF)
 	rb_sys_fail(fptr->path);
     if (fptr->mode & FMODE_SYNC) {
-	if (fflush(f) == EOF) rb_sys_fail(fptr->path);
+	io_fflush(f, fptr->path);
     }
 
     return ch;
@@ -2046,8 +2094,6 @@ rb_f_putc(recv, ch)
     return rb_io_putc(rb_defout, ch);
 }
 
-static VALUE rb_io_puts _((int, VALUE*, VALUE));
-
 static VALUE
 io_puts_ary(ary, out)
     VALUE ary, out;
@@ -2065,7 +2111,7 @@ io_puts_ary(ary, out)
     return Qnil;
 }
 
-static VALUE
+VALUE
 rb_io_puts(argc, argv, out)
     int argc;
     VALUE *argv;
@@ -2080,18 +2126,17 @@ rb_io_puts(argc, argv, out)
 	return Qnil;
     }
     for (i=0; i<argc; i++) {
-	switch (TYPE(argv[i])) {
-	  case T_NIL:
+	if (NIL_P(argv[i])) {
 	    line = rb_str_new2("nil");
-	    break;
-	  case T_ARRAY:
-	    rb_protect_inspect(io_puts_ary, argv[i], out);
-	    continue;
-	  default:
-	    line = argv[i];
-	    break;
 	}
-	line = rb_obj_as_string(line);
+	else {
+	    line = rb_check_convert_type(argv[i], T_ARRAY, "Array", "to_ary");
+	    if (!NIL_P(line)) {
+		rb_protect_inspect(io_puts_ary, line, out);
+		continue;
+	    }
+	    line = rb_obj_as_string(argv[i]);
+	}
 	rb_io_write(out, line);
 	if (RSTRING(line)->ptr[RSTRING(line)->len-1] != '\n') {
 	    rb_io_write(out, rb_default_rs);
@@ -2128,7 +2173,9 @@ rb_f_p(argc, argv)
     for (i=0; i<argc; i++) {
 	rb_p(argv[i]);
     }
-    fflush(stdout);
+    if (TYPE(rb_defout) == T_FILE) {
+    rb_io_flush(rb_defout);
+    }
     return Qnil;
 }
 
@@ -2210,8 +2257,6 @@ set_outfile(val, var, orig, stdf)
 
     GetOpenFile(val, fptr);
     rb_io_check_writable(fptr);
-
-    GetOpenFile(*var, fptr);
     f = GetWriteFile(fptr);
     dup2(fileno(f), fileno(stdf));
 
@@ -2253,7 +2298,7 @@ prep_stdio(f, mode, klass)
     return (VALUE)io;
 }
 
-static VALUE
+static void
 prep_path(io, path)
     VALUE io;
     char *path;
@@ -2449,7 +2494,11 @@ next_argv()
 		    fw = rb_fopen(fn, "w");
 #ifndef NO_SAFE_RENAME
 		    fstat(fileno(fw), &st2);
+#ifdef HAVE_FCHMOD
 		    fchmod(fileno(fw), st.st_mode);
+#else
+		    chmod(fn, st.st_mode);
+#endif
 		    if (st.st_uid!=st2.st_uid || st.st_gid!=st2.st_gid) {
 			fchown(fileno(fw), st.st_uid, st.st_gid);
 		    }
@@ -2739,6 +2788,34 @@ rb_f_select(argc, argv, obj)
     return res;			/* returns an empty array on interrupt */
 }
 
+#if !defined(MSDOS) && !defined(__human68k__)
+static int
+io_cntl(fd,cmd,narg,io_p)
+    int fd, cmd, io_p;
+    long narg;
+{
+    int retval;
+
+#ifdef HAVE_FCNTL
+    TRAP_BEG;
+# if defined(__CYGWIN__)
+    retval = io_p?ioctl(fd, cmd, (void*)narg):fcntl(fd, cmd, narg);
+# else
+    retval = io_p?ioctl(fd, cmd, narg):fcntl(fd, cmd, narg);
+# endif
+    TRAP_END;
+#else
+    if (!io_p) {
+	rb_notimplement();
+    }
+    TRAP_BEG;
+    retval = ioctl(fd, cmd, narg);
+    TRAP_END;
+#endif
+    return retval;
+}
+#endif
+
 static VALUE
 rb_io_ctl(io, req, arg, io_p)
     VALUE io, req, arg;
@@ -2748,7 +2825,6 @@ rb_io_ctl(io, req, arg, io_p)
     int cmd = NUM2ULONG(req);
     OpenFile *fptr;
     int len = 0;
-    int fd;
     long narg = 0;
     int retval;
 
@@ -2764,6 +2840,9 @@ rb_io_ctl(io, req, arg, io_p)
     else if (arg == Qtrue) {
 	narg = 1;
     }
+    else if (rb_obj_is_kind_of(arg, rb_cInteger)) {
+	narg = NUM2LONG(arg);
+    }
     else {
 	Check_Type(arg, T_STRING);
 
@@ -2788,27 +2867,16 @@ rb_io_ctl(io, req, arg, io_p)
 	RSTRING(arg)->ptr[len] = 17;	/* a little sanity check here */
 	narg = (long)RSTRING(arg)->ptr;
     }
-    fd = fileno(fptr->f);
-#ifdef HAVE_FCNTL
-    TRAP_BEG;
-# if defined(__CYGWIN__)
-    retval = io_p?ioctl(fd, cmd, (void*)narg):fcntl(fd, cmd, narg);
-# else
-    retval = io_p?ioctl(fd, cmd, narg):fcntl(fd, cmd, narg);
-# endif
-    TRAP_END;
-#else
-    if (!io_p) {
-	rb_notimplement();
-    }
-    TRAP_BEG;
-    retval = ioctl(fd, cmd, narg);
-    TRAP_END;
-#endif
+    retval = io_cntl(fileno(fptr->f), cmd, narg, io_p);
     if (retval < 0) rb_sys_fail(fptr->path);
     if (TYPE(arg) == T_STRING && RSTRING(arg)->ptr[len] != 17) {
 	rb_raise(rb_eArgError, "return value overflowed string");
     }
+
+    if (fptr->f2) {		/* call on f2 too; ignore result */
+	io_cntl(fileno(fptr->f2), cmd, narg, io_p);
+    }
+
     return INT2NUM(retval);
 #else
     rb_notimplement();
@@ -3049,8 +3117,9 @@ argf_tell()
 }
 
 static VALUE
-argf_seek(self, offset, ptrname)
-     VALUE self, offset, ptrname;
+argf_seek(argc, argv)
+    int argc;
+    VALUE *argv;
 {
     if (!next_argv()) {
 	rb_raise(rb_eArgError, "no stream to seek");
@@ -3059,7 +3128,7 @@ argf_seek(self, offset, ptrname)
     if (TYPE(current_file) != T_FILE) {
 	return argf_forward();
     }
-    return rb_io_seek(current_file, offset, ptrname);
+    return rb_io_seek(argc, argv, current_file);
 }
 
 static VALUE
@@ -3275,11 +3344,17 @@ static void
 opt_i_set(val)
     VALUE val;
 {
+    long len;
+
     if (!RTEST(val)) {
+	if (ruby_inplace_mode) free(ruby_inplace_mode);
 	ruby_inplace_mode = 0;
 	return;
     }
-    ruby_inplace_mode = STR2CSTR(val);
+    STR2CSTR(val);
+    len = RSTRING(val)->len;
+    REALLOC_N(ruby_inplace_mode, char, len + 1);
+    ((char *)MEMCPY(ruby_inplace_mode, RSTRING(val)->ptr, char, len))[len] = '\0';
 }
 
 void
@@ -3428,7 +3503,7 @@ Init_IO()
     rb_define_singleton_method(argf, "getc", argf_getc, 0);
     rb_define_singleton_method(argf, "readchar", argf_readchar, 0);
     rb_define_singleton_method(argf, "tell", argf_tell, 0);
-    rb_define_singleton_method(argf, "seek", argf_seek, 2);
+    rb_define_singleton_method(argf, "seek", argf_seek, -1);
     rb_define_singleton_method(argf, "rewind", argf_rewind, 0);
     rb_define_singleton_method(argf, "pos", argf_tell, 0);
     rb_define_singleton_method(argf, "pos=", argf_set_pos, 1);
