@@ -1,9 +1,9 @@
 #
-#   ruby-token.rb - ruby tokens 
-#   	$Release Version: 0.6$
-#   	$Revision: 1.2 $
-#   	$Date: 1999/03/05 03:14:01 $
-#   	by Keiju ISHITSUKA(Nippon Rational Inc.)
+#   irb/ruby-token.rb - ruby tokens 
+#   	$Release Version: 0.7.3$
+#   	$Revision: 1.1.2.2 $
+#   	$Date: 2001/06/27 15:46:05 $
+#   	by Keiju ISHITSUKA(keiju@ishitsuka.com)
 #
 # --
 #
@@ -17,6 +17,11 @@ module RubyToken
   EXPR_FNAME = :EXPR_FNAME
   EXPR_DOT = :EXPR_DOT
   EXPR_CLASS = :EXPR_CLASS
+
+  # for ruby 1.4X
+  if !defined?(Symbol)
+    Symbol = Integer
+  end
   
   class Token
     def initialize(seek, line_no, char_no)
@@ -59,7 +64,7 @@ module RubyToken
   class TkOPASGN < TkOp
     def initialize(seek, line_no, char_no, op)
       super(seek, line_no, char_no)
-      op = TkReading2Token[op] unless op.kind_of?(Integer)
+      op = TkReading2Token[op] unless op.kind_of?(Symbol)
       @op = op
     end
     attr :op
@@ -87,7 +92,7 @@ module RubyToken
 	tk.name = token
       end
       return tk
-    when Integer
+    when Symbol
       if (tk = TkSymbol2Token[token]).nil?
 	IRB.fail TkSymbol2TokenNoKey, token
       end
@@ -213,7 +218,7 @@ module RubyToken
 
     [:TkBACKQUOTE,  TkOp,   "`"],
 
-    [:TkASSGIN,     Token,  "="],
+    [:TkASSIGN,     Token,  "="],
     [:TkDOT,	    Token,  "."],
     [:TkLPAREN,     Token,  "("],  #(exp)
     [:TkLBRACK,     Token,  "["],  #[arry]
@@ -241,7 +246,7 @@ module RubyToken
   TkSymbol2Token = {}
 
   def RubyToken.def_token(token_n, super_token = Token, reading = nil, *opts)
-    token_n = token_n.id2name unless token_n.kind_of?(String)
+    token_n = token_n.id2name if token_n.kind_of?(Symbol)
     if RubyToken.const_defined?(token_n)
       IRB.fail AlreadyDefinedToken, token_n
     end
