@@ -40,7 +40,8 @@ module Profiler__
     MAP["#toplevel"][1] = total
 #    f = open("./rmon.out", "w")
     f = STDERR
-    data = MAP.values.sort!{|a,b| b[2] <=> a[2]}
+    data = MAP.values
+    data.sort!{|a,b| b[2] <=> a[2]}
     sum = 0
     f.printf "  %%   cumulative   self              self     total\n"           
     f.printf " time   seconds   seconds    calls  ms/call  ms/call  name\n"
@@ -49,7 +50,6 @@ module Profiler__
       f.printf "%6.2f %8.2f  %8.2f %8d ", d[2]/total*100, sum, d[2], d[0]
       f.printf "%8.2f %8.2f  %s\n", d[2]*1000/d[0], d[1]*1000/d[0], d[3]
     end
-    p total
     f.close
   }
   set_trace_func p
