@@ -1,8 +1,9 @@
 #
 #		thread.rb - thread support classes
-#			$Date: 2000/08/07 05:02:10 $
+#			$Date: 2002/09/11 07:29:56 $
 #			by Yukihiro Matsumoto <matz@netlab.co.jp>
 #
+# Copyright (C) 2001  Yukihiro Matsumoto
 # Copyright (C) 2000  Network Applied Communication Laboratory, Inc.
 # Copyright (C) 2000  Information-technology Promotion Agency, Japan
 #
@@ -74,7 +75,10 @@ class Mutex
       retry
     end
     Thread.critical = false
-    t.run if t
+    begin
+      t.run if t
+    rescue ThreadError
+    end
     self
   end
 
@@ -160,41 +164,33 @@ class Queue
     ensure
       Thread.critical = false
     end
-    t.run if t
-  end
-  def enq(obj)
-    push(obj)
-  end
-
-  def pop(non_block=false)
-    Thread.critical = true
     begin
-      loop do
-	if @que.length == 0
-	  if non_block
-	    raise ThreadError, "queue empty"
-	  end
-	  @waiting.push Thread.current
-	  Thread.stop
-	else
-	  return @que.shift
-	end
-      end
-    ensure
-      Thread.critical = false
+      t.run if t
+    rescue ThreadError
     end
   end
-  def shift(non_block=false)
-    pop(non_block=false)
+  alias << push
+  alias enq push
+
+  def pop(non_block=false)
+    while (Thread.critical = true; @que.empty?)
+      raise ThreadError, "queue empty" if non_block
+      @waiting.push Thread.current
+      Thread.stop
+    end
+    @que.shift
+  ensure
+    Thread.critical = false
   end
-  alias deq shift
+  alias shift pop
+  alias deq pop
 
   def empty?
-    @que.length == 0
+    @que.empty?
   end
 
   def clear
-    @que.replace([])
+    @que.clear
   end
 
   def length
@@ -211,6 +207,7 @@ end
 
 class SizedQueue<Queue
   def initialize(max)
+    raise ArgumentError, "queue size must be positive" unless max > 0
     @max = max
     @queue_wait = []
     @queue_wait.taint		# enable tainted comunication
@@ -223,7 +220,7 @@ class SizedQueue<Queue
 
   def max=(max)
     Thread.critical = true
-    if @max >= max
+    if max <= @max
       @max = max
       Thread.critical = false
     else
@@ -251,8 +248,11 @@ class SizedQueue<Queue
     end
     super
   end
+  alias << push
+  alias enq push
 
   def pop(*args)
+    retval = super
     Thread.critical = true
     if @que.length < @max
       begin
@@ -263,10 +263,15 @@ class SizedQueue<Queue
       ensure
 	Thread.critical = false
       end
-      t.run if t
+      begin
+	t.run if t
+      rescue ThreadError
+      end
     end
-    super
+    retval
   end
+  alias shift pop
+  alias deq pop
 
   def num_waiting
     @waiting.size + @queue_wait.size
