@@ -2,8 +2,8 @@
 
   main.c -
 
-  $Author: eban $
-  $Date: 2000/05/29 02:10:18 $
+  $Author: matz $
+  $Date: 2000/10/02 07:48:32 $
   created at: Fri Aug 19 13:19:58 JST 1994
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -28,6 +28,11 @@ int _CRT_glob = 0;
 #include <console.h>
 #endif
 
+/* to link startup code with ObjC support */
+#if defined(__APPLE__) && defined(__MACH__)
+static void objcdummyfunction( void ) { objc_msgSend(); }
+#endif
+
 int
 main(argc, argv, envp)
     int argc;
