@@ -2,8 +2,8 @@
 
   marshal.c -
 
-  $Author: matz $
-  $Date: 2000/09/19 07:54:21 $
+  $Author: nobu $
+  $Date: 2002/12/12 07:35:27 $
   created at: Thu Apr 27 16:30:01 JST 1995
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -14,12 +14,45 @@
 #include "rubyio.h"
 #include "st.h"
 
-#ifndef atof
+#if !defined(atof) && !defined(HAVE_STDLIB_H)
 double strtod();
 #endif
 
+#if SIZEOF_INT*2 <= SIZEOF_LONG_LONG || SIZEOF_INT*2 <= SIZEOF___INT64
+typedef unsigned int BDIGIT;
+#define SIZEOF_BDIGITS SIZEOF_INT
+#else
+typedef unsigned short BDIGIT;
+#define SIZEOF_BDIGITS SIZEOF_SHORT
+#endif
+
+#define BITSPERSHORT (2*CHAR_BIT)
+#define SHORTMASK ((1<<BITSPERSHORT)-1)
+#define SHORTDN(x) RSHIFT(x,BITSPERSHORT)
+
+#if SIZEOF_SHORT == SIZEOF_BDIGITS
+#define SHORTLEN(x) (x)
+#else
+static int
+shortlen(len, ds)
+    long len;
+    BDIGIT *ds;
+{
+    BDIGIT num;
+    int offset = 0;
+
+    num = ds[len-1];
+    while (num) {
+	num = SHORTDN(num);
+	offset++;
+    }
+    return (len - 1)*sizeof(BDIGIT)/2 + offset;
+}
+#define SHORTLEN(x) shortlen((x),d)
+#endif
+
 #define MARSHAL_MAJOR   4
-#define MARSHAL_MINOR   4
+#define MARSHAL_MINOR   6
 
 #define TYPE_NIL	'0'
 #define TYPE_TRUE	'T'
@@ -95,11 +128,8 @@ w_short(x, arg)
     int x;
     struct dump_arg *arg;
 {
-    int i;
-
-    for (i=0; i<sizeof(short); i++) {
-	w_byte((x >> (i*8)) & 0xff, arg);
-    }
+    w_byte((x >> 0) & 0xff, arg);
+    w_byte((x >> 8) & 0xff, arg);
 }
 
 static void
@@ -110,10 +140,25 @@ w_long(x, arg)
     char buf[sizeof(long)+1];
     int i, len = 0;
 
+#if SIZEOF_LONG > 4
+    if (!(RSHIFT(x, 31) == 0 || RSHIFT(x, 31) == -1)) {
+	/* big long does not fit in 4 bytes */
+	rb_raise(rb_eTypeError, "long too big to dump");
+    }
+#endif
+
     if (x == 0) {
 	w_byte(0, arg);
 	return;
     }
+    if (0 < x && x < 123) {
+	w_byte(x + 5, arg);
+	return;
+    }
+    if (-124 < x && x < 0) {
+	w_byte((x - 5)&0xff, arg);
+	return;
+    }
     for (i=1;i<sizeof(long)+1;i++) {
 	buf[i] = x & 0xff;
 	x = RSHIFT(x,8);
@@ -139,7 +184,7 @@ w_float(d, arg)
 {
     char buf[100];
 
-    sprintf(buf, "%.12g", d);
+    sprintf(buf, "%.16g", d);
     w_bytes(buf, strlen(buf), arg);
 }
 
@@ -167,6 +212,9 @@ w_unique(s, arg)
     char *s;
     struct dump_arg *arg;
 {
+    if (s[0] == '#') {
+	rb_raise(rb_eArgError, "can't dump anonymous class %s", s);
+    }
     w_symbol(rb_intern(s), arg);
 }
 
@@ -198,7 +246,7 @@ w_uclass(obj, klass, arg)
     VALUE obj, klass;
     struct dump_arg *arg;
 {
-    if (CLASS_OF(obj) != klass) {
+    if (rb_obj_class(obj) != klass) {
 	w_byte(TYPE_UCLASS, arg);
 	w_unique(rb_class2name(CLASS_OF(obj)), arg);
     }
@@ -244,7 +292,7 @@ w_object(obj, arg, limit)
 	w_byte(TYPE_FIXNUM, arg);
 	w_long(FIX2INT(obj), arg);
 #else
-	if (RSHIFT((long)obj, 32) == 0 || RSHIFT((long)obj, 32) == -1) {
+	if (RSHIFT((long)obj, 31) == 0 || RSHIFT((long)obj, 31) == -1) {
 	    w_byte(TYPE_FIXNUM, arg);
 	    w_long(FIX2LONG(obj), arg);
 	}
@@ -293,9 +341,16 @@ w_object(obj, arg, limit)
 
 	switch (BUILTIN_TYPE(obj)) {
 	  case T_CLASS:
+	    if (FL_TEST(obj, FL_SINGLETON)) {
+		rb_raise(rb_eTypeError, "singleton class can't be dumped");
+	    }
 	    w_byte(TYPE_CLASS, arg);
 	    {
 		VALUE path = rb_class_path(obj);
+		if (RSTRING(path)->ptr[0] == '#') {
+		    rb_raise(rb_eArgError, "can't dump anonymous class %s",
+			     RSTRING(path)->ptr);
+		}
 		w_bytes(RSTRING(path)->ptr, RSTRING(path)->len, arg);
 	    }
 	    break;
@@ -304,6 +359,10 @@ w_object(obj, arg, limit)
 	    w_byte(TYPE_MODULE, arg);
 	    {
 		VALUE path = rb_class_path(obj);
+		if (RSTRING(path)->ptr[0] == '#') {
+		    rb_raise(rb_eArgError, "can't dump anonymous module %s",
+			     RSTRING(path)->ptr);
+		}
 		w_bytes(RSTRING(path)->ptr, RSTRING(path)->len, arg);
 	    }
 	    break;
@@ -317,13 +376,24 @@ w_object(obj, arg, limit)
 	    w_byte(TYPE_BIGNUM, arg);
 	    {
 		char sign = RBIGNUM(obj)->sign?'+':'-';
-		int len = RBIGNUM(obj)->len;
-		unsigned short *d = RBIGNUM(obj)->digits;
+		long len = RBIGNUM(obj)->len;
+		BDIGIT *d = RBIGNUM(obj)->digits;
 
 		w_byte(sign, arg);
-		w_long(len, arg);
+		w_long(SHORTLEN(len), arg); /* w_short? */
 		while (len--) {
+#if SIZEOF_BDIGITS > SIZEOF_SHORT
+		    BDIGIT num = *d;
+		    int i;
+
+		    for (i=0; i<SIZEOF_BDIGITS; i+=SIZEOF_SHORT) {
+			w_short(num & SHORTMASK, arg);
+			num = SHORTDN(num);
+			if (len == 0 && num == 0) break;
+		    }
+#else
 		    w_short(*d, arg);
+#endif
 		    d++;
 		}
 	    }
@@ -346,7 +416,7 @@ w_object(obj, arg, limit)
 	    w_uclass(obj, rb_cArray, arg);
 	    w_byte(TYPE_ARRAY, arg);
 	    {
-		int len = RARRAY(obj)->len;
+		long len = RARRAY(obj)->len;
 		VALUE *ptr = RARRAY(obj)->ptr;
 
 		w_long(len, arg);
@@ -375,14 +445,16 @@ w_object(obj, arg, limit)
 	  case T_STRUCT:
 	    w_byte(TYPE_STRUCT, arg);
 	    {
-		int len = RSTRUCT(obj)->len;
-		char *path = rb_class2name(CLASS_OF(obj));
-		VALUE mem;
-		int i;
+		long len = RSTRUCT(obj)->len;
+		VALUE c, mem;
+		long i;
 
-		w_unique(path, arg);
+		c = CLASS_OF(obj);
+		w_unique(rb_class2name(c), arg);
 		w_long(len, arg);
-		mem = rb_ivar_get(CLASS_OF(obj), rb_intern("__member__"));
+		if (FL_TEST(c, FL_SINGLETON))
+		    c = RCLASS(c)->super;
+		mem = rb_ivar_get(c, rb_intern("__member__"));
 		if (mem == Qnil) {
 		    rb_raise(rb_eTypeError, "uninitialized struct");
 		}
@@ -399,10 +471,12 @@ w_object(obj, arg, limit)
 		VALUE klass = CLASS_OF(obj);
 		char *path;
 
-		if (FL_TEST(klass, FL_SINGLETON)) {
-		    if (RCLASS(klass)->m_tbl->num_entries > 0) {
+		while (FL_TEST(klass, FL_SINGLETON) || BUILTIN_TYPE(klass) == T_ICLASS) {
+		    if (RCLASS(klass)->m_tbl->num_entries > 0 ||
+			RCLASS(klass)->iv_tbl->num_entries > 1) {
 			rb_raise(rb_eTypeError, "singleton can't be dumped");
 		    }
+		    klass = RCLASS(klass)->super;
 		}
 		path = rb_class2name(klass);
 		w_unique(path, arg);
@@ -530,12 +604,9 @@ r_short(arg)
     struct load_arg *arg;
 {
     unsigned short x;
-    int i;
 
-    x = 0;
-    for (i=0; i<sizeof(short); i++) {
-	x |= r_byte(arg)<<(i*8);
-    }
+    x =  r_byte(arg);
+    x |= r_byte(arg)<<8;
 
     return x;
 }
@@ -548,28 +619,42 @@ long_toobig(size)
 	     sizeof(long), size);
 }
 
+#undef SIGN_EXTEND_CHAR
+#if __STDC__
+# define SIGN_EXTEND_CHAR(c) ((signed char)(c))
+#else  /* not __STDC__ */
+/* As in Harbison and Steele.  */
+# define SIGN_EXTEND_CHAR(c) ((((unsigned char)(c)) ^ 128) - 128)
+#endif
+
 static long
 r_long(arg)
     struct load_arg *arg;
 {
     register long x;
-    int c = (char)r_byte(arg);
-    int i;
+    int c = SIGN_EXTEND_CHAR(r_byte(arg));
+    long i;
 
     if (c == 0) return 0;
     if (c > 0) {
-	if (c > sizeof(long)) long_toobig((int)c);
+	if (4 < c && c < 128) {
+	    return c - 5;
+	}
+	if (c > sizeof(long)) long_toobig(c);
 	x = 0;
 	for (i=0;i<c;i++) {
 	    x |= (long)r_byte(arg) << (8*i);
 	}
     }
     else {
+	if (-129 < c && c < -4) {
+	    return c + 5;
+	}
 	c = -c;
-	if (c > sizeof(long)) long_toobig((int)c);
+	if (c > sizeof(long)) long_toobig(c);
 	x = -1;
 	for (i=0;i<c;i++) {
-	    x &= ~(0xff << (8*i));
+	    x &= ~((long)0xff << (8*i));
 	    x |= (long)r_byte(arg) << (8*i);
 	}
     }
@@ -583,14 +668,14 @@ r_long(arg)
 } while (0)
 
 #define r_bytes(s, arg) do {		\
-    int r_bytes_len;			\
+    long r_bytes_len;			\
     r_bytes2((s), r_bytes_len, (arg));	\
 } while (0)
 
 static void
 r_bytes0(s, len, arg)
     char *s;
-    int len;
+    long len;
     struct load_arg *arg;
 {
     if (arg->fp) {
@@ -611,7 +696,7 @@ r_symlink(arg)
     struct load_arg *arg;
 {
     ID id;
-    int num = r_long(arg);
+    long num = r_long(arg);
 
     if (st_lookup(arg->symbol, num, &id)) {
 	return id;
@@ -655,7 +740,7 @@ r_string(arg)
     struct load_arg *arg;
 {
     char *buf;
-    int len;
+    long len;
 
     r_bytes2(buf, len, arg);
     return rb_str_new(buf, len);
@@ -666,9 +751,6 @@ r_regist(v, arg)
     VALUE v;
     struct load_arg *arg;
 {
-    if (arg->proc) {
-	rb_funcall(arg->proc, rb_intern("call"), 1, v);
-    }
     rb_hash_aset(arg->data, INT2FIX(RHASH(arg->data)->tbl->num_entries), v);
     if (arg->taint) OBJ_TAINT(v);
     return v;
@@ -679,7 +761,7 @@ r_ivar(obj, arg)
     VALUE obj;
     struct load_arg *arg;
 {
-    int len;
+    long len;
 
     len = r_long(arg);
     if (len > 0) {
@@ -695,7 +777,7 @@ static VALUE
 r_object(arg)
     struct load_arg *arg;
 {
-    VALUE v;
+    VALUE v = Qnil;
     int type = r_byte(arg);
     long id;
 
@@ -707,7 +789,6 @@ r_object(arg)
 	    rb_raise(rb_eArgError, "dump format error (unlinked)");
 	}
 	return v;
-      break;
 
       case TYPE_IVAR:
 	v = r_object(arg);
@@ -717,88 +798,139 @@ r_object(arg)
       case TYPE_UCLASS:
 	{
 	    VALUE c = rb_path2class(r_unique(arg));
+
+	    if (FL_TEST(c, FL_SINGLETON)) {
+		rb_raise(rb_eTypeError, "singleton can't be loaded");
+	    }
 	    v = r_object(arg);
-	    if (rb_special_const_p(v)) {
+	    if (rb_special_const_p(v) || TYPE(v) == T_OBJECT || TYPE(v) == T_CLASS) {
+	      format_error:
 		rb_raise(rb_eArgError, "dump format error (user class)");
 	    }
+	    if (TYPE(v) == T_MODULE || !RTEST(rb_funcall(c, '<', 1, RBASIC(v)->klass))) {
+		VALUE tmp = rb_obj_alloc(c);
+
+		if (TYPE(v) != TYPE(tmp)) goto format_error;
+	    }
 	    RBASIC(v)->klass = c;
 	    return v;
 	}
 
       case TYPE_NIL:
-	return Qnil;
+	v = Qnil;
+	break;
 
       case TYPE_TRUE:
-	return Qtrue;
+	v = Qtrue;
+	break;
 
       case TYPE_FALSE:
-	return Qfalse;
+	v = Qfalse;
+	break;
 
       case TYPE_FIXNUM:
 	{
-	    int i = r_long(arg);
-	    return INT2FIX(i);
+	    long i = r_long(arg);
+	    v = INT2FIX(i);
 	}
+	break;
 
       case TYPE_FLOAT:
 	{
 	    char *buf;
+	    double d, t = 0.0;
 
 	    r_bytes(buf, arg);
-	    v = rb_float_new(strtod(buf, 0));
-	    return r_regist(v, arg);
+	    if (strcmp(buf, "nan") == 0) {
+		d = t / t;
+	    }
+	    else if (strcmp(buf, "inf") == 0) {
+		d = 1.0 / t;
+	    }
+	    else if (strcmp(buf, "-inf") == 0) {
+		d = -1.0 / t;
+	    }
+	    else {
+		/* xxx: should not use system's strtod(3) */
+		d = strtod(buf, 0);
+	    }
+	    v = rb_float_new(d);
+	    r_regist(v, arg);
 	}
+	break;
 
       case TYPE_BIGNUM:
 	{
-	    int len;
-	    unsigned short *digits;
+	    long len;
+	    BDIGIT *digits;
 
 	    NEWOBJ(big, struct RBignum);
 	    OBJSETUP(big, rb_cBignum, T_BIGNUM);
 	    big->sign = (r_byte(arg) == '+');
-	    big->len = len = r_long(arg);
-	    big->digits = digits = ALLOC_N(unsigned short, len);
-	    while (len--) {
+	    len = r_long(arg);
+#if SIZEOF_BDIGITS == SIZEOF_SHORT
+	    big->len = len;
+#else
+	    big->len = (len + 1) * 2 / sizeof(BDIGIT);
+#endif
+	    big->digits = digits = ALLOC_N(BDIGIT, big->len);
+	    while (len > 0) {
+#if SIZEOF_BDIGITS > SIZEOF_SHORT
+		BDIGIT num = 0;
+		int shift = 0;
+		int i;
+
+		for (i=0; i<SIZEOF_BDIGITS; i+=2) {
+		    int j = r_short(arg);
+		    num |= j << shift;
+		    shift += BITSPERSHORT;
+		    if (--len == 0) break;
+		}
+		*digits++ = num;
+#else
 		*digits++ = r_short(arg);
+		len--;
+#endif
 	    }
-	    big = RBIGNUM(rb_big_norm((VALUE)big));
-	    if (TYPE(big) == T_BIGNUM) {
-		r_regist((VALUE)big, arg);
+	    v = rb_big_norm((VALUE)big);
+	    if (TYPE(v) == T_BIGNUM) {
+		r_regist(v, arg);
 	    }
-	    return (VALUE)big;
 	}
+	break;
 
       case TYPE_STRING:
-	return r_regist(r_string(arg), arg);
+	v = r_regist(r_string(arg), arg);
+	break;
 
       case TYPE_REGEXP:
 	{
 	    char *buf;
-	    int len;
+	    long len;
 	    int options;
 
 	    r_bytes2(buf, len, arg);
 	    options = r_byte(arg);
-	    return r_regist(rb_reg_new(buf, len, options), arg);
+	    v = r_regist(rb_reg_new(buf, len, options), arg);
 	}
+	break;
 
       case TYPE_ARRAY:
 	{
-	    volatile int len = r_long(arg); /* gcc 2.7.2.3 -O2 bug?? */
+	    volatile long len = r_long(arg); /* gcc 2.7.2.3 -O2 bug?? */
 
 	    v = rb_ary_new2(len);
 	    r_regist(v, arg);
 	    while (len--) {
 		rb_ary_push(v, r_object(arg));
 	    }
-	    return v;
 	}
+	break;
 
       case TYPE_HASH:
       case TYPE_HASH_DEF:
 	{
-	    int len = r_long(arg);
+	    long len = r_long(arg);
 
 	    v = rb_hash_new();
 	    r_regist(v, arg);
@@ -810,14 +942,14 @@ r_object(arg)
 	    if (type == TYPE_HASH_DEF) {
 		RHASH(v)->ifnone = r_object(arg);
 	    }
-	    return v;
 	}
+	break;
 
       case TYPE_STRUCT:
 	{
 	    VALUE klass, mem, values;
-	    volatile int i;	/* gcc 2.7.2.3 -O2 bug?? */
-	    int len;
+	    volatile long i;	/* gcc 2.7.2.3 -O2 bug?? */
+	    long len;
 	    ID slot;
 
 	    klass = rb_path2class(r_unique(arg));
@@ -844,7 +976,6 @@ r_object(arg)
 		}
 		rb_struct_aset(v, INT2FIX(i), r_object(arg));
 	    }
-	    return v;
 	}
 	break;
 
@@ -853,12 +984,12 @@ r_object(arg)
 	    VALUE klass;
 
 	    klass = rb_path2class(r_unique(arg));
-	    if (rb_respond_to(klass, s_load)) {
-		v = rb_funcall(klass, s_load, 1, r_string(arg));
-		return r_regist(v, arg);
+	    if (!rb_respond_to(klass, s_load)) {
+		rb_raise(rb_eTypeError, "class %s needs to have method `_load'",
+			 rb_class2name(klass));
 	    }
-	    rb_raise(rb_eTypeError, "class %s needs to have method `_load'",
-		     rb_class2name(klass));
+	    v = rb_funcall(klass, s_load, 1, r_string(arg));
+	    r_regist(v, arg);
 	}
         break;
 
@@ -867,10 +998,15 @@ r_object(arg)
 	    VALUE klass;
 
 	    klass = rb_path2class(r_unique(arg));
+	    if (TYPE(klass) != T_CLASS) {
+		rb_raise(rb_eArgError, "dump format error");
+	    }
 	    v = rb_obj_alloc(klass);
+	    if (TYPE(v) != T_OBJECT) {
+		rb_raise(rb_eArgError, "dump format error");
+	    }
 	    r_regist(v, arg);
 	    r_ivar(v, arg);
-	    return v;
 	}
 	break;
 
@@ -878,37 +1014,37 @@ r_object(arg)
         {
 	    char *buf;
 	    r_bytes(buf, arg);
-	    return r_regist(rb_path2class(buf), arg);
+	    v = r_regist(rb_path2class(buf), arg);
 	}
+	break;
 
       case TYPE_CLASS:
         {
-	    VALUE c;
-
 	    char *buf;
 	    r_bytes(buf, arg);
-	    c = rb_path2class(buf);
-	    if (TYPE(c) != T_CLASS) {
+	    v = rb_path2class(buf);
+	    if (TYPE(v) != T_CLASS) {
 		rb_raise(rb_eTypeError, "%s is not a class", buf);
 	    }
-	    return r_regist(c, arg);
+	    r_regist(v, arg);
 	}
+	break;
 
       case TYPE_MODULE:
         {
-	    VALUE m;
-
 	    char *buf;
 	    r_bytes(buf, arg);
-	    m = rb_path2class(buf);
-	    if (TYPE(m) != T_CLASS) {
+	    v = rb_path2class(buf);
+	    if (TYPE(v) != T_MODULE) {
 		rb_raise(rb_eTypeError, "%s is not a module", buf);
 	    }
-	    return r_regist(m, arg);
+	    r_regist(v, arg);
 	}
+	break;
 
       case TYPE_SYMBOL:
-	return ID2SYM(r_symreal(arg));
+	v = ID2SYM(r_symreal(arg));
+	break;
 
       case TYPE_SYMLINK:
 	return ID2SYM(r_symlink(arg));
@@ -917,7 +1053,10 @@ r_object(arg)
 	rb_raise(rb_eArgError, "dump format error(0x%x)", type);
 	break;
     }
-    return Qnil;		/* not reached */
+    if (arg->proc) {
+	rb_funcall(arg->proc, rb_intern("call"), 1, v);
+    }
+    return v;
 }
 
 static VALUE
@@ -941,10 +1080,11 @@ marshal_load(argc, argv)
     VALUE *argv;
 {
     VALUE port, proc;
-    int major;
+    int major, minor;
     VALUE v;
     OpenFile *fptr;
     struct load_arg arg;
+    volatile VALUE hash;	/* protect from GC */
 
     rb_scan_args(argc, argv, "11", &port, &proc);
     if (rb_obj_is_kind_of(port, rb_cIO)) {
@@ -967,21 +1107,23 @@ marshal_load(argc, argv)
     }
 
     major = r_byte(&arg);
-    if (major == MARSHAL_MAJOR) {
-	volatile VALUE hash;	/* protect from GC */
+    minor = r_byte(&arg);
+    if (major != MARSHAL_MAJOR || minor > MARSHAL_MINOR) {
+	rb_raise(rb_eTypeError, "incompatible marshal file format (can't be read)\n\
+\tformat version %d.%d required; %d.%d given",
+		 MARSHAL_MAJOR, MARSHAL_MINOR, major, minor);
+    }
+    if (RTEST(ruby_verbose) && minor != MARSHAL_MINOR) {
+	rb_warn("incompatible marshal file format (can be read)\n\
+\tformat version %d.%d required; %d.%d given",
+		MARSHAL_MAJOR, MARSHAL_MINOR, major, minor);
+    }
 
-	if (r_byte(&arg) != MARSHAL_MINOR) {
-	    rb_warn("Old marshal file format (can be read)");
-	}
-	arg.symbol = st_init_numtable();
-	arg.data   = hash = rb_hash_new();
-	if (NIL_P(proc)) arg.proc = 0;
-	else             arg.proc = proc;
-	v = rb_ensure(load, (VALUE)&arg, load_ensure, (VALUE)&arg);
-    }
-    else {
-	rb_raise(rb_eTypeError, "old marshal file format (can't read)");
-    }
+    arg.symbol = st_init_numtable();
+    arg.data   = hash = rb_hash_new();
+    if (NIL_P(proc)) arg.proc = 0;
+    else             arg.proc = proc;
+    v = rb_ensure(load, (VALUE)&arg, load_ensure, (VALUE)&arg);
 
     return v;
 }
