@@ -14,6 +14,7 @@
  */
 
 #include <sys/stat.h>
+#include <errno.h>
 /*
  * Make a directory.
  */
@@ -96,7 +97,7 @@ rmdir (dpath)
   if (WIFSIGNALED (status) || WEXITSTATUS (status) != 0)
     {
       errno = EIO;		/* We don't know why, but */
-      return -1;		/* /bin/mkdir failed */
+      return -1;		/* /bin/rmdir failed */
     }
 
   return 0;
