@@ -2,8 +2,8 @@
 
   object.c -
 
-  $Author: matz $
-  $Date: 2000/09/26 07:07:12 $
+  $Author: knu $
+  $Date: 2002/10/26 15:57:23 $
   created at: Thu Jul 15 12:01:24 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -15,6 +15,7 @@
 #include "ruby.h"
 #include "st.h"
 #include <stdio.h>
+#include <errno.h>
 
 VALUE rb_mKernel;
 VALUE rb_cObject;
@@ -71,18 +72,23 @@ rb_obj_id(obj)
     return (VALUE)((long)obj|FIXNUM_FLAG);
 }
 
-static VALUE
-rb_obj_type(obj)
-    VALUE obj;
+VALUE
+rb_class_real(cl)
+    VALUE cl;
 {
-    VALUE cl = CLASS_OF(obj);
-
     while (FL_TEST(cl, FL_SINGLETON) || TYPE(cl) == T_ICLASS) {
 	cl = RCLASS(cl)->super;
     }
     return cl;
 }
 
+VALUE
+rb_obj_class(obj)
+    VALUE obj;
+{
+    return rb_class_real(CLASS_OF(obj));
+}
+
 VALUE
 rb_obj_clone(obj)
     VALUE obj;
@@ -108,11 +114,15 @@ rb_obj_dup(obj)
     VALUE dup;
 
     dup = rb_funcall(obj, clone, 0, 0);
-    if (TYPE(dup) != TYPE(obj)) {
-	rb_raise(rb_eTypeError, "dupulicated object must be same type");
+    if (TYPE(dup) != TYPE(obj) || rb_obj_class(dup) != rb_obj_class(obj)) {
+	rb_raise(rb_eTypeError, "dupulicated object must be same class");
     }
     if (!SPECIAL_CONST_P(dup)) {
-	OBJSETUP(dup, rb_obj_type(obj), BUILTIN_TYPE(obj));
+	OBJSETUP(dup, rb_obj_class(obj), BUILTIN_TYPE(obj));
+	OBJ_INFECT(dup, obj);
+	if (FL_TEST(obj, FL_EXIVAR)) {
+	    FL_SET(dup, FL_EXIVAR);
+	}
     }
     return dup;
 }
@@ -128,13 +138,12 @@ VALUE
 rb_any_to_s(obj)
     VALUE obj;
 {
-    char *s;
     char *cname = rb_class2name(CLASS_OF(obj));
     VALUE str;
 
-    s = ALLOCA_N(char, strlen(cname)+6+16+1); /* 6:tags 16:addr 1:eos */
-    sprintf(s, "#<%s:0x%lx>", cname, obj);
-    str = rb_str_new2(s);
+    str = rb_str_new(0, strlen(cname)+6+16+1); /* 6:tags 16:addr 1:eos */
+    sprintf(RSTRING(str)->ptr, "#<%s:0x%lx>", cname, obj);
+    RSTRING(str)->len = strlen(RSTRING(str)->ptr);
     if (OBJ_TAINTED(obj)) OBJ_TAINT(str);
 
     return str;
@@ -161,6 +170,7 @@ inspect_i(id, value, str)
     if (!rb_is_instance_id(id)) return ST_CONTINUE;
     if (RSTRING(str)->ptr[0] == '-') { /* first element */
 	RSTRING(str)->ptr[0] = '#';
+	rb_str_cat2(str, " ");
     }
     else {
 	rb_str_cat2(str, ", ");
@@ -181,6 +191,7 @@ inspect_obj(obj, str)
 {
     st_foreach(ROBJECT(obj)->iv_tbl, inspect_i, str);
     rb_str_cat2(str, ">");
+    RSTRING(str)->ptr[0] = '#';
     OBJ_INFECT(str, obj);
 
     return str;
@@ -194,17 +205,18 @@ rb_obj_inspect(obj)
 	&& ROBJECT(obj)->iv_tbl
 	&& ROBJECT(obj)->iv_tbl->num_entries > 0) {
 	VALUE str;
-	char *c, *b;
+	char *c;
 
 	c = rb_class2name(CLASS_OF(obj));
 	if (rb_inspecting_p(obj)) {
-	    b = ALLOCA_N(char, strlen(c)+8+16+1); /* 8:tags 16:addr 1:eos */
-	    sprintf(b, "#<%s:0x%lx ...>", c, obj);
-	    return rb_str_new2(b);
+	    str = rb_str_new(0, strlen(c)+10+16+1); /* 10:tags 16:addr 1:eos */
+	    sprintf(RSTRING(str)->ptr, "#<%s:0x%lx ...>", c, obj);
+	    RSTRING(str)->len = strlen(RSTRING(str)->ptr);
+	    return str;
 	}
-	b = ALLOCA_N(char, strlen(c)+4+16+1); 	  /* 4:tags 16:addr 1:eos */
-	sprintf(b, "-<%s:0x%lx ", c, obj);
-	str = rb_str_new2(b);
+	str = rb_str_new(0, strlen(c)+6+16+1); /* 6:tags 16:addr 1:eos */
+	sprintf(RSTRING(str)->ptr, "-<%s:0x%lx", c, obj);
+	RSTRING(str)->len = strlen(RSTRING(str)->ptr);
 	return rb_protect_inspect(inspect_obj, obj, str);
     }
     return rb_funcall(obj, rb_intern("to_s"), 0, 0);
@@ -225,18 +237,16 @@ rb_obj_is_instance_of(obj, c)
 	return Qfalse;
 
       case T_FALSE:
-	if (obj) return Qfalse;
-	return Qtrue;
+	return RTEST(obj) ? Qfalse : Qtrue;
 
       case T_TRUE:
-	if (obj) return Qtrue;
-	return Qfalse;
+	return RTEST(obj) ? Qtrue : Qfalse;
 
       default:
 	rb_raise(rb_eTypeError, "class or module required");
     }
 
-    if (rb_obj_type(obj) == c) return Qtrue;
+    if (rb_obj_class(obj) == c) return Qtrue;
     return Qfalse;
 }
 
@@ -284,7 +294,12 @@ rb_obj_taint(obj)
     VALUE obj;
 {
     rb_secure(4);
-    OBJ_TAINT(obj);
+    if (!OBJ_TAINTED(obj)) {
+	if (OBJ_FROZEN(obj)) {
+	    rb_error_frozen("object");
+	}
+	OBJ_TAINT(obj);
+    }
     return obj;
 }
 
@@ -293,7 +308,12 @@ rb_obj_untaint(obj)
     VALUE obj;
 {
     rb_secure(3);
-    FL_UNSET(obj, FL_TAINT);
+    if (OBJ_TAINTED(obj)) {
+	if (OBJ_FROZEN(obj)) {
+	    rb_error_frozen("object");
+	}
+	FL_UNSET(obj, FL_TAINT);
+    }
     return obj;
 }
 
@@ -344,13 +364,6 @@ nil_inspect(obj)
     return rb_str_new2("nil");
 }
 
-static VALUE
-nil_type(obj)
-    VALUE obj;
-{
-    return rb_cNilClass;
-}
-
 #ifdef NIL_PLUS
 static VALUE
 nil_plus(x, y)
@@ -387,13 +400,6 @@ true_to_s(obj)
     return rb_str_new2("true");
 }
 
-static VALUE
-true_type(obj)
-    VALUE obj;
-{
-    return rb_cTrueClass;
-}
-
 static VALUE
 true_and(obj, obj2)
     VALUE obj, obj2;
@@ -422,13 +428,6 @@ false_to_s(obj)
     return rb_str_new2("false");
 }
 
-static VALUE
-false_type(obj)
-    VALUE obj;
-{
-    return rb_cFalseClass;
-}
-
 static VALUE
 false_and(obj, obj2)
     VALUE obj, obj2;
@@ -474,13 +473,6 @@ rb_obj_alloc(klass)
     return (VALUE)obj;
 }
 
-static VALUE
-sym_type(sym)
-    VALUE sym;
-{
-    return rb_cSymbol;
-}
-
 static VALUE
 sym_to_i(sym)
     VALUE sym;
@@ -494,13 +486,14 @@ static VALUE
 sym_inspect(sym)
     VALUE sym;
 {
-    char *name, *buf;
+    VALUE str;
+    char *name;
 
     name = rb_id2name(SYM2ID(sym));
-    buf = ALLOCA_N(char, strlen(name)+2);
-    sprintf(buf, ":%s", name);
-
-    return rb_str_new2(buf);
+    str = rb_str_new(0, strlen(name)+2);
+    sprintf(RSTRING(str)->ptr, ":%s", name);
+    RSTRING(str)->len = strlen(RSTRING(str)->ptr);
+    return str;
 }
 
 static VALUE
@@ -510,33 +503,6 @@ sym_to_s(sym)
     return rb_str_new2(rb_id2name(SYM2ID(sym)));
 }
 
-static VALUE
-rb_mod_clone(module)
-    VALUE module;
-{
-    NEWOBJ(clone, struct RClass);
-    CLONESETUP(clone, module);
-
-    clone->super = RCLASS(module)->super;
-    if (RCLASS(module)->iv_tbl) {
-	clone->iv_tbl = st_copy(RCLASS(module)->iv_tbl);
-    }
-    if (RCLASS(module)->m_tbl) {
-	clone->m_tbl = st_copy(RCLASS(module)->m_tbl);
-    }
-
-    return (VALUE)clone;
-}
-
-static VALUE
-rb_mod_dup(module)
-    VALUE module;
-{
-    VALUE dup = rb_mod_clone(module);
-    OBJSETUP(dup, RBASIC(module)->klass, BUILTIN_TYPE(module));
-    return dup;
-}
-
 static VALUE
 rb_mod_to_s(klass)
     VALUE klass;
@@ -659,11 +625,9 @@ rb_class_s_new(argc, argv)
 	rb_raise(rb_eTypeError, "can't make subclass of virtual class");
     }
     klass = rb_class_new(super);
-    /* make metaclass */
-    RBASIC(klass)->klass = rb_singleton_class_new(RBASIC(super)->klass);
-    rb_singleton_class_attached(RBASIC(klass)->klass, klass);
+    rb_make_metaclass(klass, RBASIC(super)->klass);
     rb_obj_call_init(klass, argc, argv);
-    rb_funcall(super, rb_intern("inherited"), 1, klass);
+    rb_class_inherited(super, klass);
 
     return klass;
 }
@@ -702,7 +666,7 @@ rb_to_id(name)
       case T_FIXNUM:
 	id = FIX2INT(name);
 	if (!rb_id2name(id)) {
-	    rb_raise(rb_eArgError, "%d is not a symbol", id);
+	    rb_raise(rb_eArgError, "%lu is not a symbol", id);
 	}
 	break;
       case T_SYMBOL:
@@ -773,14 +737,24 @@ static VALUE
 rb_mod_const_get(mod, name)
     VALUE mod, name;
 {
-    return rb_const_get(mod, rb_to_id(name));
+    ID id = rb_to_id(name);
+
+    if (!rb_is_const_id(id)) {
+	rb_raise(rb_eNameError, "wrong constant name %s", rb_id2name(id));
+    }
+    return rb_const_get(mod, id);
 }
 
 static VALUE
 rb_mod_const_set(mod, name, value)
     VALUE mod, name, value;
 {
-    rb_const_set(mod, rb_to_id(name), value);
+    ID id = rb_to_id(name);
+
+    if (!rb_is_const_id(id)) {
+	rb_raise(rb_eNameError, "wrong constant name %s", rb_id2name(id));
+    }
+    rb_const_set(mod, id, value);
     return value;
 }
 
@@ -788,7 +762,12 @@ static VALUE
 rb_mod_const_defined(mod, name)
     VALUE mod, name;
 {
-    return rb_const_defined_at(mod, rb_to_id(name));
+    ID id = rb_to_id(name);
+
+    if (!rb_is_const_id(id)) {
+	rb_raise(rb_eNameError, "wrong constant name %s", rb_id2name(id));
+    }
+    return rb_const_defined_at(mod, id);
 }
 
 static VALUE
@@ -823,29 +802,29 @@ rb_obj_private_methods(obj)
     return rb_class_private_instance_methods(1, argv, CLASS_OF(obj));
 }
 
-struct arg_to {
+static VALUE
+convert_type(val, tname, method, raise)
     VALUE val;
-    const char *s;
-};
-
-static VALUE
-to_type(arg)
-    struct arg_to *arg;
+    const char *tname, *method;
+    int raise;
 {
-    return rb_funcall(arg->val, rb_intern(arg->s), 0);
-}
+    ID m;
 
-static VALUE
-fail_to_type(arg)
-    struct arg_to *arg;
-{
-    rb_raise(rb_eTypeError, "failed to convert %s into %s",
-	     NIL_P(arg->val) ? "nil" :
-	     arg->val == Qtrue ? "true" :
-	     arg->val == Qfalse ? "false" :
-	     rb_class2name(CLASS_OF(arg->val)), 
-	     arg->s);
-    return Qnil;		/* not reached */
+    m = rb_intern(method);
+    if (!rb_respond_to(val, m)) {
+	if (raise) {
+	    rb_raise(rb_eTypeError, "failed to convert %s into %s",
+		     NIL_P(val) ? "nil" :
+		     val == Qtrue ? "true" :
+		     val == Qfalse ? "false" :
+		     rb_class2name(CLASS_OF(val)), 
+		     tname);
+	}
+	else {
+	    return Qnil;
+	}
+    }
+    return rb_funcall(val, m, 0);
 }
 
 VALUE
@@ -854,27 +833,59 @@ rb_convert_type(val, type, tname, method)
     int type;
     const char *tname, *method;
 {
-    struct arg_to arg1, arg2;
+    VALUE v;
 
     if (TYPE(val) == type) return val;
-    arg1.val = arg2.val = val;
-    arg1.s = method;
-    arg2.s = tname;
-    val = rb_rescue2(to_type, (VALUE)&arg1, fail_to_type, (VALUE)&arg2,
-		     rb_eStandardError, rb_eNameError, 0);
-    if (TYPE(val) != type) {
+    v = convert_type(val, tname, method, Qtrue);
+    if (TYPE(v) != type) {
 	rb_raise(rb_eTypeError, "%s#%s should return %s",
-		 rb_class2name(CLASS_OF(arg1.val)), method, tname);
+		 rb_class2name(CLASS_OF(val)), method, tname);
     }
-    return val;
+    return v;
+}
+
+VALUE
+rb_check_convert_type(val, type, tname, method)
+    VALUE val;
+    int type;
+    const char *tname, *method;
+{
+    VALUE v;
+
+    if (TYPE(val) == type) return val;
+    v = convert_type(val, tname, method, Qfalse);
+    if (NIL_P(v)) return Qnil;
+    if (TYPE(v) != type) {
+	rb_raise(rb_eTypeError, "%s#%s should return %s",
+		 rb_class2name(CLASS_OF(val)), method, tname);
+    }
+    return v;
+}
+
+static VALUE
+rb_to_integer(val, method)
+    VALUE val;
+    char *method;
+{
+    VALUE v = convert_type(val, "Integer", method, Qtrue);
+    if (!rb_obj_is_kind_of(v, rb_cInteger)) {
+	rb_raise(rb_eTypeError, "%s#%s should return Integer",
+		 rb_class2name(CLASS_OF(val)), method);
+    }
+    return v;
+}
+
+VALUE
+rb_to_int(val)
+    VALUE val;
+{
+    return rb_to_integer(val, "to_int");
 }
 
 VALUE
 rb_Integer(val)
     VALUE val;
 {
-    struct arg_to arg1, arg2;
-
     switch (TYPE(val)) {
       case T_FLOAT:
 	if (RFLOAT(val)->value <= (double)FIXNUM_MAX
@@ -895,17 +906,7 @@ rb_Integer(val)
       default:
 	break;
     }
-
-    arg1.val = arg2.val = val;
-    arg1.s = "to_i";
-    arg2.s = "Integer";
-    val = rb_rescue2(to_type, (VALUE)&arg1, fail_to_type, (VALUE)&arg2,
-		     rb_eStandardError, rb_eNameError, 0);
-    if (!rb_obj_is_kind_of(val, rb_cInteger)) {
-	rb_raise(rb_eTypeError, "%s#to_i should return Integer",
-		 rb_class2name(CLASS_OF(arg1.val)));
-    }
-    return val;
+    return rb_to_integer(val, "to_i");
 }
 
 static VALUE
@@ -929,6 +930,55 @@ rb_Float(val)
       case T_BIGNUM:
 	return rb_float_new(rb_big2dbl(val));
 
+      case T_STRING:
+        {
+	    char *q, *p, *end;
+	    double d;
+
+	    q = p = STR2CSTR(val);
+	    while (*p && ISSPACE(*p)) p++;
+	    d = strtod(p, &end);
+	    if (p == end) {
+	      bad:
+		rb_raise(rb_eArgError, "invalid value for Float: \"%s\"", q);
+	    }
+	    if (*end) {
+		if (*end == '_') {
+		    char *buf = ALLOCA_N(char, strlen(p));
+		    char *n = buf, *last = p;
+
+		    while (p < end) *n++ = *p++;
+		    while (*p) {
+			if (*p == '_' && (n > buf && ISDIGIT(n[-1]))) {
+			    /* remove underscores between digits */
+			    last = ++p;
+			    while (*p == '_') ++p;
+			    if (!ISDIGIT(*p)) {
+				while (last < p) *n++ = *last++;
+			    continue;
+			}
+			    last = p;
+			}
+			*n++ = *p++;
+		    }
+		    while (*last && (*last == '_' || ISSPACE(*last)))
+			last++;
+		    if (!*last) goto bad;
+		    *n = '\0';
+		    p = buf;
+		    d = strtod(p, &end);
+		    if (p == end) goto bad;
+		}
+		while (*end && ISSPACE(*end)) end++;
+		if (*end) goto bad;
+	    }
+	    if (errno == ERANGE) {
+		errno = 0;
+		rb_raise(rb_eArgError, "Float %s out of range", p);
+	    }
+	    return rb_float_new(d);
+	}
+
       case T_NIL:
 	return rb_float_new(0.0);
 
@@ -976,6 +1026,9 @@ rb_str2cstr(str, len)
 	str = rb_str_to_str(str);
     }
     if (len) *len = RSTRING(str)->len;
+    else if (RTEST(ruby_verbose) && RSTRING(str)->len != strlen(RSTRING(str)->ptr)) {
+	rb_warn("string contains \\0 character");
+    }
     return RSTRING(str)->ptr;
 }
 
@@ -1045,12 +1098,9 @@ Init_Object()
     rb_cModule = boot_defclass("Module", rb_cObject);
     rb_cClass =  boot_defclass("Class",  rb_cModule);
 
-    metaclass = RBASIC(rb_cObject)->klass = rb_singleton_class_new(rb_cClass);
-    rb_singleton_class_attached(metaclass, rb_cObject);
-    metaclass = RBASIC(rb_cModule)->klass = rb_singleton_class_new(metaclass);
-    rb_singleton_class_attached(metaclass, rb_cModule);
-    metaclass = RBASIC(rb_cClass)->klass = rb_singleton_class_new(metaclass);
-    rb_singleton_class_attached(metaclass, rb_cClass);
+    metaclass = rb_make_metaclass(rb_cObject, rb_cClass);
+    metaclass = rb_make_metaclass(rb_cModule, metaclass);
+    metaclass = rb_make_metaclass(rb_cClass, metaclass);
 
     rb_mKernel = rb_define_module("Kernel");
     rb_include_module(rb_cObject, rb_mKernel);
@@ -1092,8 +1142,8 @@ Init_Object()
     rb_define_method(rb_mKernel, "hash", rb_obj_id, 0);
     rb_define_method(rb_mKernel, "id", rb_obj_id, 0);
     rb_define_method(rb_mKernel, "__id__", rb_obj_id, 0);
-    rb_define_method(rb_mKernel, "type", rb_obj_type, 0);
-    rb_define_method(rb_mKernel, "class", rb_obj_type, 0);
+    rb_define_method(rb_mKernel, "type", rb_obj_class, 0);
+    rb_define_method(rb_mKernel, "class", rb_obj_class, 0);
 
     rb_define_method(rb_mKernel, "clone", rb_obj_clone, 0);
     rb_define_method(rb_mKernel, "dup", rb_obj_dup, 0);
@@ -1132,7 +1182,6 @@ Init_Object()
     rb_define_global_function("Array", rb_f_array, 1);
 
     rb_cNilClass = rb_define_class("NilClass", rb_cObject);
-    rb_define_method(rb_cNilClass, "type", nil_type, 0);
     rb_define_method(rb_cNilClass, "to_i", nil_to_i, 0);
     rb_define_method(rb_cNilClass, "to_s", nil_to_s, 0);
     rb_define_method(rb_cNilClass, "to_a", nil_to_a, 0);
@@ -1147,8 +1196,8 @@ Init_Object()
 
     rb_cSymbol = rb_define_class("Symbol", rb_cObject);
     rb_undef_method(CLASS_OF(rb_cSymbol), "new");
-    rb_define_method(rb_cSymbol, "type", sym_type, 0);
     rb_define_method(rb_cSymbol, "to_i", sym_to_i, 0);
+    rb_define_method(rb_cSymbol, "to_int", sym_to_i, 0);
     rb_define_method(rb_cSymbol, "inspect", sym_inspect, 0);
     rb_define_method(rb_cSymbol, "to_s", sym_to_s, 0);
     rb_define_method(rb_cSymbol, "id2name", sym_to_s, 0);
@@ -1185,6 +1234,7 @@ Init_Object()
     rb_define_private_method(rb_cModule, "remove_const", rb_mod_remove_const, 1);
     rb_define_private_method(rb_cModule, "method_added", rb_obj_dummy, 1);
     rb_define_method(rb_cModule, "class_variables", rb_mod_class_variables, 0);
+    rb_define_private_method(rb_cModule, "remove_class_variable", rb_mod_remove_cvar, 1);
 
     rb_define_method(rb_cClass, "new", rb_class_new_instance, -1);
     rb_define_method(rb_cClass, "superclass", rb_class_superclass, 0);
@@ -1202,7 +1252,6 @@ Init_Object()
 
     rb_cTrueClass = rb_define_class("TrueClass", rb_cObject);
     rb_define_method(rb_cTrueClass, "to_s", true_to_s, 0);
-    rb_define_method(rb_cTrueClass, "type", true_type, 0);
     rb_define_method(rb_cTrueClass, "&", true_and, 1);
     rb_define_method(rb_cTrueClass, "|", true_or, 1);
     rb_define_method(rb_cTrueClass, "^", true_xor, 1);
@@ -1211,7 +1260,6 @@ Init_Object()
 
     rb_cFalseClass = rb_define_class("FalseClass", rb_cObject);
     rb_define_method(rb_cFalseClass, "to_s", false_to_s, 0);
-    rb_define_method(rb_cFalseClass, "type", false_type, 0);
     rb_define_method(rb_cFalseClass, "&", false_and, 1);
     rb_define_method(rb_cFalseClass, "|", false_or, 1);
     rb_define_method(rb_cFalseClass, "^", false_xor, 1);
