@@ -3,7 +3,7 @@
   pack.c -
 
   $Author: matz $
-  $Date: 2000/07/27 09:49:20 $
+  $Date: 2002/05/01 09:37:58 $
   created at: Thu Feb 10 15:17:05 JST 1994
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -22,12 +22,18 @@
 #endif
 
 #ifdef NATINT_PACK
+# define OFF16B(p) ((char*)(p) + (natint?0:(sizeof(short) - SIZE16)))
+# define OFF32B(p) ((char*)(p) + (natint?0:(sizeof(long) - SIZE32)))
+# define NATINT_I32(x) (natint?NUM2LONG(x):(NUM2I32(x)))
+# define NATINT_U32(x) (natint?NUM2ULONG(x):(NUM2U32(x)))
 # define NATINT_LEN(type,len) (natint?sizeof(type):(len))
 # ifdef WORDS_BIGENDIAN
-#   define OFF16(p) ((char*)(p) + (natint?0:(sizeof(short) - 2)))
-#   define OFF32(p) ((char*)(p) + (natint?0:(sizeof(long) - 4)))
+#   define OFF16(p) OFF16B(p)
+#   define OFF32(p) OFF32B(p)
 # endif
 #else
+# define NATINT_I32(x) NUM2I32(x)
+# define NATINT_U32(x) NUM2U32(x)
 # define NATINT_LEN(type,len) sizeof(type)
 #endif
 
@@ -36,6 +42,11 @@
 # define OFF32(p) (char*)(p)
 #endif
 
+#ifndef OFF16B
+# define OFF16B(p) (char*)(p)
+# define OFF32B(p) (char*)(p)
+#endif
+
 #define define_swapx(x, xtype)		\
 static xtype				\
 TOKEN_PASTE(swap,x)(z)			\
@@ -60,13 +71,13 @@ TOKEN_PASTE(swap,x)(z)			\
 }
 
 #if SIZEOF_SHORT == 2
-#define swaps(x)	((((x)&0xFF)<<8) + (((x)>>8)&0xFF))
+#define swaps(x)	((((x)&0xFF)<<8) | (((x)>>8)&0xFF))
 #else
 #if SIZEOF_SHORT == 4
 #define swaps(x)	((((x)&0xFF)<<24)	\
-			+(((x)>>24)&0xFF)	\
-			+(((x)&0x0000FF00)<<8)	\
-			+(((x)&0x00FF0000)>>8)	)
+			|(((x)>>24)&0xFF)	\
+			|(((x)&0x0000FF00)<<8)	\
+			|(((x)&0x00FF0000)>>8)	)
 #else
 define_swapx(s,short);
 #endif
@@ -74,19 +85,19 @@ define_swapx(s,short);
 
 #if SIZEOF_LONG == 4
 #define swapl(x)	((((x)&0xFF)<<24)	\
-			+(((x)>>24)&0xFF)	\
-			+(((x)&0x0000FF00)<<8)	\
-			+(((x)&0x00FF0000)>>8)	)
+			|(((x)>>24)&0xFF)	\
+			|(((x)&0x0000FF00)<<8)	\
+			|(((x)&0x00FF0000)>>8)	)
 #else
 #if SIZEOF_LONG == 8
 #define swapl(x)        ((((x)&0x00000000000000FF)<<56)	\
-			+(((x)&0xFF00000000000000)>>56)	\
-			+(((x)&0x000000000000FF00)<<40)	\
-			+(((x)&0x00FF000000000000)>>40)	\
-			+(((x)&0x0000000000FF0000)<<24)	\
-			+(((x)&0x0000FF0000000000)>>24)	\
-			+(((x)&0x00000000FF000000)<<8)	\
-			+(((x)&0x000000FF00000000)>>8))
+			|(((x)&0xFF00000000000000)>>56)	\
+			|(((x)&0x000000000000FF00)<<40)	\
+			|(((x)&0x00FF000000000000)>>40)	\
+			|(((x)&0x0000000000FF0000)<<24)	\
+			|(((x)&0x0000FF0000000000)>>24)	\
+			|(((x)&0x00000000FF000000)<<8)	\
+			|(((x)&0x000000FF00000000)>>8))
 #else
 define_swapx(l,long);
 #endif
@@ -293,6 +304,18 @@ endian()
 #define VTOHD(x,y)	vtohd(x)
 #endif
 
+#if SIZEOF_LONG == SIZE32
+typedef long I32;
+typedef unsigned long U32;
+#define NUM2I32(x) NUM2LONG(x)
+#define NUM2U32(x) NUM2ULONG(x)
+#elif SIZEOF_INT == SIZE32
+typedef int I32;
+typedef unsigned int U32;
+#define NUM2I32(x) NUM2INT(x)
+#define NUM2U32(x) NUM2UINT(x)
+#endif
+
 static char *toofew = "too few arguments";
 
 static void encodes _((VALUE,char*,int,int));
@@ -569,7 +592,7 @@ pack_pack(ary, fmt)
 		from = NEXTFROM;
 		if (NIL_P(from)) l = 0;
 		else {
-		    l = NUM2ULONG(from);
+		    l = NATINT_U32(from);
 		}
 		rb_str_cat(res, OFF32(&l), NATINT_LEN(long,4));
 	    }
@@ -585,7 +608,7 @@ pack_pack(ary, fmt)
 		    s = NUM2INT(from);
 		}
 		s = htons(s);
-		rb_str_cat(res, OFF16(&s), NATINT_LEN(short,2));
+		rb_str_cat(res, OFF16B(&s), NATINT_LEN(short,2));
 	    }
 	    break;
 
@@ -596,10 +619,10 @@ pack_pack(ary, fmt)
 		from = NEXTFROM;
 		if (NIL_P(from)) l = 0;
 		else {
-		    l = NUM2ULONG(from);
+		    l = NATINT_U32(from);
 		}
 		l = htonl(l);
-		rb_str_cat(res, OFF32(&l), NATINT_LEN(long,4));
+		rb_str_cat(res, OFF32B(&l), NATINT_LEN(long,4));
 	    }
 	    break;
 
@@ -624,7 +647,7 @@ pack_pack(ary, fmt)
 		from = NEXTFROM;
 		if (NIL_P(from)) l = 0;
 		else {
-		    l = NUM2ULONG(from);
+		    l = NATINT_U32(from);
 		}
 		l = htovl(l);
 		rb_str_cat(res, OFF32(&l), NATINT_LEN(long,4));
@@ -803,7 +826,7 @@ pack_pack(ary, fmt)
 	  case 'm':
 	    ptr = rb_str2cstr(NEXTFROM, &plen);
 
-	    if (len <= 1)
+	    if (len <= 2)
 		len = 45;
 	    else
 		len = len / 3 * 3;
@@ -1003,11 +1026,7 @@ qpencode(str, from, len)
     }
 }
 
-#if defined(__GNUC__) && __GNUC__ >= 2 && !defined(RUBY_NO_INLINE)
-static __inline__ int
-#else
-static int
-#endif
+static inline int
 hex2num(c)
     char c;
 {
@@ -1315,7 +1334,7 @@ pack_unpack(str, fmt)
 	    PACK_LENGTH_ADJUST(unsigned short,2);
 	    while (len-- > 0) {
 		unsigned short tmp = 0;
-		memcpy(OFF16(&tmp), s, NATINT_LEN(unsigned short,2));
+		memcpy(OFF16B(&tmp), s, NATINT_LEN(unsigned short,2));
 		s += NATINT_LEN(unsigned short,2);
 		rb_ary_push(ary, rb_uint2inum(ntohs(tmp)));
 	    }
@@ -1326,7 +1345,7 @@ pack_unpack(str, fmt)
 	    PACK_LENGTH_ADJUST(unsigned long,4);
 	    while (len-- > 0) {
 		unsigned long tmp = 0;
-		memcpy(OFF32(&tmp), s, NATINT_LEN(unsigned long,4));
+		memcpy(OFF32B(&tmp), s, NATINT_LEN(unsigned long,4));
 		s += NATINT_LEN(unsigned long,4);
 		rb_ary_push(ary, rb_uint2inum(ntohl(tmp)));
 	    }
@@ -1437,12 +1456,12 @@ pack_unpack(str, fmt)
 	    
 	  case 'U':
 	    if (len > send - s) len = send - s;
-	    while (len-- > 0 && s < send) {
-		int alen;
+	    while (len > 0 && s < send) {
+		int alen = send - s;
 		unsigned long l;
 
 		l = utf8_to_uv(s, &alen);
-		s += alen;
+		s += alen; len--;
 		rb_ary_push(ary, rb_uint2inum(l));
 	    }
 	    break;
@@ -1738,6 +1757,7 @@ utf8_to_uv(p, lenp)
     else if (c < 0xfc) n = 5;
     else if (c < 0xfe) n = 6;
     else if (c == 0xfe) n = 7;
+    if (n > *lenp) return 0;
     *lenp = n--;
 
     uv = c;
