@@ -3,7 +3,7 @@
   process.c -
 
   $Author: matz $
-  $Date: 2000/08/07 05:01:54 $
+  $Date: 2002/03/12 09:27:26 $
   created at: Tue Aug 10 14:30:50 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -40,9 +40,6 @@ struct timeval rb_time_interval _((VALUE));
 #ifdef HAVE_GETPRIORITY
 # include <sys/resource.h>
 #endif
-#ifdef HAVE_VFORK_H
-#include <vfork.h>
-#endif
 #include "st.h"
 
 #ifdef __EMX__
@@ -86,11 +83,13 @@ rb_waitpid(pid, flags, st)
     }
 
   retry:
+    TRAP_BEG;
 #ifdef HAVE_WAITPID
     result = waitpid(pid, st, flags);
 #else  /* HAVE_WAIT4 */
     result = wait4(pid, st, flags, NULL);
 #endif
+    TRAP_END;
     if (result < 0) {
 	if (errno == EINTR) {
 	    rb_thread_polling();
@@ -116,7 +115,9 @@ rb_waitpid(pid, flags, st)
     }
 
     for (;;) {
+	TRAP_BEG;
 	result = wait(st);
+	TRAP_END;
 	if (result < 0) {
 	    if (errno == EINTR) {
 		rb_thread_schedule();
@@ -170,7 +171,11 @@ proc_wait()
 	return INT2FIX(data.pid);
     }
 
-    while ((pid = wait(&state)) < 0) {
+    while (1) {
+	TRAP_BEG;
+	pid = wait(&state);
+	TRAP_END;
+	if (pid >= 0) break;
         if (errno == EINTR) {
             rb_thread_schedule();
             continue;
@@ -209,7 +214,10 @@ proc_waitpid(argc, argv)
 
     if ((pid = rb_waitpid(NUM2INT(vpid), flags, &status)) < 0)
 	rb_sys_fail(0);
-    if (pid == 0) return Qnil;
+    if (pid == 0) {
+	rb_last_status = Qnil;
+	return Qnil;
+    }
     return INT2FIX(pid);
 }
 
@@ -218,11 +226,14 @@ proc_waitpid2(argc, argv)
     int argc;
     VALUE *argv;
 {
-    VALUE pid = proc_waitpid2(argc, argv);
+    VALUE pid = proc_waitpid(argc, argv);
+    if (NIL_P(pid)) return Qnil;
     return rb_assoc_new(pid, rb_last_status);
 }
 
+#ifndef HAVE_STRING_H
 char *strtok();
+#endif
 
 #ifdef HAVE_SETITIMER
 #define before_exec() rb_thread_stop_timer()
@@ -281,7 +292,7 @@ proc_exec_v(argv, prog)
 
 	    for (n = 0; argv[n]; n++)
 		/* no-op */;
-	    new_argv = ALLOCA_N(char *, n + 2);
+	    new_argv = ALLOCA_N(char*, n + 2);
 	    for (; n > 0; n--)
 		new_argv[n + 1] = argv[n];
 	    new_argv[1] = strcpy(ALLOCA_N(char, strlen(argv[0]) + 1), argv[0]);
@@ -409,7 +420,7 @@ proc_spawn_v(argv, prog)
 
 	for (n = 0; argv[n]; n++)
 	    /* no-op */;
-	new_argv = ALLOCA_N(char *, n + 2);
+	new_argv = ALLOCA_N(char*, n + 2);
 	for (; n > 0; n--)
 	    new_argv[n + 1] = argv[n];
 	new_argv[1] = strcpy(ALLOCA_N(char, strlen(argv[0]) + 1), argv[0]);
@@ -439,13 +450,13 @@ proc_spawn_n(argc, argv, prog)
     char **args;
     int i;
 
-    args = ALLOCA_N(char *, argc + 1);
+    args = ALLOCA_N(char*, argc + 1);
     for (i = 0; i < argc; i++) {
 	Check_SafeStr(argv[i]);
 	args[i] = RSTRING(argv[i])->ptr;
     }
     Check_SafeStr(prog);
-    args[i] = (char *) 0;
+    args[i] = (char*) 0;
     if (args[0])
 	return proc_spawn_v(args, RSTRING(prog)->ptr);
     return -1;
@@ -471,7 +482,7 @@ proc_spawn(sv)
 	    return state;
 	}
     }
-    a = argv = ALLOCA_N(char *, (s - str) / 2 + 2);
+    a = argv = ALLOCA_N(char*, (s - str) / 2 + 2);
     s = ALLOCA_N(char, s - str + 1);
     strcpy(s, str);
     if (*a++ = strtok(s, " \t")) {
@@ -531,9 +542,12 @@ rb_f_fork(obj)
 #ifdef linux
 	after_exec();
 #endif
+	rb_thread_atfork();
 	if (rb_block_given_p()) {
-	    rb_yield(Qnil);
-	    _exit(0);
+	    int status;
+
+	    rb_protect(rb_yield, Qnil, &status);
+	    ruby_stop(status);
 	}
 	return Qnil;
 
@@ -574,29 +588,37 @@ void
 rb_syswait(pid)
     int pid;
 {
+    static int overriding;
     RETSIGTYPE (*hfunc)_((int)), (*qfunc)_((int)), (*ifunc)_((int));
     int status;
-    int i;
+    int i, hooked = Qfalse;
 
+    if (!overriding) {
 #ifdef SIGHUP
-    hfunc = signal(SIGHUP, SIG_IGN);
+	hfunc = signal(SIGHUP, SIG_IGN);
 #endif
 #ifdef SIGQUIT
-    qfunc = signal(SIGQUIT, SIG_IGN);
+	qfunc = signal(SIGQUIT, SIG_IGN);
 #endif
-    ifunc = signal(SIGINT, SIG_IGN);
+	ifunc = signal(SIGINT, SIG_IGN);
+	overriding = Qtrue;
+	hooked = Qtrue;
+    }
 
     do {
 	i = rb_waitpid(pid, 0, &status);
     } while (i == -1 && errno == EINTR);
 
+    if (hooked) {
 #ifdef SIGHUP
-    signal(SIGHUP, hfunc);
+	signal(SIGHUP, hfunc);
 #endif
 #ifdef SIGQUIT
-    signal(SIGQUIT, qfunc);
+	signal(SIGQUIT, qfunc);
 #endif
-    signal(SIGINT, ifunc);
+	signal(SIGINT, ifunc);
+	overriding = Qfalse;
+    }
 }
 
 static VALUE
@@ -649,7 +671,7 @@ rb_f_system(argc, argv)
 
     Check_SafeStr(cmd);
     state = system(RSTRING(cmd)->ptr);
-    rb_last_status = INT2FIX(state);
+    rb_last_status = INT2FIX((state & 0xff) << 8);
 
     if (state == 0) return Qtrue;
     return Qfalse;
@@ -710,7 +732,7 @@ rb_f_system(argc, argv)
 	Check_SafeStr(argv[i]);
     }
   retry:
-    switch (pid = vfork()) {
+    switch (pid = fork()) {
       case 0:
 	if (argc == 1 && prog == 0) {
 	    rb_proc_exec(RSTRING(argv[0])->ptr);
@@ -821,6 +843,7 @@ proc_getpgid(obj, pid)
     int i;
 
     i = getpgid(NUM2INT(pid));
+    if (i < 0) rb_sys_fail(0);
     return INT2NUM(i);
 #else
     rb_notimplement();
@@ -918,6 +941,7 @@ proc_setpriority(obj, which, who, prio)
 #else
     rb_notimplement();
 #endif
+    return INT2FIX(0);
 }
 
 static VALUE
@@ -936,16 +960,18 @@ proc_setuid(obj, id)
 
     uid = NUM2INT(id);
 #ifdef HAVE_SETREUID
-    setreuid(uid, -1);
+    if (setreuid(uid, -1) < 0) rb_sys_fail(0);
 #else
 #ifdef HAVE_SETRUID
-    setruid(uid);
+    if (setruid(uid) < 0) rb_sys_fail(0);
 #else
     {
-	if (geteuid() == uid)
-	    setuid(uid);
-	else
+	if (geteuid() == uid) {
+	    if (setuid(uid) < 0) rb_sys_fail(0);
+	}
+	else {
 	    rb_notimplement();
+	}
     }
 #endif
 #endif
@@ -967,17 +993,19 @@ proc_setgid(obj, id)
     int gid;
 
     gid = NUM2INT(id);
-#ifdef HAS_SETRGID
-	setrgid((GIDTYPE)gid);
-#else
 #ifdef HAVE_SETREGID
-    setregid(gid, -1);
+    if (setregid(gid, -1) < 0) rb_sys_fail(0);
+#else
+#ifdef HAS_SETRGID
+    if (setrgid((GIDTYPE)gid) < 0) rb_sys_fail(0);
 #else
     {
-	if (getegid() == gid)
-	    setgid(gid);
-	else
+	if (getegid() == gid) {
+	    if (setgid(gid) < 0) rb_sys_fail(0);
+	}
+	else {
 	    rb_notimplement();
+	}
     }
 #endif
 #endif
@@ -996,17 +1024,19 @@ static VALUE
 proc_seteuid(obj, euid)
     VALUE obj, euid;
 {
-#ifdef HAVE_SETEUID
-    if (seteuid(NUM2INT(euid)) < 0) rb_sys_fail(0);
-#else
 #ifdef HAVE_SETREUID
     if (setreuid(-1, NUM2INT(euid)) < 0) rb_sys_fail(0);
+#else
+#ifdef HAVE_SETEUID
+    if (seteuid(NUM2INT(euid)) < 0) rb_sys_fail(0);
 #else
     euid = NUM2INT(euid);
-    if (euid == getuid())
-	setuid(euid);
-    else
+    if (euid == getuid()) {
+	if (setuid(euid) < 0) rb_sys_fail(0);
+    }
+    else {
 	rb_notimplement();
+    }
 #endif
 #endif
     return euid;
@@ -1025,17 +1055,19 @@ proc_setegid(obj, egid)
     VALUE obj, egid;
 {
     rb_secure(2);
-#ifdef HAVE_SETEGID
-    if (setegid(NUM2INT(egid)) < 0) rb_sys_fail(0);
-#else
 #ifdef HAVE_SETREGID
     if (setregid(-1, NUM2INT(egid)) < 0) rb_sys_fail(0);
+#else
+#ifdef HAVE_SETEGID
+    if (setegid(NUM2INT(egid)) < 0) rb_sys_fail(0);
 #else
     egid = NUM2INT(egid);
-    if (egid == getgid())
-	setgid(egid);
-    else
+    if (egid == getgid()) {
+	if (setgid(egid) < 0) rb_sys_fail(0);
+    }
+    else {
 	rb_notimplement();
+    }
 #endif
 #endif
     return egid;
@@ -1077,10 +1109,10 @@ Init_process()
     rb_define_module_function(rb_mProcess, "wait2", proc_wait2, 0);
     rb_define_module_function(rb_mProcess, "waitpid", proc_waitpid, -1);
     rb_define_module_function(rb_mProcess, "waitpid2", proc_waitpid2, -1);
+#endif /* ifndef NT */
 
     rb_define_module_function(rb_mProcess, "pid", get_pid, 0);
     rb_define_module_function(rb_mProcess, "ppid", get_ppid, 0);
-#endif /* ifndef NT */
 
     rb_define_module_function(rb_mProcess, "getpgrp", proc_getpgrp, -1);
     rb_define_module_function(rb_mProcess, "setpgrp", proc_setpgrp, -1);
