@@ -2,8 +2,8 @@
 
   range.c -
 
-  $Author: matz $
-  $Date: 2000/08/31 05:29:48 $
+  $Author: knu $
+  $Date: 2002/10/26 15:57:24 $
   created at: Thu Aug 19 17:46:47 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -13,10 +13,10 @@
 #include "ruby.h"
 
 VALUE rb_cRange;
-static ID id_cmp, id_beg, id_end;
+static ID id_cmp, id_beg, id_end, id_excl;
 
-#define EXCL(r) FL_TEST((r), FL_USER1)
-#define SET_EXCL(r) FL_SET((r), FL_USER1)
+#define EXCL(r) RTEST(rb_ivar_get((r), id_excl))
+#define SET_EXCL(r,v) rb_ivar_set((r), id_excl, (v) ? Qtrue : Qfalse)
 
 static VALUE
 range_check(args)
@@ -46,10 +46,7 @@ range_init(obj, beg, end, exclude_end)
 		   rb_eStandardError, rb_eNameError, 0);
     }
 
-    if (exclude_end) {
-	SET_EXCL(obj);
-    }
-
+    SET_EXCL(obj, exclude_end);
     rb_ivar_set(obj, id_beg, beg);
     rb_ivar_set(obj, id_end, end);
 }
@@ -86,7 +83,23 @@ static VALUE
 range_exclude_end_p(range)
     VALUE range;
 {
-    return EXCL(range)?Qtrue:Qfalse;
+    return EXCL(range) ? Qtrue : Qfalse;
+}
+
+static VALUE
+range_eq(range, obj)
+    VALUE range, obj;
+{
+    if (!rb_obj_is_kind_of(obj, rb_obj_class(range))) return Qfalse;
+
+    if (!rb_equal(rb_ivar_get(range, id_beg), rb_ivar_get(obj, id_beg)))
+	return Qfalse;
+    if (!rb_equal(rb_ivar_get(range, id_end), rb_ivar_get(obj, id_end)))
+	return Qfalse;
+
+    if (EXCL(range) != EXCL(obj)) return Qfalse;
+
+    return Qtrue;
 }
 
 static VALUE
@@ -122,6 +135,38 @@ range_eqq(range, obj)
     return Qfalse;
 }
 
+static VALUE
+range_eql(range, obj)
+    VALUE range, obj;
+{
+    if (range == obj) return Qtrue;
+    if (!rb_obj_is_kind_of(obj, rb_obj_class(range))) return Qfalse;
+
+    if (!rb_eql(rb_ivar_get(range, id_beg), rb_ivar_get(obj, id_beg)))
+	return Qfalse;
+    if (!rb_eql(rb_ivar_get(range, id_end), rb_ivar_get(obj, id_end)))
+	return Qfalse;
+
+    if (EXCL(range) != EXCL(obj)) return Qfalse;
+
+    return Qtrue;
+}
+
+static VALUE
+range_hash(range, obj)
+    VALUE range, obj;
+{
+    long hash = EXCL(range);
+    VALUE v;
+
+    v = rb_hash(rb_ivar_get(range, id_beg));
+    hash ^= v << 1;
+    v = rb_hash(rb_ivar_get(range, id_end));
+    hash ^= v << 9;
+
+    return INT2FIX(hash);
+}
+
 static VALUE
 range_each(range)
     VALUE range;
@@ -243,8 +288,8 @@ rb_range_beg_len(range, begp, lenp, len, err)
 
   out_of_range:
     if (err) {
-	rb_raise(rb_eRangeError, "%d..%s%d out of range",
-		 b, EXCL(range)?".":"", e);
+	rb_raise(rb_eRangeError, "%ld..%s%ld out of range",
+		 b, EXCL(range) ? "." : "", e);
     }
     return Qnil;
 }
@@ -258,7 +303,7 @@ range_to_s(range)
     str = rb_obj_as_string(rb_ivar_get(range, id_beg));
     str2 = rb_obj_as_string(rb_ivar_get(range, id_end));
     str = rb_str_dup(str);
-    rb_str_cat(str, "...", EXCL(range)?3:2);
+    rb_str_cat(str, "...", EXCL(range) ? 3 : 2);
     rb_str_append(str, str2);
     OBJ_INFECT(str, str2);
 
@@ -274,7 +319,7 @@ range_inspect(range)
     str = rb_inspect(rb_ivar_get(range, id_beg));
     str2 = rb_inspect(rb_ivar_get(range, id_end));
     str = rb_str_dup(str);
-    rb_str_cat(str, "...", EXCL(range)?3:2);
+    rb_str_cat(str, "...", EXCL(range) ? 3 : 2);
     rb_str_append(str, str2);
     OBJ_INFECT(str, str2);
 
@@ -338,7 +383,10 @@ Init_Range()
     rb_cRange = rb_define_class("Range", rb_cObject);
     rb_include_module(rb_cRange, rb_mEnumerable);
     rb_define_method(rb_cRange, "initialize", range_initialize, -1);
+    rb_define_method(rb_cRange, "==", range_eq, 1);
     rb_define_method(rb_cRange, "===", range_eqq, 1);
+    rb_define_method(rb_cRange, "eql?", range_eql, 1);
+    rb_define_method(rb_cRange, "hash", range_hash, 0);
     rb_define_method(rb_cRange, "each", range_each, 0);
     rb_define_method(rb_cRange, "first", range_first, 0);
     rb_define_method(rb_cRange, "last", range_last, 0);
@@ -355,4 +403,5 @@ Init_Range()
     id_cmp = rb_intern("<=>");
     id_beg = rb_intern("begin");
     id_end = rb_intern("end");
+    id_excl = rb_intern("excl");
 }
