@@ -2,7 +2,7 @@
 
   re.c -
 
-  $Author: matz $
+  $Author: knu $
   created at: Mon Aug  9 18:24:49 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -73,7 +73,7 @@ static const char casetable[] = {
 int
 rb_memcmp(p1, p2, len)
     char *p1, *p2;
-    size_t len;
+    long len;
 {
     int tmp;
 
@@ -214,13 +214,13 @@ rb_reg_expr_str(str, s, len)
     const char *p, *pend;
     int need_escape = 0;
 
-    p = s; pend = p + len;
+    p = s; pend = p + len; 
     while (p<pend) {
 	if (*p == '/' || (!ISPRINT(*p) && !ismbchar(*p))) {
 	    need_escape = 1;
 	    break;
 	}
-	p++;
+	p += mbclen(*p);
     }
     if (!need_escape) {
 	rb_str_cat(str, s, len);
@@ -228,7 +228,14 @@ rb_reg_expr_str(str, s, len)
     else {
 	p = s; 
 	while (p<pend) {
-	    if (*p == '/') {
+	    if (*p == '\\') {
+		rb_str_cat(str, p, 1);
+		p++;
+	    	rb_str_cat(str, p, mbclen(*p));
+		p += mbclen(*p);
+		continue;
+	    }
+	    else if (*p == '/') {
 		char c = '\\';
 		rb_str_cat(str, &c, 1);
 		rb_str_cat(str, p, 1);
@@ -241,35 +248,14 @@ rb_reg_expr_str(str, s, len)
 	    else if (ISPRINT(*p)) {
 		rb_str_cat(str, p, 1);
 	    }
-	    else {
+	    else if (!ISSPACE(*p)) {
 		char b[8];
-		switch (*p) {
-		case '\r':
-		    rb_str_cat(str, "\\r", 2);
-		    break;
-		case '\n':
-		    rb_str_cat(str, "\\n", 2);
-		    break;
-		case '\t':
-		    rb_str_cat(str, "\\t", 2);
-		    break;
-		case '\f':
-		    rb_str_cat(str, "\\f", 2);
-		    break;
-		case 007:
-		    rb_str_cat(str, "\\a", 2);
-		    break;
-		case 013:
-		    rb_str_cat(str, "\\v", 2);
-		    break;
-		case 033:
-		    rb_str_cat(str, "\\e", 2);
-		    break;
-		default:
-		    sprintf(b, "\\%03o", *p & 0377);
-		    rb_str_cat(str, b, 4);
-		    break;
-		}
+
+		sprintf(b, "\\%03o", *p & 0377);
+		rb_str_cat(str, b, 4);
+	    }
+	    else {
+		rb_str_cat(str, p, 1);
 	    }
 	    p++;
 	}
@@ -288,11 +274,11 @@ rb_reg_desc(s, len, re)
     rb_str_cat2(str, "/");
     if (re) {
 	rb_reg_check(re);
-	if (RREGEXP(re)->ptr->options & RE_OPTION_MULTILINE)
-	    rb_str_cat2(str, "m");
 	/* /p is obsolete; to be removed */
 	if ((RREGEXP(re)->ptr->options & RE_OPTION_POSIXLINE) == RE_OPTION_POSIXLINE)
 	    rb_str_cat2(str, "p");
+	else if (RREGEXP(re)->ptr->options & RE_OPTION_MULTILINE)
+	    rb_str_cat2(str, "m");
 	if (RREGEXP(re)->ptr->options & RE_OPTION_IGNORECASE)
 	    rb_str_cat2(str, "i");
 	if (RREGEXP(re)->ptr->options & RE_OPTION_EXTENDED)
@@ -595,7 +581,10 @@ rb_reg_search(re, str, pos, reverse)
     static struct re_registers regs;
     int range;
 
-    if (pos > RSTRING(str)->len) return -1;
+    if (pos > RSTRING(str)->len) {
+	rb_backref_set(Qnil);
+	return -1;
+    }
 
     rb_reg_check(re);
     if (may_need_recompile) rb_reg_prepare_re(re);
@@ -636,6 +625,13 @@ rb_reg_search(re, str, pos, reverse)
     if (NIL_P(match) || FL_TEST(match, MATCH_BUSY)) {
 	match = match_alloc();
     }
+    else {
+	if (rb_safe_level() >= 3) 
+	    OBJ_TAINT(match);
+	else
+	    FL_UNSET(match, FL_TAINT);
+    }
+
     re_copy_registers(RMATCH(match)->regs, &regs);
     RMATCH(match)->str = rb_str_new4(str);
     rb_backref_set(match);
@@ -652,7 +648,11 @@ rb_reg_nth_defined(nth, match)
 {
     if (NIL_P(match)) return Qnil;
     if (nth >= RMATCH(match)->regs->num_regs) {
-	return Qfalse;
+	return Qnil;
+    }
+    if (nth < 0) {
+	nth += RMATCH(match)->regs->num_regs;
+	if (nth <= 0) return Qnil;
     }
     if (RMATCH(match)->BEG(nth) == -1) return Qfalse;
     return Qtrue;
@@ -670,6 +670,10 @@ rb_reg_nth_match(nth, match)
     if (nth >= RMATCH(match)->regs->num_regs) {
 	return Qnil;
     }
+    if (nth < 0) {
+	nth += RMATCH(match)->regs->num_regs;
+	if (nth <= 0) return Qnil;
+    }
     start = RMATCH(match)->BEG(nth);
     if (start == -1) return Qnil;
     end = RMATCH(match)->END(nth);
@@ -690,19 +694,27 @@ VALUE
 rb_reg_match_pre(match)
     VALUE match;
 {
+    VALUE str;
+
     if (NIL_P(match)) return Qnil;
     if (RMATCH(match)->BEG(0) == -1) return Qnil;
-    return rb_str_new(RSTRING(RMATCH(match)->str)->ptr, RMATCH(match)->BEG(0));
+    str = rb_str_new(RSTRING(RMATCH(match)->str)->ptr, RMATCH(match)->BEG(0));
+    if (OBJ_TAINTED(match)) OBJ_TAINT(str);
+    return str;
 }
 
 VALUE
 rb_reg_match_post(match)
     VALUE match;
 {
+    VALUE str;
+
     if (NIL_P(match)) return Qnil;
     if (RMATCH(match)->BEG(0) == -1) return Qnil;
-    return rb_str_new(RSTRING(RMATCH(match)->str)->ptr+RMATCH(match)->END(0),
-		      RSTRING(RMATCH(match)->str)->len-RMATCH(match)->END(0));
+    str = rb_str_new(RSTRING(RMATCH(match)->str)->ptr+RMATCH(match)->END(0),
+		     RSTRING(RMATCH(match)->str)->len-RMATCH(match)->END(0));
+    if (OBJ_TAINTED(match)) OBJ_TAINT(str);
+    return str;
 }
 
 VALUE
@@ -752,11 +764,16 @@ match_to_a(match)
     VALUE ary = rb_ary_new2(regs->num_regs);
     char *ptr = RSTRING(RMATCH(match)->str)->ptr;
     int i;
-
+    int taint = OBJ_TAINTED(match);
+    
     for (i=0; i<regs->num_regs; i++) {
-	if (regs->beg[i] == -1) rb_ary_push(ary, Qnil);
-	else rb_ary_push(ary, rb_str_new(ptr+regs->beg[i],
-				   regs->end[i]-regs->beg[i]));
+	if (regs->beg[i] == -1) {
+	    rb_ary_push(ary, Qnil);
+	} else {
+	    VALUE str = rb_str_new(ptr+regs->beg[i], regs->end[i]-regs->beg[i]);
+	    if (taint) OBJ_TAINT(str);
+	    rb_ary_push(ary, str);
+	}
     }
     return ary;
 }
@@ -768,7 +785,6 @@ match_aref(argc, argv, match)
     VALUE match;
 {
     VALUE idx, rest;
-    struct re_registers *regs;
     char *ptr;
     int i;
 
@@ -905,18 +921,13 @@ static VALUE
 rb_reg_equal(re1, re2)
     VALUE re1, re2;
 {
-    int min;
-
     if (re1 == re2) return Qtrue;
     if (TYPE(re2) != T_REGEXP) return Qfalse;
     rb_reg_check(re1); rb_reg_check(re2);
     if (RREGEXP(re1)->len != RREGEXP(re2)->len) return Qfalse;
-    min = RREGEXP(re1)->len;
-    if (min > RREGEXP(re2)->len) min = RREGEXP(re2)->len;
-    if (memcmp(RREGEXP(re1)->str, RREGEXP(re2)->str, min) == 0 &&
+    if (memcmp(RREGEXP(re1)->str, RREGEXP(re2)->str, RREGEXP(re1)->len) == 0 &&
 	rb_reg_cur_kcode(re1) == rb_reg_cur_kcode(re2) &&
-	!((RREGEXP(re1)->ptr->options & RE_OPTION_IGNORECASE) ^
-	  (RREGEXP(re2)->ptr->options & RE_OPTION_IGNORECASE))) {
+	RREGEXP(re1)->ptr->options == RREGEXP(re2)->ptr->options) {
 	return Qtrue;
     }
     return Qfalse;
@@ -928,7 +939,10 @@ rb_reg_match(re, str)
 {
     int start;
 
-    if (NIL_P(str)) return Qnil;
+    if (NIL_P(str)) {
+	rb_backref_set(Qnil);
+	return Qnil;
+    }
     str = rb_str_to_str(str);
     start = rb_reg_search(re, str, 0, 0);
     if (start < 0) {
@@ -944,8 +958,10 @@ rb_reg_match2(re)
     int start;
     VALUE line = rb_lastline_get();
 
-    if (TYPE(line) != T_STRING)
+    if (TYPE(line) != T_STRING) {
+	rb_backref_set(Qnil);
 	return Qnil;
+    }
 
     start = rb_reg_search(re, line, 0, 0);
     if (start < 0) {
@@ -1003,6 +1019,9 @@ rb_reg_initialize_m(argc, argv, self)
 	}
     }
 
+    if (OBJ_FROZEN(self)) {
+	rb_error_frozen("Regexp");
+    }
     src = argv[0];
     if (TYPE(src) == T_REGEXP) {
 	rb_reg_check(src);
@@ -1015,6 +1034,7 @@ rb_reg_initialize_m(argc, argv, self)
 	p = rb_str2cstr(src, &len);
 	rb_reg_initialize(self, p, len, flag);
     }
+    return self;
 }
 
 static VALUE
@@ -1038,8 +1058,8 @@ rb_reg_s_quote(argc, argv)
     VALUE str, kcode;
     int kcode_saved = reg_kcode;
     char *s, *send, *t;
-    char *tmp;
-    int len;
+    VALUE tmp;
+    int len, c;
 
     rb_scan_args(argc, argv, "11", &str, &kcode);
     if (!NIL_P(kcode)) {
@@ -1049,32 +1069,50 @@ rb_reg_s_quote(argc, argv)
     }
     s = rb_str2cstr(str, &len);
     send = s + len;
-    tmp = ALLOCA_N(char, len*2);
-    t = tmp;
+    tmp = rb_str_new(0, len*2);
+    t = RSTRING(tmp)->ptr;
 
     for (; s < send; s++) {
-	if (ismbchar(*s)) {
-	    size_t n = mbclen(*s);
+	c = *s;
+	if (ismbchar(c)) {
+	    int n = mbclen(c);
 
 	    while (n-- && s < send)
 		*t++ = *s++;
 	    s--;
 	    continue;
 	}
-	if (*s == '[' || *s == ']'
-	    || *s == '{' || *s == '}'
-	    || *s == '(' || *s == ')'
-	    || *s == '|' || *s == '-'
-	    || *s == '*' || *s == '.' || *s == '\\'
-	    || *s == '?' || *s == '+'
-	    || *s == '^' || *s == '$') {
+	switch (c) {
+	  case '\t':
+	    c = 't';
 	    *t++ = '\\';
+	    break;
+	  case '\f':
+	    c = 'f';
+	    *t++ = '\\';
+	    break;
+	  case '\r':
+	    c = 'r';
+	    *t++ = '\\';
+	    break;
+	  case '\n':
+	    c = 'n';
+	    *t++ = '\\';
+	    break;
+	  case '[': case ']': case '{': case '}':
+	  case '(': case ')': case '|': case '-':
+	  case '*': case '.': case '\\':
+	  case '?': case '+': case '^': case '$':
+	  case ' ': case '#':
+	    *t++ = '\\';
+	    break;
 	}
-	*t++ = *s;
+	*t++ = c;
     }
     kcode_reset_option();
-
-    return rb_str_new(tmp, t - tmp);
+    rb_str_resize(tmp, t - RSTRING(tmp)->ptr);
+    OBJ_INFECT(tmp, str);
+    return tmp;
 }
 
 int
@@ -1122,6 +1160,12 @@ rb_reg_options(re)
     rb_reg_check(re);
     if (RREGEXP(re)->ptr->options & RE_OPTION_IGNORECASE)
 	options |= RE_OPTION_IGNORECASE;
+    if ((RREGEXP(re)->ptr->options & RE_OPTION_POSIXLINE) == RE_OPTION_POSIXLINE)
+	options |= RE_OPTION_POSIXLINE;
+    else if (RREGEXP(re)->ptr->options & RE_OPTION_MULTILINE)
+	options |= RE_OPTION_MULTILINE;
+    if (RREGEXP(re)->ptr->options & RE_OPTION_EXTENDED)
+	options |= RE_OPTION_EXTENDED;
     if (FL_TEST(re, KCODE_FIXED)) {
 	options |= rb_reg_get_kcode(re);
     }
@@ -1138,7 +1182,7 @@ rb_reg_clone(re)
     clone->ptr = 0; clone->len = 0; clone->str = 0;
     rb_reg_initialize(clone, RREGEXP(re)->str, RREGEXP(re)->len,
 		      rb_reg_options(re));
-    return (VALUE)re;
+    return (VALUE)clone;
 }
 
 VALUE
@@ -1307,7 +1351,9 @@ static void
 match_setter(val)
     VALUE val;
 {
-    Check_Type(val, T_MATCH);
+    if (!NIL_P(val)) {
+	Check_Type(val, T_MATCH);
+    }
     rb_backref_set(val);
 }
 
