@@ -2,17 +2,17 @@
 
   ruby.c -
 
-  $Author: matz $
-  $Date: 2000/09/21 06:37:16 $
+  $Author: nobu $
+  $Date: 2002/09/05 04:12:40 $
   created at: Tue Aug 10 12:47:31 JST 1993
 
-  Copyright (C) 1993-2000 Yukihiro Matsumoto
+  Copyright (C) 1993-2001 Yukihiro Matsumoto
   Copyright (C) 2000  Network Applied Communication Laboratory, Inc.
   Copyright (C) 2000  Information-technology Promotion Agency, Japan
 
 **********************************************************************/
 
-#ifdef _WIN32
+#if defined _WIN32 || defined __CYGWIN__
 #include <windows.h>
 #endif
 #include "ruby.h"
@@ -106,7 +106,7 @@ extern VALUE rb_load_path;
 
 #define STATIC_FILE_LENGTH 255
 
-#if defined(_WIN32) || defined(DJGPP)
+#if defined _WIN32 || defined __CYGWIN__ || defined __DJGPP__
 static char *
 rubylib_mangle(s, l)
     char *s;
@@ -172,7 +172,7 @@ ruby_incpush(path)
     const char sep = PATH_SEP_CHAR;
 
     if (path == 0) return;
-#if defined(__CYGWIN32__)
+#if defined(__CYGWIN__)
     {
 	char rubylib[FILENAME_MAX];
 	conv_to_posix_path(path, rubylib, FILENAME_MAX);
@@ -202,23 +202,44 @@ ruby_incpush(path)
     }
 }
 
+#if defined _WIN32 || defined __CYGWIN__ || defined __DJGPP__ || defined __EMX__
+#define LOAD_RELATIVE 1
+#endif
+
 void
 ruby_init_loadpath()
 {
-#if defined(_WIN32) || defined(DJGPP)
+#if defined LOAD_RELATIVE
     char libpath[FILENAME_MAX+1];
     char *p;
-    size_t rest;
-#if defined(_WIN32)
-    GetModuleFileName(NULL, libpath, sizeof libpath);
+    int rest;
+#if defined _WIN32 || defined __CYGWIN__
+# if defined LIBRUBY_SO
+    HMODULE libruby = GetModuleHandle(LIBRUBY_SO);
+# else
+    HMODULE libruby = NULL;
+# endif
+    GetModuleFileName(libruby, libpath, sizeof libpath);
 #elif defined(DJGPP)
     extern char *__dos_argv0;
     strncpy(libpath, __dos_argv0, FILENAME_MAX);
+#define CharNext(p) ((p) + mblen(p, MB_CUR_MAX))
+#elif defined(__EMX__)
+    _execname(libpath, FILENAME_MAX);
 #endif
-    p = strrchr(libpath, '\\');
+
+#ifndef CharNext		/* defined as CharNext[AW] on Windows. */
+#define CharNext(p) ((p) + 1)
+#endif
+
+    for (p = libpath; *p; p = CharNext(p))
+	if (*p == '\\')
+	    *p = '/';
+
+    p = strrchr(libpath, '/');
     if (p) {
 	*p = 0;
-	if (p-libpath > 3 && !strcasecmp(p-4, "\\bin")) {
+	if (p-libpath > 3 && !strcasecmp(p-4, "/bin")) {
 	    p -= 4;
 	    *p = 0;
 	}
@@ -227,14 +248,6 @@ ruby_init_loadpath()
 	p = libpath + 1;
     }
 
-#if !defined(__CYGWIN32__)
-#ifndef CharNext		/* defined as CharNext[AW] on Windows. */
-#define CharNext(p) ((p) + 1)
-#endif
-    for (p = libpath; *p; p = CharNext(p))
-	if (*p == '\\')
-	    *p = '/';
-#endif
     rest = FILENAME_MAX - (p - libpath);
 
 #define RUBY_RELATIVE(path) (strncpy(p, (path), rest), libpath)
@@ -250,20 +263,19 @@ ruby_init_loadpath()
     ruby_incpush(RUBY_RELATIVE(RUBY_SEARCH_PATH));
 #endif
 
+    ruby_incpush(RUBY_RELATIVE(RUBY_SITE_LIB2));
 #ifdef RUBY_SITE_THIN_ARCHLIB
     ruby_incpush(RUBY_RELATIVE(RUBY_SITE_THIN_ARCHLIB));
 #endif
     ruby_incpush(RUBY_RELATIVE(RUBY_SITE_ARCHLIB));
-    ruby_incpush(RUBY_RELATIVE(RUBY_SITE_LIB2));
     ruby_incpush(RUBY_RELATIVE(RUBY_SITE_LIB));
 
+    ruby_incpush(RUBY_RELATIVE(RUBY_LIB));
 #ifdef RUBY_THIN_ARCHLIB
     ruby_incpush(RUBY_RELATIVE(RUBY_THIN_ARCHLIB));
 #endif
     ruby_incpush(RUBY_RELATIVE(RUBY_ARCHLIB));
 
-    ruby_incpush(RUBY_RELATIVE(RUBY_LIB));
-
     if (rb_safe_level() == 0) {
 	ruby_incpush(".");
     }
@@ -272,8 +284,8 @@ ruby_init_loadpath()
 struct req_list {
     char *name;
     struct req_list *next;
-} req_list_head;
-struct req_list *req_list_last = &req_list_head;
+};
+static struct req_list req_list_head, *req_list_last = &req_list_head;
 
 static void
 add_modules(mod)
@@ -291,7 +303,7 @@ add_modules(mod)
 
 extern void Init_ext _((void));
 
-void
+static void
 require_libraries()
 {
     extern NODE *ruby_eval_tree;
@@ -302,7 +314,6 @@ require_libraries()
     struct req_list *tmp;
 
     Init_ext();		/* should be called here for some reason :-( */
-    ruby_sourcefile = 0;
     save[0] = ruby_eval_tree;
     save[1] = ruby_eval_tree_begin;
     ruby_eval_tree = ruby_eval_tree_begin = 0;
@@ -426,6 +437,10 @@ proc_options(argc, argv)
 	    goto reswitch;
 
 	  case 'v':
+	    if (argv0 == 0 || verbose) {
+		s++;
+		goto reswitch;
+	    }
 	    ruby_show_version();
 	    verbose = 1;
 	  case 'w':
@@ -539,10 +554,11 @@ proc_options(argc, argv)
 		if (*++s) {
 		    v = scan_oct(s, 2, &numlen);
 		    if (numlen == 0) v = 1;
+		    s += numlen;
 		}
 		rb_set_safe_level(v);
 	    }
-	    break;
+	    goto reswitch;
 
 	  case 'I':
 	    forbid_setid("-I");
@@ -574,15 +590,17 @@ proc_options(argc, argv)
 	    goto reswitch;
 
 	  case '-':
-	    if (!s[1]) {
+	    if (!s[1] || s[1] == '\r' && !s[2]) {
 		argc--,argv++;
 		goto switch_end;
 	    }
 	    s++;
 	    if (strcmp("copyright", s) == 0)
 		copyright = 1;
-	    else if (strcmp("debug", s) == 0)
-		ruby_debug = 1;
+	    else if (strcmp("debug", s) == 0) {
+		ruby_debug = Qtrue;
+                ruby_verbose = Qtrue;
+            }
 	    else if (strcmp("version", s) == 0)
 		version = 1;
 	    else if (strcmp("verbose", s) == 0) {
@@ -602,11 +620,6 @@ proc_options(argc, argv)
 	    }
 	    break;
 
-	  case '*':
-	  case ' ':
-	    if (s[1] == '-') s+=2;
-	    break;
-
 	  default:
 	    fprintf(stderr, "%s: invalid option -%c  (-h will show valid options)\n",
 		    origargv[0], *s);
@@ -648,11 +661,6 @@ proc_options(argc, argv)
 	}
     }
 
-    if (e_script) {
-	argc++, argv--;
-	argv[0] = script;
-    }
-
     if (version) {
 	ruby_show_version();
 	exit(0);
@@ -661,12 +669,19 @@ proc_options(argc, argv)
 	ruby_show_copyright();
     }
 
+    if (rb_safe_level() >= 4) {
+      OBJ_TAINT(rb_argv);
+      OBJ_TAINT(rb_load_path);
+    }
+
     if (!e_script && argc == 0) { /* no more args */
 	if (verbose) exit(0);
 	script = "-";
     }
     else {
-	script = argv[0];
+	if (!e_script) {
+	    script = argv[0];
+	}
 	if (script[0] == '\0') {
 	    script = "-";
 	}
@@ -684,7 +699,9 @@ proc_options(argc, argv)
 		if (!script) script = argv[0];
 	    }
 	}
-	argc--; argv++;
+	if (!e_script) {
+	    argc--; argv++;
+	}
     }
 
     ruby_script(script);
@@ -692,7 +709,7 @@ proc_options(argc, argv)
     process_sflag();
 
     ruby_init_loadpath();
-    ruby_sourcefile = argv0;
+    ruby_sourcefile = rb_source_filename(argv0);
     if (e_script) {
 	require_libraries();
 	rb_compile_string(script, e_script, 1);
@@ -706,6 +723,11 @@ proc_options(argc, argv)
 
     process_sflag();
     xflag = 0;
+
+    if (rb_safe_level() >= 4) {
+      FL_UNSET(rb_argv, FL_TAINT);
+      FL_UNSET(rb_load_path, FL_TAINT);
+    }
 }
 
 extern int ruby__end__seen;
@@ -726,7 +748,7 @@ load_file(fname, script)
 	FILE *fp = fopen(fname, "r");
 
 	if (fp == NULL) {
-	    rb_raise(rb_eLoadError, "No such file to load -- %s", fname);
+	    rb_load_fail(fname);
 	}
 	fclose(fp);
 
@@ -741,7 +763,7 @@ load_file(fname, script)
     }
 
     if (script) {
-	VALUE c;
+	VALUE c = 1;		/* something not nil */
 	VALUE line;
 	char *p;
 
@@ -764,6 +786,7 @@ load_file(fname, script)
 	c = rb_io_getc(f);
 	if (c == INT2FIX('#')) {
 	    line = rb_io_gets(f);
+	    if (NIL_P(line)) return;
 	    line_start++;
 
 	    if (RSTRING(line)->len > 2 && RSTRING(line)->ptr[0] == '!') {
@@ -794,7 +817,7 @@ load_file(fname, script)
 		    argv[0] = path;
 		    execv(path, argv);
 
-		    ruby_sourcefile = fname;
+		    ruby_sourcefile = rb_source_filename(fname);
 		    ruby_sourceline = 1;
 		    rb_fatal("Can't exec %s", path);
 		}
@@ -825,6 +848,10 @@ load_file(fname, script)
     else if (f != rb_stdin) {
 	rb_io_close(f);
     }
+
+    if (ruby_parser_stack_on_heap()) {
+        rb_gc();
+    }
 }
 
 void
@@ -855,33 +882,8 @@ set_arg0(val, id)
     static int len;
 
     if (origargv == 0) rb_raise(rb_eRuntimeError, "$0 not initialized");
-#ifndef __hpux
-    if (len == 0) {
-	s = origargv[0];
-	s += strlen(s);
-	/* See if all the arguments are contiguous in memory */
-	for (i = 1; i < origargc; i++) {
-	    if (origargv[i] == s + 1)
-		s += strlen(++s);	/* this one is ok too */
-	}
-	len = s - origargv[0];
-    }
-#endif
     s = rb_str2cstr(val, &i);
-#ifndef __hpux
-    if (i > len) {
-	memcpy(origargv[0], s, len);
-	origargv[0][len] = '\0';
-    }
-    else {
-	memcpy(origargv[0], s, i);
-	s = origargv[0]+i;
-	*s++ = '\0';
-	while (++i < len)
-	    *s++ = ' ';
-    }
-    rb_progname = rb_tainted_str_new2(origargv[0]);
-#else
+#ifdef __hpux
     if (i >= PST_CLEN) {
       union pstun j;
       j.pst_command = s;
@@ -895,6 +897,37 @@ set_arg0(val, id)
       pstat(PSTAT_SETCMD, j, i, 0, 0);
     }
     rb_progname = rb_tainted_str_new(s, i);
+#elif defined(HAVE_SETPROCTITLE)
+    setproctitle("%.*s", i, s);
+    rb_progname = rb_tainted_str_new(s, i);
+#else
+    if (len == 0) {
+	char *s = origargv[0];
+	int i;
+
+	s += strlen(s);
+	/* See if all the arguments are contiguous in memory */
+	for (i = 1; i < origargc; i++) {
+	    if (origargv[i] == s + 1)
+		s += strlen(++s);	/* this one is ok too */
+	}
+	len = s - origargv[0];
+    }
+    if (i >= len) {
+	i = len;
+	memcpy(origargv[0], s, i);
+	origargv[0][i] = '\0';
+    }
+    else {
+	memcpy(origargv[0], s, i);
+	s = origargv[0]+i;
+	*s++ = '\0';
+	while (++i < len)
+	    *s++ = ' ';
+	for (i = 1; i < origargc; i++)
+	    origargv[i] = 0;
+    }
+    rb_progname = rb_tainted_str_new2(origargv[0]);
 #endif
 }
 
@@ -904,7 +937,7 @@ ruby_script(name)
 {
     if (name) {
 	rb_progname = rb_tainted_str_new2(name);
-	ruby_sourcefile = name;
+	ruby_sourcefile = rb_source_filename(name);
     }
 }
 
@@ -943,7 +976,7 @@ ruby_prog_init()
 {
     init_ids();
 
-    ruby_sourcefile = "ruby";
+    ruby_sourcefile = rb_source_filename("ruby");
     rb_define_variable("$VERBOSE", &ruby_verbose);
     rb_define_variable("$-v", &ruby_verbose);
     rb_define_variable("$-w", &ruby_verbose);
@@ -981,6 +1014,7 @@ ruby_set_argv(argc, argv)
     if (origargv) dln_argv0 = origargv[0];
     else          dln_argv0 = argv[0];
 #endif
+    rb_ary_clear(rb_argv);
     for (i=0; i < argc; i++) {
 	rb_ary_push(rb_argv, rb_tainted_str_new2(argv[i]));
     }
