@@ -385,6 +385,18 @@ $z = [1,2]
 $y[$z] = 256
 test_ok($y[$z] == 256)
 
+$x = [1,2,3]
+$x[1,0] = $x
+test_ok($x == [1,1,2,3,2,3])
+
+$x = [1,2,3]
+$x[-1,0] = $x
+test_ok($x == [1,2,1,2,3,3])
+
+$x = [1,2,3]
+$x.concat($x)
+test_ok($x == [1,2,3,1,2,3])
+
 test_check "iterator"
 
 test_ok(!iterator?)
@@ -571,7 +583,6 @@ def fact(n)
   end
   return f
 end
-fact(3)
 $x = fact(40)
 test_ok($x == $x)
 test_ok($x == fact(40))
@@ -613,6 +624,7 @@ $good = true;
 for i in 4000..4096
   n1 = 1 << i;
   if (n1**2-1) / (n1+1) != (n1-1)
+    p i
     $good = false
   end
 end
@@ -629,6 +641,9 @@ test_ok(7 == a.remainder(-b))
 test_ok(-7 == (-a).remainder(b))
 test_ok(-7 == (-a).remainder(-b))
 
+test_ok(10**40+10**20 == 10000000000000000000100000000000000000000)
+test_ok(10**40/10**20 == 100000000000000000000)
+
 test_check "string & char"
 
 test_ok("abcd" == "abcd")
@@ -991,9 +1006,21 @@ File.unlink "script_tmp" or `/bin/rm -f "script_tmp"`
 File.unlink "script_tmp.bak" or `/bin/rm -f "script_tmp.bak"`
 
 $bad = false
-for script in Dir["{lib,sample,ext}/**/*.rb"]
-  `./miniruby -c #{script}`
-  unless $?
+if (dir = File.dirname(File.dirname($0))) == '.'
+  dir = ""
+else
+  dir << "/"
+end
+
+def valid_syntax?(code)
+  eval("BEGIN {return true}\n#{code}")
+rescue Exception
+ensure
+  false
+end
+
+for script in Dir["#{dir}{lib,sample,ext}/**/*.rb"]
+  unless valid_syntax? open(script).read
     $bad = true
   end
 end
