@@ -1,11 +1,12 @@
 #! /usr/local/bin/ruby
-path = $ENV['PATH'].split(/:/)
+
+path = ENV['PATH'].split(File::PATH_SEPARATOR)
 
 for dir in path
-  if File.d(dir)
+  if File.directory?(dir)
     for f in d = Dir.open(dir)
       fpath = dir+"/"+f  
-      if File.f(fpath) && (File.stat(fpath).mode & 022) != 0
+      if File.file?(fpath) && (File.stat(fpath).mode & 022) != 0
 	printf("file %s is writable from other users\n", fpath)
       end
     end
