@@ -3,7 +3,7 @@
   signal.c -
 
   $Author: matz $
-  $Date: 2000/09/27 09:11:12 $
+  $Date: 2002/03/08 06:59:53 $
   created at: Tue Dec 20 10:13:44 JST 1994
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -162,6 +162,9 @@ static struct signals {
 #endif
 #ifdef SIGSOUND
     "SOUND", SIGSOUND,
+#endif
+#ifdef SIGINFO
+    "INFO", SIGINFO,
 #endif
     NULL, 0,
 };
@@ -252,7 +255,7 @@ rb_f_kill(argc, argv)
     else {
 	for (i=1; i<argc; i++) {
 	    Check_Type(argv[i], T_FIXNUM);
-	    if (kill(FIX2UINT(argv[i]), sig) < 0)
+	    if (kill(FIX2INT(argv[i]), sig) < 0)
 		rb_sys_fail(0);
 	}
     }
@@ -260,9 +263,9 @@ rb_f_kill(argc, argv)
 }
 
 static VALUE trap_list[NSIG];
-static int trap_pending_list[NSIG];
-int rb_trap_pending;
-int rb_trap_immediate;
+static rb_atomic_t trap_pending_list[NSIG];
+rb_atomic_t rb_trap_pending;
+rb_atomic_t rb_trap_immediate;
 int rb_prohibit_interrupt;
 
 void
@@ -290,7 +293,12 @@ posix_signal(signum, handler)
     sigemptyset(&sigact.sa_mask);
     sigact.sa_flags = 0;
 #ifdef SA_RESTART
-    sigact.sa_flags |= SA_RESTART; /* SVR4, 4.3+BSD */
+    /* All other signals but VTALRM shall restart restartable syscall
+       VTALRM will cause EINTR to syscall if interrupted.
+    */
+    if (signum != SIGVTALRM) {
+        sigact.sa_flags |= SA_RESTART; /* SVR4, 4.3+BSD */
+    }
 #endif
 #ifdef SA_NOCLDWAIT
     if (signum == SIGCHLD && handler == SIG_IGN)
@@ -303,6 +311,7 @@ posix_signal(signum, handler)
 #define ruby_signal(sig,handle) signal((sig),(handle))
 #endif
 
+static void signal_exec _((int sig));
 static void
 signal_exec(sig)
     int sig;
@@ -342,31 +351,27 @@ sighandle(sig)
     int sig;
 {
 #ifdef NT
-#define end_interrupt() win32_thread_resume_main()
-    if (win32_main_context(sig, sighandle)) return;
-
+#define IN_MAIN_CONTEXT(f, a) (win32_main_context(a, f) ? (void)0 : f(a))
 #else
-#define end_interrupt() (void)0
+#define IN_MAIN_CONTEXT(f, a) f(a)
 #endif
 
     if (sig >= NSIG) {
 	rb_bug("trap_handler: Bad signal %d", sig);
     }
 
-#if !defined(BSD_SIGNAL)
+#if !defined(BSD_SIGNAL) && !defined(POSIX_SIGNAL)
     ruby_signal(sig, sighandle);
 #endif
 
-    if (rb_trap_immediate) {
-	rb_trap_immediate = 0;
-	signal_exec(sig);
-	rb_trap_immediate = 1;
+    if (ATOMIC_TEST(rb_trap_immediate)) {
+	IN_MAIN_CONTEXT(signal_exec, sig);
+	ATOMIC_SET(rb_trap_immediate, 1);
     }
     else {
-	rb_trap_pending++;
-	trap_pending_list[sig]++;
+	ATOMIC_INC(rb_trap_pending);
+	ATOMIC_INC(trap_pending_list[sig]);
     }
-    end_interrupt();
 }
 
 #ifdef SIGBUS
@@ -389,12 +394,25 @@ sigsegv(sig)
 }
 #endif
 
+#ifdef SIGPIPE
+static RETSIGTYPE sigpipe _((int));
+static RETSIGTYPE
+sigpipe(sig)
+    int sig;
+{
+    /* do nothing */
+}
+#endif
+
 void
 rb_trap_exit()
 {
 #ifndef MACOS_UNUSE_SIGNAL
     if (trap_list[0]) {
-	rb_eval_cmd(trap_list[0], rb_ary_new3(1, INT2FIX(0)));
+	VALUE trap_exit = trap_list[0];
+
+	trap_list[0] = 0;
+	rb_eval_cmd(trap_exit, rb_ary_new3(1, INT2FIX(0)));
     }
 #endif
 }
@@ -503,7 +521,7 @@ trap(arg)
 	    s += 3;
 	sig = signm2signo(s);
 	if (sig == 0 && strcmp(s, "EXIT") != 0)
-	    rb_raise(rb_eArgError, "invalid signal SIG%s", s);
+	    rb_raise(rb_eArgError, "unsupported signal SIG%s", s);
     }
 
     if (sig < 0 || sig > NSIG) {
@@ -546,7 +564,7 @@ trap(arg)
 #endif
 #ifdef SIGPIPE
 	  case SIGPIPE:
-	    func = SIG_IGN;
+	    func = sigpipe;
 	    break;
 #endif
 	}
@@ -659,7 +677,7 @@ Init_signal()
     ruby_signal(SIGSEGV, sigsegv);
 #endif
 #ifdef SIGPIPE
-    ruby_signal(SIGPIPE, SIG_IGN);
+    ruby_signal(SIGPIPE, sigpipe);
 #endif
 #endif /* MACOS_UNUSE_SIGNAL */
 }
