@@ -2,8 +2,8 @@
 
   string.c -
 
-  $Author: eban $
-  $Date: 2000/09/26 09:11:47 $
+  $Author: knu $
+  $Date: 2002/10/26 15:57:24 $
   created at: Mon Aug  9 17:12:58 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -54,6 +54,9 @@ VALUE
 rb_str_new2(ptr)
     const char *ptr;
 {
+    if (!ptr) {
+	rb_raise(rb_eArgError, "NULL pointer given");
+    }
     return rb_str_new(ptr, strlen(ptr));
 }
 
@@ -97,19 +100,29 @@ VALUE
 rb_str_new4(orig)
     VALUE orig;
 {
-    if (OBJ_FROZEN(orig)) return orig;
+    VALUE klass;
+
+    klass = CLASS_OF(orig);
+    while (TYPE(klass) == T_ICLASS || FL_TEST(klass, FL_SINGLETON)) {
+	klass = (VALUE)RCLASS(klass)->super;
+    }
+
     if (RSTRING(orig)->orig) {
+	VALUE str;
+
 	if (FL_TEST(orig, STR_NO_ORIG)) {
-	    orig = rb_str_new(RSTRING(orig)->ptr, RSTRING(orig)->len);
-	    OBJ_FREEZE(orig);
-	    return orig;
+	    str = rb_str_new(RSTRING(orig)->ptr, RSTRING(orig)->len);
 	}
-	OBJ_FREEZE(RSTRING(orig)->orig);
-	return RSTRING(orig)->orig;
+	else {
+	    str = rb_str_new3(RSTRING(orig)->orig);
+	}
+	OBJ_FREEZE(str);
+	RBASIC(str)->klass = klass;
+	return str;
     }
     else {
 	NEWOBJ(str, struct RString);
-	OBJSETUP(str, rb_cString, T_STRING);
+	OBJSETUP(str, klass, T_STRING);
 
 	str->len = RSTRING(orig)->len;
 	str->ptr = RSTRING(orig)->ptr;
@@ -164,7 +177,7 @@ rb_str_associate(str, add)
     rb_ary_push(RSTRING(str)->orig, add);
 }
 
-static ID to_str;
+static ID id_to_s;
 
 VALUE
 rb_obj_as_string(obj)
@@ -175,7 +188,7 @@ rb_obj_as_string(obj)
     if (TYPE(obj) == T_STRING) {
 	return obj;
     }
-    str = rb_funcall(obj, to_str, 0);
+    str = rb_funcall(obj, id_to_s, 0);
     if (TYPE(str) != T_STRING)
 	return rb_any_to_s(obj);
     if (OBJ_TAINTED(obj)) OBJ_TAINT(str);
@@ -201,18 +214,15 @@ rb_str_dup(str)
     }
     else if (RSTRING(str)->orig) {
 	str2 = rb_str_new3(RSTRING(str)->orig);
+	FL_UNSET(str2, FL_TAINT);
+	OBJ_INFECT(str2, str);
     }
     else {
-	NEWOBJ(dup, struct RString);
-	OBJSETUP(dup, klass, T_STRING);
-
-	str2 = rb_str_new4(str);
-	dup->len = RSTRING(str2)->len;
-	dup->ptr = RSTRING(str2)->ptr;
-	dup->orig = str2;
-	str2 = (VALUE)dup;
+	str2 = rb_str_new3(rb_str_new4(str));
     }
     OBJ_INFECT(str2, str);
+    if (FL_TEST(str, FL_EXIVAR))
+	rb_clone_generic_ivar(str2, str);
     RBASIC(str2)->klass = klass;
     return str2;
 }
@@ -301,9 +311,8 @@ rb_str_times(str, times)
     }
     RSTRING(str2)->ptr[RSTRING(str2)->len] = '\0';
 
-    if (OBJ_TAINTED(str)) {
-	OBJ_TAINT(str2);
-    }
+    OBJ_INFECT(str2, str);
+    RBASIC(str2)->klass = rb_obj_class(str);
 
     return str2;
 }
@@ -336,7 +345,6 @@ rb_str_substr(str, beg, len)
 
     if (len < 0) return Qnil;
     if (beg > RSTRING(str)->len) return Qnil;
-    if (beg == RSTRING(str)->len && len > 0) return Qnil;
     if (beg < 0) {
 	beg += RSTRING(str)->len;
 	if (beg < 0) return Qnil;
@@ -350,6 +358,7 @@ rb_str_substr(str, beg, len)
     if (len == 0) return rb_str_new(0,0);
 
     str2 = rb_str_new(RSTRING(str)->ptr+beg, len);
+    RBASIC(str2)->klass = rb_obj_class(str);
     if (OBJ_TAINTED(str)) OBJ_TAINT(str2);
 
     return str2;
@@ -364,7 +373,6 @@ str_independent(str)
 	rb_raise(rb_eSecurityError, "Insecure: can't modify string");
     if (!RSTRING(str)->orig || FL_TEST(str, STR_NO_ORIG)) return 1;
     if (TYPE(RSTRING(str)->orig) != T_STRING) rb_bug("non string str->orig");
-    RSTRING(str)->orig = 0;
     return 0;
 }
 
@@ -381,6 +389,7 @@ rb_str_modify(str)
     }
     ptr[RSTRING(str)->len] = 0;
     RSTRING(str)->ptr = ptr;
+    RSTRING(str)->orig = 0;
 }
 
 VALUE
@@ -553,7 +562,6 @@ rb_str_cmp(str1, str2)
 	if (RSTRING(str1)->len > RSTRING(str2)->len) return 1;
 	return -1;
     }
-    if (retval == 0) return 0;
     if (retval > 0) return 1;
     return -1;
 }
@@ -562,8 +570,11 @@ static VALUE
 rb_str_equal(str1, str2)
     VALUE str1, str2;
 {
-    if (TYPE(str2) != T_STRING)
-	return Qfalse;
+    if (str1 == str2) return Qtrue;
+    if (TYPE(str2) != T_STRING) {
+	str2 = rb_check_convert_type(str2, T_STRING, "String", "to_str");
+	if (NIL_P(str2)) return Qfalse;
+    }
 
     if (RSTRING(str1)->len == RSTRING(str2)->len
 	&& rb_str_cmp(str1, str2) == 0) {
@@ -659,7 +670,12 @@ rb_str_index_m(argc, argv, str)
     }
     if (pos < 0) {
 	pos += RSTRING(str)->len;
-	if (pos < 0) return Qnil;
+	if (pos < 0) {
+	    if (TYPE(sub) == T_REGEXP) {
+		rb_backref_set(Qnil);
+	    }
+		return Qnil;
+	}
     }
 
     switch (TYPE(sub)) {
@@ -708,7 +724,12 @@ rb_str_rindex(argc, argv, str)
 	pos = NUM2INT(position);
         if (pos < 0) {
 	    pos += RSTRING(str)->len;
-	    if (pos < 0) return Qnil;
+	    if (pos < 0) {
+		if (TYPE(sub) == T_REGEXP) {
+		    rb_backref_set(Qnil);
+		}
+		return Qnil;
+	    }
         }
 	if (pos > RSTRING(str)->len) pos = RSTRING(str)->len;
     }
@@ -803,6 +824,7 @@ rb_str_succ(orig)
     int n = 0;
 
     str = rb_str_new(RSTRING(orig)->ptr, RSTRING(orig)->len);
+    RBASIC(str)->klass = rb_obj_class(orig);
     OBJ_INFECT(str, orig);
     if (RSTRING(str)->len == 0) return str;
 
@@ -819,17 +841,17 @@ rb_str_succ(orig)
 	sbeg = RSTRING(str)->ptr; s = sbeg + RSTRING(str)->len - 1;
 	c = '\001';
 	while (sbeg <= s) {
-	    *s += 1;
-	    if (*s-- != 0) break;
+           if ((*s += 1) != 0) break;
+           s--;
 	}
     }
     if (s < sbeg) {
-	REALLOC_N(RSTRING(str)->ptr, char, RSTRING(str)->len + 1);
+	REALLOC_N(RSTRING(str)->ptr, char, RSTRING(str)->len + 2);
 	s = RSTRING(str)->ptr + n;
 	memmove(s+1, s, RSTRING(str)->len - n);
 	*s = c;
 	RSTRING(str)->len += 1;
-	
+	RSTRING(str)->ptr[RSTRING(str)->len] = '\0';
     }
 
     return str;
@@ -966,6 +988,7 @@ rb_str_replace(str, beg, len, val)
     }
     RSTRING(str)->len += RSTRING(val)->len - len;
     RSTRING(str)->ptr[RSTRING(str)->len] = '\0';
+    OBJ_INFECT(str, val);
 }
 
 static VALUE rb_str_sub_bang _((int, VALUE*, VALUE));
@@ -981,12 +1004,15 @@ rb_str_aset(str, indx, val)
       case T_FIXNUM:
       num_index:
 	idx = NUM2INT(indx);
-	if (idx < 0) {
-	    idx += RSTRING(str)->len;
-	}
-	if (idx < 0 || RSTRING(str)->len <= idx) {
+	if (RSTRING(str)->len <= idx) {
+	  out_of_range:
 	    rb_raise(rb_eIndexError, "index %d out of string", idx);
 	}
+	if (idx < 0) {
+	    if (-idx > RSTRING(str)->len)
+		goto out_of_range;
+	    idx += RSTRING(str)->len;
+	}
 	if (FIXNUM_P(val)) {
 	    if (RSTRING(str)->len == idx) {
 		RSTRING(str)->len += 1;
@@ -1011,10 +1037,11 @@ rb_str_aset(str, indx, val)
 
       case T_STRING:
 	beg = rb_str_index(str, indx, 0);
-	if (beg != -1) {
-	    if (TYPE(val) != T_STRING) val = rb_str_to_str(val);
-	    rb_str_replace(str, beg, RSTRING(indx)->len, val);
+	if (beg < 0) {
+	    rb_raise(rb_eIndexError, "string not matched");
 	}
+	if (TYPE(val) != T_STRING) val = rb_str_to_str(val);
+	rb_str_replace(str, beg, RSTRING(indx)->len, val);
 	return val;
 
       default:
@@ -1045,15 +1072,16 @@ rb_str_aset_m(argc, argv, str)
 	if (TYPE(argv[2]) != T_STRING) argv[2] = rb_str_to_str(argv[2]);
 	beg = NUM2INT(argv[0]);
 	len = NUM2INT(argv[1]);
-	if (len < 0) rb_raise(rb_eIndexError, "negative length %d", len);
-	if (beg < 0) {
-	    beg += RSTRING(str)->len;
+	if (len < 0) rb_raise(rb_eIndexError, "negative length %ld", len);
+	if (RSTRING(str)->len < beg) {
+	  out_of_range:
+	    rb_raise(rb_eIndexError, "index %ld out of string", beg);
 	}
-	if (beg < 0 || RSTRING(str)->len < beg) {
-	    if (beg < 0) {
-		beg -= RSTRING(str)->len;
+	if (beg < 0) {
+	    if (-beg > RSTRING(str)->len) {
+		goto out_of_range;
 	    }
-	    rb_raise(rb_eIndexError, "index %d out of string", beg);
+	    beg += RSTRING(str)->len;
 	}
 	if (beg + len > RSTRING(str)->len) {
 	    len = RSTRING(str)->len - beg;
@@ -1085,7 +1113,9 @@ rb_str_slice_bang(argc, argv, str)
     }
     buf[i] = rb_str_new(0,0);
     result = rb_str_aref_m(argc, buf, str);
-    rb_str_aset_m(argc+1, buf, str);
+    if (!NIL_P(result)) {
+	rb_str_aset_m(argc+1, buf, str);
+    }
     return result;
 }
 
@@ -1124,7 +1154,7 @@ rb_str_sub_bang(argc, argv, str)
 	iter = 1;
     }
     else if (argc == 2) {
-	repl = rb_obj_as_string(argv[1]);;
+	repl = rb_str_to_str(argv[1]);;
 	if (OBJ_TAINTED(repl)) tainted = 1;
     }
     else {
@@ -1197,7 +1227,7 @@ str_gsub(argc, argv, str, bang)
 	iter = 1;
     }
     else if (argc == 2) {
-	repl = rb_obj_as_string(argv[1]);
+	repl = rb_str_to_str(argv[1]);
 	if (OBJ_TAINTED(repl)) tainted = 1;
     }
     else {
@@ -1275,11 +1305,15 @@ str_gsub(argc, argv, str, bang)
 	if (str_independent(str)) {
 	    free(RSTRING(str)->ptr);
 	}
+	else {
+	    RSTRING(str)->orig = 0;
+	}
     }
     else {
 	NEWOBJ(dup, struct RString);
 	OBJSETUP(dup, rb_cString, T_STRING);
 	OBJ_INFECT(dup, str);
+	RBASIC(dup)->klass = rb_obj_class(str);
 	str = (VALUE)dup;
 	dup->orig = 0;
     }
@@ -1313,12 +1347,24 @@ static VALUE
 rb_str_replace_m(str, str2)
     VALUE str, str2;
 {
+    if (str == str2) return str;
     if (TYPE(str2) != T_STRING) str2 = rb_str_to_str(str2);
-    rb_str_modify(str);
-    rb_str_resize(str, RSTRING(str2)->len);
-    memcpy(RSTRING(str)->ptr, RSTRING(str2)->ptr, RSTRING(str2)->len);
-    if (OBJ_TAINTED(str2)) OBJ_TAINT(str);
 
+    if (RSTRING(str2)->orig && !FL_TEST(str2, STR_NO_ORIG)) {
+	if (str_independent(str)) {
+	    free(RSTRING(str)->ptr);
+	}
+	RSTRING(str)->len = RSTRING(str2)->len;
+	RSTRING(str)->ptr = RSTRING(str2)->ptr;
+	RSTRING(str)->orig = RSTRING(str2)->orig;
+    }
+    else {
+	rb_str_modify(str);
+	rb_str_resize(str, RSTRING(str2)->len);
+	memcpy(RSTRING(str)->ptr, RSTRING(str2)->ptr, RSTRING(str2)->len);
+    }
+
+    if (OBJ_TAINTED(str2)) OBJ_TAINT(str);
     return str;
 }
 
@@ -1330,7 +1376,7 @@ uscore_get()
     line = rb_lastline_get();
     if (TYPE(line) != T_STRING) {
 	rb_raise(rb_eTypeError, "$_ value need to be String (%s given)",
-		 NIL_P(line)?"nil":rb_class2name(CLASS_OF(line)));
+		 NIL_P(line) ? "nil" : rb_class2name(CLASS_OF(line)));
     }
     return line;
 }
@@ -1381,16 +1427,17 @@ static VALUE
 rb_str_reverse_bang(str)
     VALUE str;
 {
-    char *s, *e, *p, *q;
+    char *s, *e;
+    char c;
 
+    rb_str_modify(str);
     s = RSTRING(str)->ptr;
     e = s + RSTRING(str)->len - 1;
-    p = q = ALLOCA_N(char, RSTRING(str)->len);
-
-    while (e >= s) {
-	*p++ = *e--;
+    while (s < e) {
+	c = *s;
+	*s++ = *e;
+	*e-- = c;
     }
-    MEMCPY(RSTRING(str)->ptr, q, char, RSTRING(str)->len);
 
     return str;
 }
@@ -1411,6 +1458,8 @@ rb_str_reverse(str)
     while (e >= s) {
 	*p++ = *e--;
     }
+    OBJ_INFECT(obj, str);
+    RBASIC(obj)->klass = rb_obj_class(str);
 
     return obj;
 }
@@ -1468,90 +1517,63 @@ VALUE
 rb_str_inspect(str)
     VALUE str;
 {
-#define STRMAX 80
-    char buf[STRMAX];
     char *p, *pend;
-    char *b;
-    VALUE inspect;
+    VALUE result = rb_str_new2("\"");
+    char s[5];
 
     p = RSTRING(str)->ptr; pend = p + RSTRING(str)->len;
-    b = buf;
-    *b++ = '"';
-
-#define CHECK(n) {\
-    if (b - buf + n > STRMAX - 4) {\
-	strcpy(b, "...");\
-	b += 3;\
-        break;\
-    }\
-}
-
     while (p < pend) {
 	char c = *p++;
 	if (ismbchar(c) && p < pend) {
 	    int len = mbclen(c);
-
-	    CHECK(len);
-	    *b++ = c;
-	    while (--len) {
-		*b++ = *p++;
-	    }
+	    rb_str_cat(result, p - 1, len);
+	    p += len - 1;
 	}
 	else if (c == '"'|| c == '\\') {
-	    CHECK(2);
-	    *b++ = '\\';
-	    *b++ = c;
+	    s[0] = '\\'; s[1] = c;
+	    rb_str_cat(result, s, 2);
 	}
 	else if (ISPRINT(c)) {
-	    CHECK(1);
-	    *b++ = c;
+	    s[0] = c;
+	    rb_str_cat(result, s, 1);
 	}
 	else if (c == '\n') {
-	    CHECK(2);
-	    *b++ = '\\';
-	    *b++ = 'n';
+	    s[0] = '\\'; s[1] = 'n';
+	    rb_str_cat(result, s, 2);
 	}
 	else if (c == '\r') {
-	    CHECK(2);
-	    *b++ = '\\';
-	    *b++ = 'r';
+	    s[0] = '\\'; s[1] = 'r';
+	    rb_str_cat(result, s, 2);
 	}
 	else if (c == '\t') {
-	    CHECK(2);
-	    *b++ = '\\';
-	    *b++ = 't';
+	    s[0] = '\\'; s[1] = 't';
+	    rb_str_cat(result, s, 2);
 	}
 	else if (c == '\f') {
-	    CHECK(2);
-	    *b++ = '\\';
-	    *b++ = 'f';
+	    s[0] = '\\'; s[1] = 'f';
+	    rb_str_cat(result, s, 2);
 	}
 	else if (c == '\013') {
-	    CHECK(2);
-	    *b++ = '\\';
-	    *b++ = 'v';
+	    s[0] = '\\'; s[1] = 'v';
+	    rb_str_cat(result, s, 2);
 	}
 	else if (c == '\007') {
-	    CHECK(2);
-	    *b++ = '\\';
-	    *b++ = 'a';
+	    s[0] = '\\'; s[1] = 'a';
+	    rb_str_cat(result, s, 2);
 	}
 	else if (c == 033) {
-	    CHECK(2);
-	    *b++ = '\\';
-	    *b++ = 'e';
+	    s[0] = '\\'; s[1] = 'e';
+	    rb_str_cat(result, s, 2);
 	}
 	else {
-	    CHECK(4);
-	    *b++ = '\\';
-	    sprintf(b, "%03o", c & 0377);
-	    b += 3;
+	    sprintf(s, "\\%03o", c & 0377);
+	    rb_str_cat2(result, s);
 	}
     }
-    *b++ = '"';
-    inspect = rb_str_new(buf, b - buf);
-    OBJ_INFECT(inspect, str);
-    return inspect;
+    rb_str_cat2(result, "\"");
+
+    OBJ_INFECT(result, str);
+    return result;
 }
 
 static VALUE
@@ -1641,6 +1663,7 @@ rb_str_dump(str)
     }
     *q++ = '"';
 
+    OBJ_INFECT(result, str);
     return result;
 }
 
@@ -1657,7 +1680,7 @@ rb_str_upcase_bang(str)
 	if (ismbchar(*s)) {
 	    s+=mbclen(*s) - 1;
 	}
-	else if (islower(*s)) {
+	else if (ISLOWER(*s)) {
 	    *s = toupper(*s);
 	    modify = 1;
 	}
@@ -1796,8 +1819,11 @@ trnext(t)
     for (;;) {
 	if (!t->gen) {
 	    if (t->p == t->pend) return -1;
+	    if (t->p < t->pend - 1 && *t->p == '\\') {
+		t->p++;
+	    }
 	    t->now = *(USTR)t->p++;
-	    if (t->p < t->pend && *t->p == '-') {
+	    if (t->p < t->pend - 1 && *t->p == '-') {
 		t->p++;
 		if (t->p < t->pend) {
 		    if (t->now > *(USTR)t->p) {
@@ -1978,6 +2004,9 @@ rb_str_delete_bang(argc, argv, str)
     int init = 1;
     int i;
 
+    if (argc < 1) {
+	rb_raise(rb_eArgError, "wrong # of arguments");
+    }
     for (i=0; i<argc; i++) {
 	VALUE s = argv[i];
 
@@ -2145,7 +2174,7 @@ rb_str_split_m(argc, argv, str)
 	i = 1;
     }
 
-    if (argc == 0) {
+    if (NIL_P(spat)) {
 	if (!NIL_P(rb_fs)) {
 	    spat = rb_fs;
 	    goto fs_set;
@@ -2211,7 +2240,6 @@ rb_str_split_m(argc, argv, str)
 		    if (!NIL_P(limit) && lim <= ++i) break;
 		}
 		end++;
-		if (ismbchar(*ptr)) {ptr++; end++;}
 	    }
 	}
     }
@@ -2326,10 +2354,11 @@ rb_str_each_line(argc, argv, str)
 	    if (*++p != '\n') continue;
 	    while (*p == '\n') p++;
 	}
-	if (p[-1] == newline &&
+	if (RSTRING(str)->ptr < p && p[-1] == newline &&
 	    (rslen <= 1 ||
 	     rb_memcmp(RSTRING(rs)->ptr, p-rslen, rslen) == 0)) {
 	    line = rb_str_new(s, p - s);
+	    OBJ_INFECT(line, str);
 	    rb_yield(line);
 	    if (RSTRING(str)->ptr != ptr || RSTRING(str)->len != len)
 		rb_raise(rb_eArgError, "string modified");
@@ -2634,11 +2663,16 @@ rb_str_crypt(str, salt)
     VALUE str, salt;
 {
     extern char *crypt();
+    VALUE result;
 
     if (TYPE(salt) != T_STRING) salt = rb_str_to_str(salt);
     if (RSTRING(salt)->len < 2)
 	rb_raise(rb_eArgError, "salt too short(need >=2 bytes)");
-    return rb_str_new2(crypt(RSTRING(str)->ptr, RSTRING(salt)->ptr));
+
+    result = rb_str_new2(crypt(RSTRING(str)->ptr, RSTRING(salt)->ptr));
+    OBJ_INFECT(result, str);
+    OBJ_INFECT(result, salt);
+    return result;
 }
 
 static VALUE
@@ -2708,13 +2742,15 @@ rb_str_ljust(str, w)
     VALUE res;
     char *p, *pend;
 
-    if (width < 0 || RSTRING(str)->len >= width) return str;
+    if (width < 0 || RSTRING(str)->len >= width) return rb_str_dup(str);
     res = rb_str_new(0, width);
+    RBASIC(res)->klass = rb_obj_class(str);
     memcpy(RSTRING(res)->ptr, RSTRING(str)->ptr, RSTRING(str)->len);
     p = RSTRING(res)->ptr + RSTRING(str)->len; pend = RSTRING(res)->ptr + width;
     while (p < pend) {
 	*p++ = ' ';
     }
+    OBJ_INFECT(res, str);
     return res;
 }
 
@@ -2727,13 +2763,15 @@ rb_str_rjust(str, w)
     VALUE res;
     char *p, *pend;
 
-    if (width < 0 || RSTRING(str)->len >= width) return str;
+    if (width < 0 || RSTRING(str)->len >= width) return rb_str_dup(str);
     res = rb_str_new(0, width);
+    RBASIC(res)->klass = rb_obj_class(str);
     p = RSTRING(res)->ptr; pend = p + width - RSTRING(str)->len;
     while (p < pend) {
 	*p++ = ' ';
     }
     memcpy(pend, RSTRING(str)->ptr, RSTRING(str)->len);
+    OBJ_INFECT(res, str);
     return res;
 }
 
@@ -2747,8 +2785,9 @@ rb_str_center(str, w)
     char *p, *pend;
     long n;
 
-    if (width < 0 || RSTRING(str)->len >= width) return str;
+    if (width < 0 || RSTRING(str)->len >= width) return rb_str_dup(str);
     res = rb_str_new(0, width);
+    RBASIC(res)->klass = rb_obj_class(str);
     n = (width - RSTRING(str)->len)/2;
     p = RSTRING(res)->ptr; pend = p + n;
     while (p < pend) {
@@ -2759,6 +2798,7 @@ rb_str_center(str, w)
     while (p < pend) {
 	*p++ = ' ';
     }
+    OBJ_INFECT(res, str);
     return res;
 }
 
@@ -2890,7 +2930,7 @@ Init_String()
     rb_define_method(rb_cString, "slice", rb_str_aref_m, -1);
     rb_define_method(rb_cString, "slice!", rb_str_slice_bang, -1);
 
-    to_str = rb_intern("to_s");
+    id_to_s = rb_intern("to_s");
 
     rb_fs = Qnil;
     rb_define_hooked_variable("$;", &rb_fs, 0, rb_str_setter);
