@@ -3,7 +3,7 @@
   util.c -
 
   $Author: matz $
-  $Date: 2000/07/27 09:49:24 $
+  $Date: 2002/12/02 18:19:04 $
   created at: Fri Mar 10 17:22:34 JST 1995
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -11,50 +11,16 @@
 **********************************************************************/
 
 #include <stdio.h>
+#include <ctype.h>
+#include <errno.h>
 
 #ifdef NT
 #include "missing/file.h"
 #endif
 
-#define RUBY_NO_INLINE
+#define INLINE_DEFINE
 #include "ruby.h"
 
-VALUE
-rb_class_of(obj)
-    VALUE obj;
-{
-    if (FIXNUM_P(obj)) return rb_cFixnum;
-    if (obj == Qnil) return rb_cNilClass;
-    if (obj == Qfalse) return rb_cFalseClass;
-    if (obj == Qtrue) return rb_cTrueClass;
-    if (SYMBOL_P(obj)) return rb_cSymbol;
-
-    return RBASIC(obj)->klass;
-}
-
-int
-rb_type(obj)
-    VALUE obj;
-{
-    if (FIXNUM_P(obj)) return T_FIXNUM;
-    if (obj == Qnil) return T_NIL;
-    if (obj == Qfalse) return T_FALSE;
-    if (obj == Qtrue) return T_TRUE;
-    if (obj == Qundef) return T_UNDEF;
-    if (SYMBOL_P(obj)) return T_SYMBOL;
-
-    return BUILTIN_TYPE(obj);
-}
-
-int
-rb_special_const_p(obj)
-    VALUE obj;
-{
-    if (SPECIAL_CONST_P(obj)) return Qtrue;
-
-    return Qfalse;
-}
-
 #include "util.h"
 #ifndef HAVE_STRING_H
 char *strchr _((char*,char));
@@ -83,7 +49,7 @@ const char *start;
 int len;
 int *retlen;
 {
-    static char hexdigit[] = "0123456789abcdef0123456789ABCDEFx";
+    static char hexdigit[] = "0123456789abcdef0123456789ABCDEF";
     register const char *s = start;
     register unsigned long retval = 0;
     char *tmp;
@@ -218,7 +184,7 @@ ruby_add_suffix(str, suffix)
     char buf[1024];
 
     if (RSTRING(str)->len > 1000)
-        rb_fatal("Cannot do inplace edit on long filename (%d characters)",
+        rb_fatal("Cannot do inplace edit on long filename (%ld characters)",
 		 RSTRING(str)->len);
 
 #if defined(DJGPP) || defined(__CYGWIN32__) || defined(NT)
@@ -299,231 +265,131 @@ valid_filename(char *s)
 #endif
 #endif
 
-#ifdef DJGPP
-#include <libc/stubs.h>
-#include <stdio.h>		/* For FILENAME_MAX */
-#include <errno.h>		/* For errno */
-#include <fcntl.h>		/* For LFN stuff */
-#include <go32.h>
-#include <dpmi.h>		/* For dpmisim */
-#include <crt0.h>		/* For crt0 flags */
-#include <libc/dosio.h>
+#if defined __DJGPP__
 
-static unsigned use_lfn;
+#include <dpmi.h>
 
-static char *__get_current_directory(char *out, int drive_number);
+static char dbcs_table[256];
 
-static char *
-__get_current_directory(char *out, int drive_number)
+int
+make_dbcs_table()
 {
-  __dpmi_regs r;
-  char tmpbuf[FILENAME_MAX];
-
-  memset(&r, 0, sizeof(r));
-  if(use_lfn)
-    r.x.ax = 0x7147;
-  else
-    r.h.ah = 0x47;
-  r.h.dl = drive_number + 1;
-  r.x.si = __tb_offset;
-  r.x.ds = __tb_segment;
-  __dpmi_int(0x21, &r);
-
-  if (r.x.flags & 1)
-  {
-    errno = r.x.ax;
-    return out;
-  }
-  else
-  {
-    dosmemget(__tb, sizeof(tmpbuf), tmpbuf);
-    strcpy(out+1,tmpbuf);
-
-    /* Root path, don't insert "/", it'll be added later */
-    if (*(out + 1) != '\0')
-      *out = '/';
-    else
-      *out = '\0';
-    return out + strlen(out);
-  }
-}
-
-__inline__ static int
-is_slash(int c)
-{
-  return c == '/' || c == '\\';
-}
-
-__inline__ static int
-is_term(int c)
-{
-  return c == '/' || c == '\\' || c == '\0';
-}
-
-#ifdef SJIS
-__inline__ static int
-is_sjis1(int c)
-{
-  return 0x81 <= c && (c <= 0x9f || 0xe0 <= c);
-}
-#endif
-
-/* Takes as input an arbitrary path.  Fixes up the path by:
-   1. Removing consecutive slashes
-   2. Removing trailing slashes
-   3. Making the path absolute if it wasn't already
-   4. Removing "." in the path
-   5. Removing ".." entries in the path (and the directory above them)
-   6. Adding a drive specification if one wasn't there
-   7. Converting all slashes to '/'
- */
-void
-fixpath(const char *in, char *out)
-{
-  int		drive_number;
-  const char	*ip = in;
-  char		*op = out;
-  int		preserve_case = _preserve_fncase();
-  char		*name_start;
-
-  use_lfn = _USE_LFN;
-
-  /* Add drive specification to output string */
-  if (((*ip >= 'a' && *ip <= 'z') ||
-       (*ip >= 'A' && *ip <= 'Z'))
-      && (*(ip + 1) == ':'))
-  {
-    if (*ip >= 'a' && *ip <= 'z')
-    {
-      drive_number = *ip - 'a';
-      *op++ = *ip++;
-    }
-    else
-    {
-      drive_number = *ip - 'A';
-      if (*ip <= 'Z')
-	*op++ = drive_number + 'a';
-      else
-	*op++ = *ip;
-      ++ip;
-    }
-    *op++ = *ip++;
-  }
-  else
-  {
     __dpmi_regs r;
-    r.h.ah = 0x19;
+    struct {
+	unsigned char start;
+	unsigned char end;
+    } vec;
+    int offset;
+
+    memset(&r, 0, sizeof(r));
+    r.x.ax = 0x6300;
     __dpmi_int(0x21, &r);
-    drive_number = r.h.al;
-    *op++ = drive_number + (drive_number < 26 ? 'a' : 'A');
-    *op++ = ':';
-  }
+    offset = r.x.ds * 16 + r.x.si;
 
-  /* Convert relative path to absolute */
-  if (!is_slash(*ip))
-    op = __get_current_directory(op, drive_number);
-
-  /* Step through the input path */
-  while (*ip)
-  {
-    /* Skip input slashes */
-    if (is_slash(*ip))
-    {
-      ip++;
-      continue;
+    for (;;) {
+	int i;
+	dosmemget(offset, sizeof vec, &vec);
+	if (!vec.start && !vec.end)
+	    break;
+	for (i = vec.start; i <= vec.end; i++)
+	    dbcs_table[i] = 1;
+	offset += 2;
     }
-
-    /* Skip "." and output nothing */
-    if (*ip == '.' && is_term(*(ip + 1)))
-    {
-      ip++;
-      continue;
-    }
-
-    /* Skip ".." and remove previous output directory */
-    if (*ip == '.' && *(ip + 1) == '.' && is_term(*(ip + 2)))
-    {
-      ip += 2;
-      /* Don't back up over drive spec */
-      if (op > out + 2)
-	/* This requires "/" to follow drive spec */
-	while (!is_slash(*--op));
-      continue;
-    }
-
-    /* Copy path component from in to out */
-    *op++ = '/';
-#ifndef SJIS
-    while (!is_term(*ip)) *op++ = *ip++;
-#else
-    while (!is_term(*ip)) {
-      if (is_sjis1((unsigned char)*ip))
-	*op++ = *ip++;
-      *op++ = *ip++;
-    }
-#endif
-  }
-
-  /* If root directory, insert trailing slash */
-  if (op == out + 2) *op++ = '/';
-
-  /* Null terminate the output */
-  *op = '\0';
-
-  /* switch FOO\BAR to foo/bar, downcase where appropriate */
-  for (op = out + 3, name_start = op - 1; *name_start; op++)
-  {
-    char long_name[FILENAME_MAX], short_name[13];
-
-#ifdef SJIS
-    if (is_sjis1((unsigned char)*op)) {
-      op++;
-      continue;
-    }
-#endif
-    if (*op == '\\')
-      *op = '/';
-    if (!preserve_case && (*op == '/' || *op == '\0'))
-    {
-      memcpy(long_name, name_start+1, op - name_start - 1);
-      long_name[op - name_start - 1] = '\0';
-      if (!strcmp(_lfn_gen_short_fname(long_name, short_name), long_name))
-      {
-#ifndef SJIS
-	while (++name_start < op)
-	  if (*name_start >= 'A' && *name_start <= 'Z')
-	    *name_start += 'a' - 'A';
-#else
-	while (++name_start < op) {
-	  if (is_sjis1((unsigned char)*name_start))
-	    name_start++;
-	  else if (*name_start >= 'A' && *name_start <= 'Z')
-	    *name_start += 'a' - 'A';
-	}
-#endif
-      }
-      else
-	name_start = op;
-    }
-    else if (*op == '\0')
-      break;
-  }
 }
 
-#ifdef TEST
-
-int main (int argc, char *argv[])
+int
+mblen(const char *s, size_t n)
 {
-  char fixed[FILENAME_MAX];
-  if (argc > 1)
-    {
-      _fixpath (argv[1], fixed);
-      printf ("You mean %s?\n", fixed);
+    static int need_init = 1;
+    if (need_init) {
+	make_dbcs_table();
+	need_init = 0;
     }
-  return 0;
+    if (s) {
+	if (n == 0 || *s == 0)
+	    return 0;
+	return dbcs_table[(unsigned char)*s] + 1;
+    }
+    else
+	return 1;
+}
+
+struct PathList {
+    struct PathList *next;
+    char *path;
+};
+
+struct PathInfo {
+    struct PathList *head;
+    int count;
+};
+
+static void
+push_element(const char *path, VALUE vinfo)
+{
+    struct PathList *p;
+    struct PathInfo *info = (struct PathInfo *)vinfo;
+
+    p = ALLOC(struct PathList);
+    MEMZERO(p, struct PathList, 1);
+    p->path = ruby_strdup(path);
+    p->next = info->head;
+    info->head = p;
+    info->count++;
+}
+
+#include <dirent.h>
+int __opendir_flags = __OPENDIR_PRESERVE_CASE;
+
+char **
+__crt0_glob_function(char *path)
+{
+    int len = strlen(path);
+    int i;
+    char **rv;
+    char path_buffer[PATH_MAX];
+    char *buf = path_buffer;
+    char *p;
+    struct PathInfo info;
+    struct PathList *plist;
+
+    if (PATH_MAX <= len)
+	buf = ruby_xmalloc(len + 1);
+
+    strncpy(buf, path, len);
+    buf[len] = '\0';
+
+    for (p = buf; *p; p += mblen(p, MB_CUR_MAX))
+	if (*p == '\\')
+	    *p = '/';
+
+    info.count = 0;
+    info.head = 0;
+
+    rb_iglob(buf, push_element, (VALUE)&info);
+
+    if (buf != path_buffer)
+	ruby_xfree(buf);
+
+    if (info.count == 0)
+	return 0;
+
+    rv = ruby_xmalloc((info.count + 1) * sizeof (char *));
+
+    plist = info.head;
+    i = 0;
+    while (plist) {
+	struct PathList *cur;
+	rv[i] = plist->path;
+	cur = plist;
+	plist = plist->next;
+	ruby_xfree(cur);
+	i++;
+    }
+    rv[i] = 0;
+    return rv;
 }
 
-#endif
 #endif
 
 /* mm.c */
@@ -770,3 +636,256 @@ ruby_strdup(str)
 
     return tmp;
 }
+
+/* copyright notice for strtod implementation --
+ *
+ * Copyright (c) 1988-1993 The Regents of the University of California.
+ * Copyright (c) 1994 Sun Microsystems, Inc.
+ *
+ * Permission to use, copy, modify, and distribute this
+ * software and its documentation for any purpose and without
+ * fee is hereby granted, provided that the above copyright
+ * notice appear in all copies.  The University of California
+ * makes no representations about the suitability of this
+ * software for any purpose.  It is provided "as is" without
+ * express or implied warranty.
+ *
+ */
+
+#define TRUE 1
+#define FALSE 0
+
+static int maxExponent = 511;	/* Largest possible base 10 exponent.  Any
+				 * exponent larger than this will already
+				 * produce underflow or overflow, so there's
+				 * no need to worry about additional digits.
+				 */
+static double powersOf10[] = {	/* Table giving binary powers of 10.  Entry */
+    10.0,			/* is 10^2^i.  Used to convert decimal */
+    100.0,			/* exponents into floating-point numbers. */
+    1.0e4,
+    1.0e8,
+    1.0e16,
+    1.0e32,
+    1.0e64,
+    1.0e128,
+    1.0e256
+};
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * strtod --
+ *
+ *	This procedure converts a floating-point number from an ASCII
+ *	decimal representation to internal double-precision format.
+ *
+ * Results:
+ *	The return value is the double-precision floating-point
+ *	representation of the characters in string.  If endPtr isn't
+ *	NULL, then *endPtr is filled in with the address of the
+ *	next character after the last one that was part of the
+ *	floating-point number.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+double
+ruby_strtod(string, endPtr)
+    const char *string;		/* A decimal ASCII floating-point number,
+				 * optionally preceded by white space.
+				 * Must have form "-I.FE-X", where I is the
+				 * integer part of the mantissa, F is the
+				 * fractional part of the mantissa, and X
+				 * is the exponent.  Either of the signs
+				 * may be "+", "-", or omitted.  Either I
+				 * or F may be omitted, or both.  The decimal
+				 * point isn't necessary unless F is present.
+				 * The "E" may actually be an "e".  E and X
+				 * may both be omitted (but not just one).
+				 */
+    char **endPtr;		/* If non-NULL, store terminating character's
+				 * address here. */
+{
+    int sign, expSign = FALSE;
+    double fraction, dblExp, *d;
+    register const char *p;
+    register int c;
+    int exp = 0;		/* Exponent read from "EX" field. */
+    int fracExp = 0;		/* Exponent that derives from the fractional
+				 * part.  Under normal circumstatnces, it is
+				 * the negative of the number of digits in F.
+				 * However, if I is very long, the last digits
+				 * of I get dropped (otherwise a long I with a
+				 * large negative exponent could cause an
+				 * unnecessary overflow on I alone).  In this
+				 * case, fracExp is incremented one for each
+				 * dropped digit. */
+    int mantSize;		/* Number of digits in mantissa. */
+    int decPt;			/* Number of mantissa digits BEFORE decimal
+				 * point. */
+    const char *pExp;		/* Temporarily holds location of exponent
+				 * in string. */
+
+    /*
+     * Strip off leading blanks and check for a sign.
+     */
+
+    errno = 0;
+    p = string;
+    while (ISSPACE(*p)) {
+	p += 1;
+    }
+    if (*p == '-') {
+	sign = TRUE;
+	p += 1;
+    }
+    else {
+	if (*p == '+') {
+	    p += 1;
+	}
+	sign = FALSE;
+    }
+
+    /*
+     * Count the number of digits in the mantissa (including the decimal
+     * point), and also locate the decimal point.
+     */
+
+    decPt = -1;
+    for (mantSize = 0; ; mantSize += 1) {
+	c = *p;
+	if (!ISDIGIT(c)) {
+	    if ((c != '.') || (decPt >= 0)) {
+		break;
+	    }
+	    decPt = mantSize;
+	}
+	p += 1;
+    }
+
+    /*
+     * Now suck up the digits in the mantissa.  Use two integers to
+     * collect 9 digits each (this is faster than using floating-point).
+     * If the mantissa has more than 18 digits, ignore the extras, since
+     * they can't affect the value anyway.
+     */
+    
+    pExp  = p;
+    p -= mantSize;
+    if (decPt < 0) {
+	decPt = mantSize;
+    }
+    else {
+	mantSize -= 1;			/* One of the digits was the point. */
+    }
+    if (mantSize > 18) {
+	fracExp = decPt - 18;
+	mantSize = 18;
+    }
+    else {
+	fracExp = decPt - mantSize;
+    }
+    if (mantSize == 0) {
+	fraction = 0.0;
+	p = string;
+	goto done;
+    }
+    else {
+	int frac1, frac2;
+	frac1 = 0;
+	for ( ; mantSize > 9; mantSize -= 1) {
+	    c = *p;
+	    p += 1;
+	    if (c == '.') {
+		c = *p;
+		p += 1;
+	    }
+	    frac1 = 10*frac1 + (c - '0');
+	}
+	frac2 = 0;
+	for (; mantSize > 0; mantSize -= 1) {
+	    c = *p;
+	    p += 1;
+	    if (c == '.') {
+		c = *p;
+		p += 1;
+	    }
+	    frac2 = 10*frac2 + (c - '0');
+	}
+	fraction = (1.0e9 * frac1) + frac2;
+    }
+
+    /*
+     * Skim off the exponent.
+     */
+
+    p = pExp;
+    if ((*p == 'E') || (*p == 'e')) {
+	p += 1;
+	if (*p == '-') {
+	    expSign = TRUE;
+	    p += 1;
+	}
+	else {
+	    if (*p == '+') {
+		p += 1;
+	    }
+	    expSign = FALSE;
+	}
+	while (ISDIGIT(*p)) {
+	    exp = exp * 10 + (*p - '0');
+	    p += 1;
+	}
+    }
+    if (expSign) {
+	exp = fracExp - exp;
+    }
+    else {
+	exp = fracExp + exp;
+    }
+
+    /*
+     * Generate a floating-point number that represents the exponent.
+     * Do this by processing the exponent one bit at a time to combine
+     * many powers of 2 of 10. Then combine the exponent with the
+     * fraction.
+     */
+    
+    if (exp < 0) {
+	expSign = TRUE;
+	exp = -exp;
+    }
+    else {
+	expSign = FALSE;
+    }
+    if (exp > maxExponent) {
+	exp = maxExponent;
+	errno = ERANGE;
+    }
+    dblExp = 1.0;
+    for (d = powersOf10; exp != 0; exp >>= 1, d += 1) {
+	if (exp & 01) {
+	    dblExp *= *d;
+	}
+    }
+    if (expSign) {
+	fraction /= dblExp;
+    }
+    else {
+	fraction *= dblExp;
+    }
+
+done:
+    if (endPtr != NULL) {
+	*endPtr = (char *) p;
+    }
+
+    if (sign) {
+	return -fraction;
+    }
+    return fraction;
+}
