@@ -3,7 +3,7 @@
   version.c -
 
   $Author: matz $
-  $Date: 2000/05/01 09:41:31 $
+  $Date: 2002/01/16 09:22:39 $
   created at: Thu Sep 30 20:08:01 JST 1993
 
   Copyright (C) 1993-2000 Yukihiro Matsumoto
@@ -17,9 +17,9 @@
 void
 Init_version()
 {
-    VALUE v = rb_str_new2(RUBY_VERSION);
-    VALUE d = rb_str_new2(RUBY_RELEASE_DATE);
-    VALUE p = rb_str_new2(RUBY_PLATFORM);
+    VALUE v = rb_obj_freeze(rb_str_new2(RUBY_VERSION));
+    VALUE d = rb_obj_freeze(rb_str_new2(RUBY_RELEASE_DATE));
+    VALUE p = rb_obj_freeze(rb_str_new2(RUBY_PLATFORM));
 
     rb_define_global_const("RUBY_VERSION", v);
     rb_define_global_const("RUBY_RELEASE_DATE", d);
@@ -40,6 +40,6 @@ ruby_show_version()
 void
 ruby_show_copyright()
 {
-    printf("ruby - Copyright (C) 1993-2000 Yukihiro Matsumoto\n");
+    printf("ruby - Copyright (C) 1993-2002 Yukihiro Matsumoto\n");
     exit(0);
 }
