@@ -24,16 +24,50 @@
 #include <windows.h>
 #include <winbase.h>
 #include <wincon.h>
+#ifdef __MINGW32__
+#include <mswsock.h>
+#endif
 #include "win32.h"
 #include "dir.h"
 #ifndef index
 #define index(x, y) strchr((x), (y))
 #endif
+#define isdirsep(x) ((x) == '/' || (x) == '\\')
+
+#undef fclose
+#undef close
+#undef setsockopt
 
 #ifndef bool
 #define bool int
 #endif
 
+#if USE_INTERRUPT_WINSOCK
+
+# if defined(_MSC_VER) && _MSC_VER <= 1000
+/* VC++4.0 doesn't have this. */
+extern DWORD WSAWaitForMultipleEvents(DWORD nevent, const HANDLE *events,
+				      BOOL waitall, DWORD timeout,
+				      BOOL alertable);
+# endif
+
+# define WaitForMultipleEvents WSAWaitForMultipleEvents
+# define CreateSignal() (HANDLE)WSACreateEvent()
+# define SetSignal(ev) WSASetEvent(ev)
+# define ResetSignal(ev) WSAResetEvent(ev)
+#else  /* USE_INTERRUPT_WINSOCK */
+# define WaitForMultipleEvents WaitForMultipleObjectsEx
+# define CreateSignal() CreateEvent(NULL, FALSE, FALSE, NULL);
+# define SetSignal(ev) SetEvent(ev)
+# define ResetSignal(ev) (void)0
+#endif /* USE_INTERRUPT_WINSOCK */
+
+#ifdef WIN32_DEBUG
+#define Debug(something) something
+#else
+#define Debug(something) /* nothing */
+#endif
+
 #define TO_SOCKET(x)	_get_osfhandle(x)
 
 bool NtSyncProcess = TRUE;
@@ -46,6 +80,7 @@ static bool NtHasRedirection (char *);
 static int valid_filename(char *s);
 static void StartSockets ();
 static char *str_grow(struct RString *str, size_t new_size);
+static DWORD wait_events(HANDLE event, DWORD timeout);
 
 char *NTLoginName;
 
@@ -76,17 +111,23 @@ IsWinNT(void) {
 }
 
 /* main thread constants */
-HANDLE rb_CurrentProcessHandle;
-HANDLE rb_MainThreadHandle;
-DWORD rb_MainThreadId;
-HANDLE rb_InterruptEvent;
+static struct {
+    HANDLE handle;
+    DWORD id;
+} main_thread;
+
+/* interrupt stuff */
+static HANDLE interrupted_event;
 
 HANDLE GetCurrentThreadHandle(void)
 {
+    static HANDLE current_process_handle = NULL;
     HANDLE h;
-    HANDLE proc = rb_CurrentProcessHandle;
 
-    if (!DuplicateHandle(proc, GetCurrentThread(), proc, &h,
+    if (!current_process_handle)
+	current_process_handle = GetCurrentProcess();
+    if (!DuplicateHandle(current_process_handle, GetCurrentThread(),
+			 current_process_handle, &h,
 			 0, FALSE, DUPLICATE_SAME_ACCESS))
 	return NULL;
     return h;
@@ -98,68 +139,78 @@ HANDLE GetCurrentThreadHandle(void)
 #define LK_ERR(f,i) ((f) ? (i = 0) : (errno = GetLastError()))
 #define LK_LEN      0xffff0000
 
-int
-flock(int fd, int oper)
+static VALUE
+flock_winnt(VALUE self, int argc, VALUE* argv)
 {
     OVERLAPPED o;
     int i = -1;
-    HANDLE fh;
+    const HANDLE fh = (HANDLE)self;
+    const int oper = argc;
 
-    fh = (HANDLE)_get_osfhandle(fd);
     memset(&o, 0, sizeof(o));
 
-    if(IsWinNT()) {
-        switch(oper) {
-        case LOCK_SH:       /* shared lock */
-            LK_ERR(LockFileEx(fh, 0, 0, LK_LEN, 0, &o),i);
-            break;
-        case LOCK_EX:       /* exclusive lock */
-            LK_ERR(LockFileEx(fh, LOCKFILE_EXCLUSIVE_LOCK, 0, LK_LEN, 0, &o),i);
-            break;
-        case LOCK_SH|LOCK_NB:   /* non-blocking shared lock */
-            LK_ERR(LockFileEx(fh, LOCKFILE_FAIL_IMMEDIATELY, 0, LK_LEN, 0, &o),i);
-            break;
-        case LOCK_EX|LOCK_NB:   /* non-blocking exclusive lock */
-            LK_ERR(LockFileEx(fh,
-                   LOCKFILE_EXCLUSIVE_LOCK|LOCKFILE_FAIL_IMMEDIATELY,
-                   0, LK_LEN, 0, &o),i);
-	    if(errno == EDOM) errno = EWOULDBLOCK;
-            break;
-        case LOCK_UN:       /* unlock lock */
-	    if (UnlockFileEx(fh, 0, LK_LEN, 0, &o)) {
-		i = 0;
-	    }
-	    else {
-		/* GetLastError() must returns `ERROR_NOT_LOCKED' */
+    switch(oper) {
+      case LOCK_SH:		/* shared lock */
+	LK_ERR(LockFileEx(fh, 0, 0, LK_LEN, 0, &o), i);
+	break;
+      case LOCK_EX:		/* exclusive lock */
+	LK_ERR(LockFileEx(fh, LOCKFILE_EXCLUSIVE_LOCK, 0, LK_LEN, 0, &o), i);
+	break;
+      case LOCK_SH|LOCK_NB:	/* non-blocking shared lock */
+	LK_ERR(LockFileEx(fh, LOCKFILE_FAIL_IMMEDIATELY, 0, LK_LEN, 0, &o), i);
+	break;
+      case LOCK_EX|LOCK_NB:	/* non-blocking exclusive lock */
+	LK_ERR(LockFileEx(fh,
+			  LOCKFILE_EXCLUSIVE_LOCK|LOCKFILE_FAIL_IMMEDIATELY,
+			  0, LK_LEN, 0, &o), i);
+	if (errno == EDOM)
+	    errno = EWOULDBLOCK;
+	break;
+      case LOCK_UN:		/* unlock lock */
+	if (UnlockFileEx(fh, 0, LK_LEN, 0, &o)) {
+	    i = 0;
+	    if (errno == EDOM)
 		errno = EWOULDBLOCK;
-	    }
-	    if(errno == EDOM) errno = EWOULDBLOCK;
-            break;
-        default:            /* unknown */
-            errno = EINVAL;
-            break;
-        }
+	}
+	else {
+	    /* GetLastError() must returns `ERROR_NOT_LOCKED' */
+	    errno = EWOULDBLOCK;
+	}
+	break;
+      default:            /* unknown */
+	errno = EINVAL;
+	break;
     }
-    else if(IsWin95()) {
-        switch(oper) {
-        case LOCK_EX:
-	    while(i == -1) {
-	        LK_ERR(LockFile(fh, 0, 0, LK_LEN, 0), i);
-		if(errno != EDOM && i == -1) break;
-	    }
-	    break;
-	case LOCK_EX | LOCK_NB:
+    return i;
+}
+
+static VALUE
+flock_win95(VALUE self, int argc, VALUE* argv)
+{
+    int i = -1;
+    const HANDLE fh = (HANDLE)self;
+    const int oper = argc;
+
+    switch(oper) {
+      case LOCK_EX:
+	while(i == -1) {
 	    LK_ERR(LockFile(fh, 0, 0, LK_LEN, 0), i);
-	    if(errno == EDOM) errno = EWOULDBLOCK;
-            break;
-        case LOCK_UN:
-            LK_ERR(UnlockFile(fh, 0, 0, LK_LEN, 0), i);
-	    if(errno == EDOM) errno = EWOULDBLOCK;
-            break;
-        default:
-            errno = EINVAL;
-            break;
-        }
+	    if (errno != EDOM && i == -1) break;
+	}
+	break;
+      case LOCK_EX | LOCK_NB:
+	LK_ERR(LockFile(fh, 0, 0, LK_LEN, 0), i);
+	if (errno == EDOM)
+	    errno = EWOULDBLOCK;
+	break;
+      case LOCK_UN:
+	LK_ERR(UnlockFile(fh, 0, 0, LK_LEN, 0), i);
+	if (errno == EDOM)
+	    errno = EWOULDBLOCK;
+	break;
+      default:
+	errno = EINVAL;
+	break;
     }
     return i;
 }
@@ -167,6 +218,22 @@ flock(int fd, int oper)
 #undef LK_ERR
 #undef LK_LEN
 
+int
+flock(int fd, int oper)
+{
+    static asynchronous_func_t locker = NULL;
+
+    if (!locker) {
+	if (IsWinNT())
+	    locker = flock_winnt;
+	else
+	    locker = flock_win95;
+    }
+
+    return win32_asynchronize(locker,
+			      (VALUE)_get_osfhandle(fd), oper, NULL,
+			      (DWORD)-1);
+}
 
 //#undef const
 //FILE *fdopen(int, const char *);
@@ -293,20 +360,22 @@ static char *szInternalCmds[] = {
 int
 isInternalCmd(char *cmd)
 {
-        int i, fRet=0;
-	char **vec;
-        int vecc = NtMakeCmdVector(cmd, &vec, FALSE);
+    int i, fRet=0;
+    char **vec;
+    int vecc = NtMakeCmdVector(cmd, &vec, FALSE);
 
-        for( i = 0; szInternalCmds[i] ; i++){
-	    if(!strcmp(szInternalCmds[i], vec[0])){
-		fRet = 1;
-		break;
-	    }
-        }
- 
-	SafeFree (vec, vecc);
+    if (vecc == 0)
+	return 0;
+    for( i = 0; szInternalCmds[i] ; i++){
+	if(!strcasecmp(szInternalCmds[i], vec[0])){
+	    fRet = 1;
+	    break;
+	}
+    }
 
-	return fRet;
+    SafeFree(vec, vecc);
+
+    return fRet;
 }
 
 
@@ -653,7 +722,7 @@ mypclose(FILE *fp)
     MyPopenRecord[i].pipe  = NULL;
     MyPopenRecord[i].pid   = 0;
 
-    return exitcode;
+    return (int)((exitcode & 0xff) << 8);
 }
 #endif
 
@@ -673,10 +742,12 @@ char *cmd;
     register char **a;
     register char *s;
     char **argv;
-    int status;
+    int status = -1;
     char *shell, *cmd2;
     int mode = NtSyncProcess ? P_WAIT : P_NOWAIT;
+    char **env = NULL;
 
+    env = win32_get_environ();
     /* save an extra exec if possible */
     if ((shell = getenv("RUBYSHELL")) != 0) {
 	if (NtHasRedirection(cmd)) {
@@ -700,16 +771,18 @@ char *cmd;
 	    argv[1] = "-c";
 	    argv[2] = cmdline;
 	    argv[4] = NULL;
-	    status = spawnvpe(mode, argv[0], argv, environ);
+	    status = spawnvpe(mode, argv[0], argv, env);
 	    /* return spawnle(mode, shell, shell, "-c", cmd, (char*)0, environ); */
 	    free(cmdline);
-	    return status;
+	    if (env) win32_free_environ(env);
+	    return (int)((status & 0xff) << 8);
 	} 
     }
     else if ((shell = getenv("COMSPEC")) != 0) {
 	if (NtHasRedirection(cmd) /* || isInternalCmd(cmd) */) {
-	  do_comspec_shell:
-	    return spawnle(mode, shell, shell, "/c", cmd, (char*)0, environ);
+	    status = spawnle(mode, shell, shell, "/c", cmd, (char*)0, env);
+	    if (env) win32_free_environ(env);
+	    return (int)((status & 0xff) << 8);
 	}
     }
 
@@ -718,24 +791,26 @@ char *cmd;
     strcpy(cmd2, cmd);
     a = argv;
     for (s = cmd2; *s;) {
-	while (*s && isspace(*s)) s++;
+	while (*s && ISSPACE(*s)) s++;
 	if (*s)
 	    *(a++) = s;
-	while (*s && !isspace(*s)) s++;
+	while (*s && !ISSPACE(*s)) s++;
 	if (*s)
 	    *s++ = '\0';
     }
     *a = NULL;
     if (argv[0]) {
-	if ((status = spawnvpe(mode, argv[0], argv, environ)) == -1) {
+	if ((status = spawnvpe(mode, argv[0], argv, env)) == -1) {
 	    free(argv);
 	    free(cmd2);
+	    if (env) win32_free_environ(env);
 	    return -1;
 	}
     }
     free(cmd2);
     free(argv);
-    return status;
+    if (env) win32_free_environ(env);
+    return (int)((status & 0xff) << 8);
 }
 
 #endif
@@ -859,9 +934,10 @@ typedef struct {
 } ListInfo;
 
 static void
-insert(char *path, ListInfo *listinfo)
+insert(const char *path, VALUE vinfo)
 {
     NtCmdLineElement *tmpcurr;
+    ListInfo *listinfo = (ListInfo *)vinfo;
 
     tmpcurr = ALLOC(NtCmdLineElement);
     MEMZERO(tmpcurr, NtCmdLineElement, 1);
@@ -902,7 +978,7 @@ NtCmdGlob (NtCmdLineElement *patt)
     for (p = buf; *p; p = CharNext(p))
 	if (*p == '\\')
 	    *p = '/';
-    rb_glob(buf, insert, (VALUE)&listinfo);
+    rb_iglob(buf, insert, (VALUE)&listinfo);
     if (buf != buffer)
 	free(buf);
 
@@ -995,7 +1071,7 @@ NtMakeCmdVector (char *cmdline, char ***vec, int InputCmd)
     //
 
     ptr = cmdline+(cmdlen - 1);
-    while(ptr >= cmdline && isspace(*ptr))
+    while(ptr >= cmdline && ISSPACE(*ptr))
         --ptr;
     *++ptr = '\0';
 
@@ -1015,7 +1091,7 @@ NtMakeCmdVector (char *cmdline, char ***vec, int InputCmd)
 	// zap any leading whitespace
 	//
 
-	while(isspace(*ptr))
+	while(ISSPACE(*ptr))
 	    ptr++;
 	base = ptr;
 
@@ -1218,7 +1294,6 @@ NtMakeCmdVector (char *cmdline, char ***vec, int InputCmd)
 }
 
 
-#if !defined __MINGW32__
 //
 // UNIX compatible directory access functions for NT
 //
@@ -1232,7 +1307,7 @@ NtMakeCmdVector (char *cmdline, char ***vec, int InputCmd)
 //
 
 DIR *
-opendir(char *filename)
+opendir(const char *filename)
 {
     DIR            *p;
     long            len;
@@ -1240,8 +1315,8 @@ opendir(char *filename)
     char            scannamespc[PATHLEN];
     char	   *scanname = scannamespc;
     struct stat	    sbuf;
-    WIN32_FIND_DATA FindData;
-    HANDLE          fh;
+    struct _finddata_t fd;
+    long               fh;
     char            root[PATHLEN];
     char            volname[PATHLEN];
     DWORD           serial, maxname, flags;
@@ -1250,9 +1325,9 @@ opendir(char *filename)
     // check to see if we\'ve got a directory
     //
 
-    if ((stat (filename, &sbuf) < 0 ||
+    if ((win32_stat (filename, &sbuf) < 0 ||
 	sbuf.st_mode & _S_IFDIR == 0) &&
-	(!isalpha(filename[0]) || filename[1] != ':' || filename[2] != '\0' ||
+	(!ISALPHA(filename[0]) || filename[1] != ':' || filename[2] != '\0' ||
 	((1 << (filename[0] & 0x5f) - 'A') & GetLogicalDrives()) == 0)) {
 	return NULL;
     }
@@ -1280,8 +1355,8 @@ opendir(char *filename)
     // do the FindFirstFile call
     //
 
-    fh = FindFirstFile (scanname, &FindData);
-    if (fh == INVALID_HANDLE_VALUE) {
+    fh = _findfirst(scanname, &fd);
+    if (fh == -1) {
 	return NULL;
     }
 
@@ -1290,9 +1365,9 @@ opendir(char *filename)
     // filenames that we find.
     //
 
-    idx = strlen(FindData.cFileName)+1;
+    idx = strlen(fd.name)+1;
     p->start = ALLOC_N(char, idx);
-    strcpy (p->start, FindData.cFileName);
+    strcpy(p->start, fd.name);
     p->nfiles++;
     
     //
@@ -1301,8 +1376,8 @@ opendir(char *filename)
     // the variable idx should point one past the null terminator
     // of the previous string found.
     //
-    while (FindNextFile(fh, &FindData)) {
-	len = strlen (FindData.cFileName);
+    while (_findnext(fh, &fd) == 0) {
+	len = strlen(fd.name);
 
 	//
 	// bump the string table size by enough for the
@@ -1315,11 +1390,11 @@ opendir(char *filename)
 	if (p->start == NULL) {
             rb_fatal ("opendir: malloc failed!\n");
 	}
-	strcpy(&p->start[idx], FindData.cFileName);
+	strcpy(&p->start[idx], fd.name);
 	p->nfiles++;
 	idx += len+1;
     }
-    FindClose(fh);
+    _findclose(fh);
     p->size = idx;
     p->curr = p->start;
     return p;
@@ -1407,7 +1482,6 @@ closedir(DIR *dirp)
 	free(dirp->start);
 	free(dirp);
 }
-#endif
 
 
 //
@@ -1667,6 +1741,8 @@ my_open_osfhandle(long osfhandle, int flags)
     return fh;			/* return handle */
 }
 
+#undef getsockopt
+
 static int
 is_socket(SOCKET fd)
 {
@@ -1691,30 +1767,6 @@ is_socket(SOCKET fd)
     return TRUE;
 }
 
-FILE *
-myfdopen (int fd, const char *mode)
-{
-    if (is_socket((SOCKET)fd)) {
-	int fh;
-
-	fh = my_open_osfhandle((SOCKET)fd, O_RDWR|O_BINARY);
-	return _fdopen(fh, mode);		// return file pointer
-    }
-    else {
-	return (_fdopen(fd, mode));
-    }
-}
-
-
-void
-myfdclose(FILE *fp)
-{
-#if !defined __MINGW32__
-    _free_osfhnd(fileno(fp));
-#endif
-    fclose(fp);
-}
-
 
 //
 // Since the errors returned by the socket error function 
@@ -1724,26 +1776,31 @@ myfdclose(FILE *fp)
 
 #undef strerror
 
-char * 
-mystrerror(int e) 
+char *
+mystrerror(int e)
 {
     static char buffer[512];
 #if !defined __MINGW32__
     extern int sys_nerr;
 #endif
     DWORD source = 0;
+    char *p;
 
     if (e < 0 || e > sys_nerr) {
 	if (e < 0)
 	    e = GetLastError();
-	if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, &source, e, 0,
+	if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
+			  FORMAT_MESSAGE_IGNORE_INSERTS, &source, e, 0,
 			  buffer, 512, NULL) == 0) {
-	    strcpy (buffer, "Unknown Error");
+	    strcpy(buffer, "Unknown Error");
+	}
+	for (p = buffer + strlen(buffer) - 1; buffer <= p; p--) {
+	    if (*p != '\r' && *p != '\n') break;
+	    *p = 0;
 	}
 	return buffer;
     }
     return strerror(e);
-	
 }
 
 //
@@ -1904,13 +1961,19 @@ myselect (int nfds, fd_set *rd, fd_set *wr, fd_set *ex,
     long r;
     fd_set file_rd;
     fd_set file_wr;
+#ifdef USE_INTERRUPT_WINSOCK
     fd_set trap;
+#endif /* USE_INTERRUPT_WINSOCK */
     int file_nfds;
-    int trap_immediate = rb_trap_immediate;
 
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
+    r = 0;
+    if (rd && rd->fd_count > r) r = rd->fd_count;
+    if (wr && wr->fd_count > r) r = wr->fd_count;
+    if (ex && ex->fd_count > r) r = ex->fd_count;
+    if (nfds > r) nfds = r;
     if (nfds == 0 && timeout) {
 	Sleep(timeout->tv_sec * 1000 + timeout->tv_usec / 1000);
 	return 0;
@@ -1925,25 +1988,26 @@ myselect (int nfds, fd_set *rd, fd_set *wr, fd_set *ex,
 	if (wr) *wr = file_wr;
 	return file_nfds;
     }
+
+#if USE_INTERRUPT_WINSOCK
     if (ex)
 	trap = *ex;
     else
 	trap.fd_count = 0;
     if (trap.fd_count < FD_SETSIZE)
-	trap.fd_array[trap.fd_count++] = rb_InterruptEvent;
+	trap.fd_array[trap.fd_count++] = (SOCKET)interrupted_event;
     // else unable to catch interrupt.
-    if (trap_immediate)
-	TRAP_END;
-    if ((r = select (nfds, rd, wr, &trap, timeout)) == SOCKET_ERROR) {
+    ex = &trap;
+#endif /* USE_INTERRUPT_WINSOCK */
+
+    if ((r = select (nfds, rd, wr, ex, timeout)) == SOCKET_ERROR) {
 	errno = WSAGetLastError();
 	switch (errno) {
-	case WSAEINTR:
+	  case WSAEINTR:
 	    errno = EINTR;
 	    break;
 	}
     }
-    if (trap_immediate)
-	TRAP_BEG;
     return r;
 }
 
@@ -1953,7 +2017,7 @@ StartSockets ()
     WORD version;
     WSADATA retdata;
     int ret;
-	int iSockOpt;
+    int iSockOpt;
     
     //
     // initalize the winsock interface and insure that it\'s
@@ -1961,46 +2025,44 @@ StartSockets ()
     //
     version = MAKEWORD(1, 1);
     if (ret = WSAStartup(version, &retdata))
-        rb_fatal ("Unable to locate winsock library!\n");
+	rb_fatal ("Unable to locate winsock library!\n");
     if (LOBYTE(retdata.wVersion) != 1)
-        rb_fatal("could not find version 1 of winsock dll\n");
+	rb_fatal("could not find version 1 of winsock dll\n");
 
     if (HIBYTE(retdata.wVersion) != 1)
-        rb_fatal("could not find version 1 of winsock dll\n");
+	rb_fatal("could not find version 1 of winsock dll\n");
 
     atexit((void (*)(void)) WSACleanup);
 
-	iSockOpt = SO_SYNCHRONOUS_NONALERT;
+    iSockOpt = SO_SYNCHRONOUS_NONALERT;
     /*
      * Enable the use of sockets as filehandles
      */
     setsockopt(INVALID_SOCKET, SOL_SOCKET, SO_OPENTYPE,
-		(char *)&iSockOpt, sizeof(iSockOpt));
+	       (char *)&iSockOpt, sizeof(iSockOpt));
 
-    rb_CurrentProcessHandle = GetCurrentProcess();
-    rb_MainThreadHandle = GetCurrentThreadHandle();
-    rb_MainThreadId = GetCurrentThreadId();
-    rb_InterruptEvent = CreateEvent(NULL, FALSE, TRUE, NULL);
+    main_thread.handle = GetCurrentThreadHandle();
+    main_thread.id = GetCurrentThreadId();
+
+    interrupted_event = CreateSignal();
+    if (!interrupted_event)
+	rb_fatal("Unable to create interrupt event!\n");
+    NtSocketsInitialized = 1;
 }
 
 #undef accept
 
-SOCKET 
+SOCKET
 myaccept (SOCKET s, struct sockaddr *addr, int *addrlen)
 {
     SOCKET r;
-    int trap_immediate = rb_trap_immediate;
 
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
-    if (trap_immediate)
-	TRAP_END;
     if ((r = accept (TO_SOCKET(s), addr, addrlen)) == INVALID_SOCKET)
 	errno = WSAGetLastError();
-    if (trap_immediate)
-	TRAP_BEG;
-    return r;
+    return my_open_osfhandle(r, O_RDWR|O_BINARY);
 }
 
 #undef bind
@@ -2010,10 +2072,10 @@ mybind (SOCKET s, struct sockaddr *addr, int addrlen)
 {
     int r;
 
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
-    if ((r = bind (s, addr, addrlen)) == SOCKET_ERROR)
+    if ((r = bind (TO_SOCKET(s), addr, addrlen)) == SOCKET_ERROR)
 	errno = WSAGetLastError();
     return r;
 }
@@ -2024,10 +2086,10 @@ int
 myconnect (SOCKET s, struct sockaddr *addr, int addrlen)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
-    if ((r = connect (s, addr, addrlen)) == SOCKET_ERROR)
+    if ((r = connect (TO_SOCKET(s), addr, addrlen)) == SOCKET_ERROR)
 	errno = WSAGetLastError();
     return r;
 }
@@ -2039,7 +2101,7 @@ int
 mygetpeername (SOCKET s, struct sockaddr *addr, int *addrlen)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = getpeername (TO_SOCKET(s), addr, addrlen)) == SOCKET_ERROR)
@@ -2053,7 +2115,7 @@ int
 mygetsockname (SOCKET s, struct sockaddr *addr, int *addrlen)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = getsockname (TO_SOCKET(s), addr, addrlen)) == SOCKET_ERROR)
@@ -2061,16 +2123,14 @@ mygetsockname (SOCKET s, struct sockaddr *addr, int *addrlen)
     return r;
 }
 
-#undef getsockopt
-
 int 
 mygetsockopt (SOCKET s, int level, int optname, char *optval, int *optlen)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
-    if ((r = getsockopt (s, level, optname, optval, optlen)) == SOCKET_ERROR)
+    if ((r = getsockopt (TO_SOCKET(s), level, optname, optval, optlen)) == SOCKET_ERROR)
 	errno = WSAGetLastError();
     return r;
 }
@@ -2081,7 +2141,7 @@ int
 myioctlsocket (SOCKET s, long cmd, u_long *argp)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = ioctlsocket (TO_SOCKET(s), cmd, argp)) == SOCKET_ERROR)
@@ -2095,10 +2155,10 @@ int
 mylisten (SOCKET s, int backlog)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
-    if ((r = listen (s, backlog)) == SOCKET_ERROR)
+    if ((r = listen (TO_SOCKET(s), backlog)) == SOCKET_ERROR)
 	errno = WSAGetLastError();
     return r;
 }
@@ -2109,7 +2169,7 @@ int
 myrecv (SOCKET s, char *buf, int len, int flags)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = recv (TO_SOCKET(s), buf, len, flags)) == SOCKET_ERROR)
@@ -2124,7 +2184,7 @@ myrecvfrom (SOCKET s, char *buf, int len, int flags,
 		struct sockaddr *from, int *fromlen)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = recvfrom (TO_SOCKET(s), buf, len, flags, from, fromlen)) == SOCKET_ERROR)
@@ -2138,7 +2198,7 @@ int
 mysend (SOCKET s, char *buf, int len, int flags)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = send (TO_SOCKET(s), buf, len, flags)) == SOCKET_ERROR)
@@ -2153,7 +2213,7 @@ mysendto (SOCKET s, char *buf, int len, int flags,
 		struct sockaddr *to, int tolen)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = sendto (TO_SOCKET(s), buf, len, flags, to, tolen)) == SOCKET_ERROR)
@@ -2167,10 +2227,10 @@ int
 mysetsockopt (SOCKET s, int level, int optname, char *optval, int optlen)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
-    if ((r = setsockopt (s, level, optname, optval, optlen))
+    if ((r = setsockopt (TO_SOCKET(s), level, optname, optval, optlen))
     		 == SOCKET_ERROR)
 	errno = WSAGetLastError();
     return r;
@@ -2182,7 +2242,7 @@ int
 myshutdown (SOCKET s, int how)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = shutdown (TO_SOCKET(s), how)) == SOCKET_ERROR)
@@ -2196,14 +2256,14 @@ SOCKET
 mysocket (int af, int type, int protocol)
 {
     SOCKET s;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((s = socket (af, type, protocol)) == INVALID_SOCKET) {
 	errno = WSAGetLastError();
 	//fprintf(stderr, "socket fail (%d)", WSAGetLastError());
     }
-    return s;
+    return my_open_osfhandle(s, O_RDWR|O_BINARY);
 }
 
 #undef gethostbyaddr
@@ -2212,7 +2272,7 @@ struct hostent *
 mygethostbyaddr (char *addr, int len, int type)
 {
     struct hostent *r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = gethostbyaddr (addr, len, type)) == NULL)
@@ -2226,7 +2286,7 @@ struct hostent *
 mygethostbyname (char *name)
 {
     struct hostent *r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = gethostbyname (name)) == NULL)
@@ -2240,7 +2300,7 @@ int
 mygethostname (char *name, int len)
 {
     int r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = gethostname (name, len)) == SOCKET_ERROR)
@@ -2254,7 +2314,7 @@ struct protoent *
 mygetprotobyname (char *name)
 {
     struct protoent *r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = getprotobyname (name)) == NULL)
@@ -2268,7 +2328,7 @@ struct protoent *
 mygetprotobynumber (int num)
 {
     struct protoent *r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = getprotobynumber (num)) == NULL)
@@ -2282,7 +2342,7 @@ struct servent *
 mygetservbyname (char *name, char *proto)
 {
     struct servent *r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = getservbyname (name, proto)) == NULL)
@@ -2296,7 +2356,7 @@ struct servent *
 mygetservbyport (int port, char *proto)
 {
     struct servent *r;
-    if (!NtSocketsInitialized++) {
+    if (!NtSocketsInitialized) {
 	StartSockets();
     }
     if ((r = getservbyport (port, proto)) == NULL)
@@ -2346,8 +2406,11 @@ waitpid (pid_t pid, int *stat_loc, int options)
     } else {
 	timeout = INFINITE;
     }
-    if (WaitForSingleObject((HANDLE) pid, timeout) == WAIT_OBJECT_0) {
+    if (wait_events((HANDLE)pid, timeout) == WAIT_OBJECT_0) {
 	pid = _cwait(stat_loc, pid, 0);
+#if !defined __BORLANDC__
+	*stat_loc <<= 8;
+#endif
 	return pid;
     }
     return 0;
@@ -2358,13 +2421,23 @@ waitpid (pid_t pid, int *stat_loc, int options)
 int _cdecl
 gettimeofday(struct timeval *tv, struct timezone *tz)
 {                                
-    struct timeb tb;
+    SYSTEMTIME st;
+    time_t t;
+    struct tm tm;
 
-    ftime(&tb);
-    tv->tv_sec = tb.time;
-    tv->tv_usec = tb.millitm * 1000;
+    GetLocalTime(&st);
+    tm.tm_sec = st.wSecond;
+    tm.tm_min = st.wMinute;
+    tm.tm_hour = st.wHour;
+    tm.tm_mday = st.wDay;
+    tm.tm_mon = st.wMonth - 1;
+    tm.tm_year = st.wYear - 1900;
+    tm.tm_isdst = -1;
+    t = mktime(&tm);
+    tv->tv_sec = t;
+    tv->tv_usec = st.wMilliseconds * 1000;
 
-	return 0;
+    return 0;
 }
 
 char *
@@ -2445,14 +2518,16 @@ wait()
 char *
 win32_getenv(const char *name)
 {
-    char *curitem = NULL;	/* XXX threadead */
-    DWORD curlen = 0;		/* XXX threadead */
+    static char *curitem = NULL;
+    static DWORD curlen = 0;
     DWORD needlen;
 
-    curlen = 512;
-    curitem = ALLOC_N(char, curlen);
+    if (curitem == NULL || curlen == 0) {
+	curlen = 512;
+	curitem = ALLOC_N(char, curlen);
+    }
 
-    needlen = GetEnvironmentVariable(name,curitem,curlen);
+    needlen = GetEnvironmentVariable(name, curitem, curlen);
     if (needlen != 0) {
 	while (needlen > curlen) {
 	    REALLOC_N(curitem, char, needlen);
@@ -2478,7 +2553,7 @@ myrename(const char *oldpath, const char *newpath)
     newatts = GetFileAttributes(newpath);
 
     if (oldatts == -1) {
-	printf("file to move doesn't exist");
+	errno = GetLastError();
 	return -1;
     }
 
@@ -2488,25 +2563,26 @@ myrename(const char *oldpath, const char *newpath)
     if (!MoveFile(oldpath, newpath))
 	res = -1;
 
-    if (res == 0 || (GetLastError() != ERROR_ALREADY_EXISTS
-		  && GetLastError() != ERROR_FILE_EXISTS))
-	goto done;
-
-    if (IsWinNT()) {
-	if (MoveFileEx(oldpath, newpath, MOVEFILE_REPLACE_EXISTING))
-	    res = 0;
-    } else {
-	for (;;) {
-	    if (!DeleteFile(newpath) && GetLastError() != ERROR_FILE_NOT_FOUND)
-		break;
-	    else if (MoveFile(oldpath, newpath)) {
-		  res = 0;
-		  break;
+    if (res) {
+	switch (GetLastError()) {
+	  case ERROR_ALREADY_EXISTS:
+	  case ERROR_FILE_EXISTS:
+	    if (IsWinNT()) {
+		if (MoveFileEx(oldpath, newpath, MOVEFILE_REPLACE_EXISTING))
+		    res = 0;
+	    } else {
+		for (;;) {
+		    if (!DeleteFile(newpath) && GetLastError() != ERROR_FILE_NOT_FOUND)
+			break;
+		    else if (MoveFile(oldpath, newpath)) {
+			res = 0;
+			break;
+		    }
+		}
 	    }
 	}
     }
 
-done:
     if (res)
 	errno = GetLastError();
     else
@@ -2515,6 +2591,67 @@ done:
     return res;
 }
 
+static int
+isUNCRoot(const char *path)
+{
+    if (path[0] == '\\' && path[1] == '\\') {
+	const char *p;
+	for (p = path + 3; *p; p = CharNext(p)) {
+	    if (*p == '\\')
+		break;
+	}
+	if (p[0] && p[1]) {
+	    for (p++; *p; p = CharNext(p)) {
+		if (*p == '\\')
+		    break;
+	    }
+	    if (!p[0] || !p[1])
+		return 1;
+	}
+    }
+    return 0;
+}
+
+int
+win32_stat(const char *path, struct stat *st)
+{
+    const char *p;
+    char *buf1 = ALLOCA_N(char, strlen(path) + 2);
+    char *buf2 = ALLOCA_N(char, MAXPATHLEN);
+    char *s;
+    int len;
+    int ret;
+
+    for (p = path, s = buf1; *p; p++, s++) {
+	if (*p == '/')
+	    *s = '\\';
+	else
+	    *s = *p;
+    }
+    *s = '\0';
+    len = s - buf1;
+    if (!len || '\"' == *(--s)) {
+	errno = ENOENT;
+	return -1;
+    }
+    p = CharPrev(buf1, buf1 + len);
+
+    if (isUNCRoot(buf1)) {
+	if (*p != '\\')
+	    strcat(buf1, "\\");
+    } else if (*p == '\\' || *p == ':')
+	strcat(buf1, ".");
+    if (_fullpath(buf2, buf1, MAXPATHLEN)) {
+	ret = stat(buf2, st);
+	if (ret == 0) {
+	    st->st_mode &= ~(S_IWGRP | S_IWOTH);
+	}
+	return ret;
+    }
+    else
+	return -1;
+}
+
 static long
 filetime_to_clock(FILETIME *ft)
 {
@@ -2546,88 +2683,354 @@ mytimes(struct tms *tmbuf)
 }
 
 
-static int win32_thread_exclusive(void)
-{
-    if (GetCurrentThreadId() == rb_MainThreadId) return FALSE;
+#undef Sleep
+#define yield_once() Sleep(0)
+#define yield_until(condition) do yield_once(); while (!(condition))
 
-    SuspendThread(rb_MainThreadHandle);
-    return TRUE;
+static DWORD wait_events(HANDLE event, DWORD timeout)
+{
+    HANDLE events[2];
+    int count = 0;
+    DWORD ret;
+
+    if (event) {
+	events[count++] = event;
+    }
+    events[count++] = interrupted_event;
+
+    ret = WaitForMultipleEvents(count, events, FALSE, timeout, TRUE);
+
+    if (ret == WAIT_OBJECT_0 + count - 1) {
+	ResetSignal(interrupted_event);
+	errno = EINTR;
+    }
+
+    return ret;
 }
 
-void win32_thread_resume_main(void)
+static CRITICAL_SECTION* system_state(void)
 {
-    if (GetCurrentThreadId() != rb_MainThreadId)
-	ResumeThread(rb_MainThreadHandle);
+    static int initialized = 0;
+    static CRITICAL_SECTION syssect;
+
+    if (!initialized) {
+	InitializeCriticalSection(&syssect);
+	initialized = 1;
+    }
+    return &syssect;
 }
 
-static void win32_suspend_self(void)
+static LONG flag_interrupt = -1;
+static volatile DWORD tlsi_interrupt = TLS_OUT_OF_INDEXES;
+
+void win32_disable_interrupt(void)
 {
-    SuspendThread(GetCurrentThread());
+    if (IsWinNT()) {
+	EnterCriticalSection(system_state());
+	return;
+    }
+
+    if (tlsi_interrupt == TLS_OUT_OF_INDEXES) {
+	tlsi_interrupt = TlsAlloc();
+    }
+
+    {
+	DWORD ti = (DWORD)TlsGetValue(tlsi_interrupt);
+	while (InterlockedIncrement(&flag_interrupt) > 0 && !ti) {
+	    InterlockedDecrement(&flag_interrupt);
+	    Sleep(1);
+	}
+	TlsSetValue(tlsi_interrupt, (PVOID)++ti);
+    }
 }
 
-static void win32_call_handler(int arg, void (*handler)(int), CONTEXT ctx)
+void win32_enable_interrupt(void)
 {
-    handler(arg);
-    ctx.ContextFlags = CONTEXT_FULL | CONTEXT_FLOATING_POINT;
-    SetThreadContext(rb_MainThreadHandle, &ctx);
+    if (IsWinNT()) {
+	LeaveCriticalSection(system_state());
+	return;
+    }
+
+    InterlockedDecrement(&flag_interrupt);
+    TlsSetValue(tlsi_interrupt, (PVOID)((DWORD)TlsGetValue(tlsi_interrupt) - 1));
 }
 
-static int catch_interrupt(unsigned long msec)
+struct handler_arg_t {
+    void (*handler)(int);
+    int arg;
+    int status;
+    int userstate;
+    HANDLE handshake;
+};
+
+static void win32_call_handler(struct handler_arg_t* h)
 {
-    return !WaitForSingleObject(rb_InterruptEvent, msec);
+    int status;
+    RUBY_CRITICAL(rb_protect((VALUE (*)())h->handler, (VALUE)h->arg, &h->status);
+		  status = h->status;
+		  SetEvent(h->handshake));
+    if (status) {
+	rb_jump_tag(status);
+    }
+    h->userstate = 1;		/* never syscall after here */
+    for (;;);			/* wait here in user state */
 }
 
-int win32_interruptible(void)
+static struct handler_arg_t* setup_handler(struct handler_arg_t *harg,
+					   int arg,
+					   void (*handler)(int),
+					   HANDLE handshake)
 {
-    if (catch_interrupt(0)) return TRUE;
-    SetEvent(rb_InterruptEvent);
-    return FALSE;
+    harg->handler = handler;
+    harg->arg = arg;
+    harg->status = 0;
+    harg->userstate = 0;
+    harg->handshake = handshake;
+    return harg;
+}
+
+static void setup_call(CONTEXT* ctx, struct handler_arg_t *harg)
+{
+#ifdef _M_IX86
+    DWORD *esp = (DWORD *)ctx->Esp;
+    *--esp = (DWORD)harg;
+    *--esp = ctx->Eip;
+    ctx->Esp = (DWORD)esp;
+    ctx->Eip = (DWORD)win32_call_handler;
+#else
+#error unsupported processor
+#endif
 }
 
 int win32_main_context(int arg, void (*handler)(int))
 {
-    if (!win32_thread_exclusive()) return FALSE;
+    static HANDLE interrupt_done = NULL;
+    struct handler_arg_t harg;
+    CONTEXT ctx_orig;
+    HANDLE current_thread = GetCurrentThread();
+    int old_priority = GetThreadPriority(current_thread);
 
-    if (!catch_interrupt(0)) {
-	SetEvent(rb_InterruptEvent);
-	return FALSE;
-    }
+    if (GetCurrentThreadId() == main_thread.id) return FALSE;
 
-    {
+    SetSignal(interrupted_event);
+
+    RUBY_CRITICAL({		/* the main thread must be in user state */
 	CONTEXT ctx;
 
+	SuspendThread(main_thread.handle);
+	SetThreadPriority(current_thread, GetThreadPriority(main_thread.handle));
+
 	ZeroMemory(&ctx, sizeof(CONTEXT));
 	ctx.ContextFlags = CONTEXT_FULL | CONTEXT_FLOATING_POINT;
-	GetThreadContext(rb_MainThreadHandle, &ctx);
-#ifdef _M_IX86
-	{
-	    DWORD *esp = (DWORD *)(ctx.Esp - sizeof(CONTEXT));
-	    *(CONTEXT *)esp = ctx;
-	    *--esp = (DWORD)handler;
-	    *--esp = arg;
-	    *--esp = ctx.Eip;
-	    ctx.Esp = (DWORD)esp;
+	GetThreadContext(main_thread.handle, &ctx);
+	ctx_orig = ctx;
+
+	/* handler context setup */
+	if (!interrupt_done) {
+	    interrupt_done = CreateEvent(NULL, FALSE, FALSE, NULL);
+	    /* anonymous one-shot event */
 	}
-	ctx.Eip = (DWORD)win32_call_handler;
-#else
-#error
-#endif
+	else {
+	    ResetEvent(interrupt_done);
+	}
+	setup_call(&ctx, setup_handler(&harg, arg, handler, interrupt_done));
 
 	ctx.ContextFlags = CONTEXT_CONTROL;
-	SetThreadContext(rb_MainThreadHandle, &ctx);
+	SetThreadContext(main_thread.handle, &ctx);
+	ResumeThread(main_thread.handle);
+    });
+
+    /* give a chance to the main thread */
+    yield_once();
+    WaitForSingleObject(interrupt_done, INFINITE); /* handshaking */
+
+    if (!harg.status) {
+	/* no exceptions raised, restore old context. */
+	RUBY_CRITICAL({
+	    /* ensure the main thread is in user state. */
+	    yield_until(harg.userstate);
+
+	    SuspendThread(main_thread.handle);
+	    ctx_orig.ContextFlags = CONTEXT_FULL | CONTEXT_FLOATING_POINT;
+	    SetThreadContext(main_thread.handle, &ctx_orig);
+	    ResumeThread(main_thread.handle);
+	});
     }
-    ResumeThread(rb_MainThreadHandle);
+    /* otherwise leave the main thread raised */
+
+    SetThreadPriority(current_thread, old_priority);
 
     return TRUE;
 }
 
-void win32_sleep(unsigned long msec)
+int win32_sleep(unsigned long msec)
 {
-    int trap_immediate = rb_trap_immediate;
+    return wait_events(NULL, msec) != WAIT_TIMEOUT;
+}
 
-    if (trap_immediate)
-	TRAP_END;
-    catch_interrupt(msec);
-    if (trap_immediate)
-	TRAP_BEG;
+static void catch_interrupt(void)
+{
+    yield_once();
+    win32_sleep(0);
+    CHECK_INTS;
+}
+
+void win32_enter_syscall(void)
+{
+    InterlockedExchange(&rb_trap_immediate, 1);
+    catch_interrupt();
+    win32_disable_interrupt();
+}
+
+void win32_leave_syscall(void)
+{
+    win32_enable_interrupt();
+    catch_interrupt();
+    InterlockedExchange(&rb_trap_immediate, 0);
+}
+
+struct asynchronous_arg_t {
+    /* output field */
+    void* stackaddr;
+
+    /* input field */
+    VALUE (*func)(VALUE self, int argc, VALUE* argv);
+    VALUE self;
+    int argc;
+    VALUE* argv;
+};
+
+static DWORD WINAPI
+call_asynchronous(PVOID argp)
+{
+    struct asynchronous_arg_t *arg = argp;
+    arg->stackaddr = &argp;
+    return (DWORD)arg->func(arg->self, arg->argc, arg->argv);
+}
+
+VALUE win32_asynchronize(asynchronous_func_t func,
+			 VALUE self, int argc, VALUE* argv, VALUE intrval)
+{
+    DWORD val;
+    BOOL interrupted = FALSE;
+    HANDLE thr;
+
+    RUBY_CRITICAL({
+	struct asynchronous_arg_t arg;
+
+	arg.stackaddr = NULL;
+	arg.func = func;
+	arg.self = self;
+	arg.argc = argc;
+	arg.argv = argv;
+
+	thr = CreateThread(NULL, 0, call_asynchronous, &arg, 0, &val);
+
+	if (thr) {
+	    yield_until(arg.stackaddr);
+
+	    if (wait_events(thr, INFINITE) != WAIT_OBJECT_0) {
+		interrupted = TRUE;
+
+		if (TerminateThread(thr, intrval)) {
+		    yield_once();
+		}
+	    }
+
+	    GetExitCodeThread(thr, &val);
+	    CloseHandle(thr);
+
+	    if (interrupted) {
+		/* must release stack of killed thread, why doesn't Windows? */
+		MEMORY_BASIC_INFORMATION m;
+
+		memset(&m, 0, sizeof(m));
+		if (!VirtualQuery(arg.stackaddr, &m, sizeof(m))) {
+		    Debug(fprintf(stderr, "couldn't get stack base:%p:%d\n",
+				  arg.stackaddr, GetLastError()));
+		}
+		else if (!VirtualFree(m.AllocationBase, 0, MEM_RELEASE)) {
+		    Debug(fprintf(stderr, "couldn't release stack:%p:%d\n",
+				  m.AllocationBase, GetLastError()));
+		}
+	    }
+	}
+    });
+
+    if (!thr) {
+	rb_fatal("failed to launch waiter thread:%d", GetLastError());
+    }
+
+    if (interrupted) {
+	errno = EINTR;
+	CHECK_INTS;
+    }
+
+    return val;
+}
+
+char **win32_get_environ(void)
+{
+    char *envtop, *env;
+    char **myenvtop, **myenv;
+    int num;
+
+    envtop = GetEnvironmentStrings();
+    for (env = envtop, num = 0; *env; env += strlen(env) + 1)
+	if (*env != '=') num++;
+
+    myenvtop = ALLOC_N(char*, num + 1);
+    for (env = envtop, myenv = myenvtop; *env; env += strlen(env) + 1) {
+	if (*env != '=') {
+	    *myenv = ALLOC_N(char, strlen(env) + 1);
+	    strcpy(*myenv, env);
+	    myenv++;
+	}
+    }
+    *myenv = NULL;
+    FreeEnvironmentStrings(envtop);
+
+    return myenvtop;
+}
+
+void win32_free_environ(char **env)
+{
+    char **t = env;
+
+    while (*t) free(*t++);
+    free(env);
+}
+
+int
+win32_fclose(FILE *fp)
+{
+    int fd = fileno(fp);
+    SOCKET sock = TO_SOCKET(fd);
+
+    if (fflush(fp)) return -1;
+    if (!is_socket(sock)) {
+	return fclose(fp);
+    }
+    _set_osfhnd(fd, (SOCKET)INVALID_HANDLE_VALUE);
+    fclose(fp);
+    if (closesocket(sock) == SOCKET_ERROR) {
+	errno = WSAGetLastError();
+	return -1;
+    }
+    return 0;
+}
+
+int
+win32_close(int fd)
+{
+    SOCKET sock = TO_SOCKET(fd);
+
+    if (!is_socket(sock)) {
+	return _close(fd);
+    }
+    if (closesocket(sock) == SOCKET_ERROR) {
+	errno = WSAGetLastError();
+	return -1;
+    }
+    return 0;
 }
