@@ -1,5 +1,5 @@
-#ifndef EXT_NT_H
-#define EXT_NT_H
+#ifndef RUBY_WIN32_H
+#define RUBY_WIN32_H
 
 /*
  *  Copyright (c) 1993, Intergraph Corporation
@@ -99,7 +99,11 @@ extern "C++" {
 #include <time.h>
 #include <math.h>
 #include <sys/types.h>
-#include <sys/utime.h>
+#if !defined(__BORLANDC__)
+# include <sys/utime.h>
+#else
+# include <utime.h>
+#endif
 #include <io.h>
 #include <malloc.h>
 
@@ -115,7 +119,8 @@ extern "C++" {
 #define access	   _access
 #define chmod	   _chmod
 #define chsize	   _chsize
-#define close	   _close
+#define close	   win32_close
+#define fclose	   win32_fclose
 #define creat	   _creat
 #define dup	   _dup
 #define dup2	   _dup2
@@ -162,6 +167,8 @@ extern "C++" {
 #define pclose     _pclose
 #define strcasecmp _stricmp
 #define strncasecmp _strnicmp
+#undef rb_sys_stat
+#define rb_sys_stat win32_stat
 /* these are defined in nt.c */
 
 #ifdef __MINGW32__
@@ -177,8 +184,6 @@ extern char *NtGetBin(void);
 extern FILE *mypopen(char *, char *);
 extern int   mypclose(FILE *);
 extern int  flock(int fd, int oper);
-extern FILE *  myfdopen(int, const char *);
-extern void  myfdclose(FILE *);
 extern SOCKET  myaccept(SOCKET, struct sockaddr *, int *);
 extern int  mybind(SOCKET, struct sockaddr *, int);
 extern int  myconnect(SOCKET, struct sockaddr *, int);
@@ -208,6 +213,8 @@ extern struct servent *  mygetservbyname(char *, char *);
 extern struct servent * mygetservbyport(int, char *);
 extern char *win32_getenv(const char *);
 extern int myrename(const char *, const char *);
+extern char **win32_get_environ(void);
+extern void win32_free_environ(char **);
 
 extern int chown(const char *, int, int);
 extern int link(char *, char *);
@@ -219,6 +226,36 @@ extern int isinf(double);
 extern int isnan(double);
 
 
+#if !defined S_IRUSR && !defined __MINGW32__
+#define S_IRUSR 0400
+#endif
+#ifndef S_IRGRP
+#define S_IRGRP 0040
+#endif
+#ifndef S_IROTH
+#define S_IROTH 0004
+#endif
+
+#if !defined S_IWUSR && !defined __MINGW32__
+#define S_IWUSR 0200
+#endif
+#ifndef S_IWGRP
+#define S_IWGRP 0020
+#endif
+#ifndef S_IWOTH
+#define S_IWOTH 0002
+#endif
+
+#if !defined S_IXUSR && !defined __MINGW32__
+#define S_IXUSR 0100
+#endif
+#ifndef S_IXGRP
+#define S_IXGRP 0010
+#endif
+#ifndef S_IXOTH
+#define S_IXOTH 0001
+#endif
+
 //
 // define this so we can do inplace editing
 //
@@ -228,7 +265,9 @@ extern int isnan(double);
 //
 // stubs
 //
+#if !defined(__BORLANDC__)
 extern int       ioctl (int, unsigned int, long);
+#endif
 extern UIDTYPE   getuid (void);
 extern UIDTYPE   geteuid (void);
 extern GIDTYPE   getgid (void);
@@ -267,11 +306,6 @@ extern char *mystrerror(int);
 /* #undef va_start */
 /* #undef va_end */
 
-#ifdef fdopen
-#undef fdopen
-#endif
-#define fdopen myfdopen
-
 #ifdef accept
 #undef accept
 #endif
@@ -427,17 +461,25 @@ struct tms {
 #define times mytimes
 
 /* thread stuff */
-/* initialized by NtInitialize() */
-HANDLE rb_CurrentProcessHandle;
-HANDLE rb_MainThreadHandle;
-HANDLE rb_InterruptEvent;
-DWORD rb_MainThreadId;
-
 HANDLE GetCurrentThreadHandle(void);
 int win32_main_context(int arg, void (*handler)(int));
-int win32_interruptible(void);
-void win32_thread_resume_main(void);
-void win32_sleep(unsigned long msec);
-#define Sleep(msec) win32_sleep(msec)
+int win32_sleep(unsigned long msec);
+void win32_enter_syscall(void);
+void win32_leave_syscall(void);
+void win32_disable_interrupt(void);
+void win32_enable_interrupt(void);
+int  win32_close(int);
+int  win32_fclose(FILE*);
+#define Sleep(msec) (void)win32_sleep(msec)
+
+/*
+== ***CAUTION***
+Since this function is very dangerous, ((*NEVER*))
+* lock any HANDLEs(i.e. Mutex, Semaphore, CriticalSection and so on) or,
+* use anything like TRAP_BEG...TRAP_END block structure,
+in asynchronous_func_t.
+*/
+typedef DWORD (*asynchronous_func_t)(DWORD self, int argc, DWORD* argv);
+DWORD win32_asynchronize(asynchronous_func_t func, DWORD self, int argc, DWORD* argv, DWORD intrval);
 
 #endif
